#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"

#include "php_yami.h"

#include "yamic.h"
#include "yamilimits.h"
#include "yamierrors.h"
#include "yamiparams.h"
#include "yamisynchro.h"

#ifdef COMPILE_DL_YAMI
    ZEND_GET_MODULE(yami)
#endif

/* {{{ PHP Yami Module Globals */
ZEND_DECLARE_MODULE_GLOBALS(yami)

static void php_yami_init_globals(zend_yami_globals *yami_globals)
{
    yami_globals->default_response_timeout = 5;
    yami_globals->last_error = YAMIOK;
}
/* }}} */

/* {{{ PHP INI Entries */
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("yami.response.timeout", "5", PHP_INI_ALL, OnUpdateLong, default_response_timeout, zend_yami_globals, yami_globals)
PHP_INI_END()
/* }}} */

/* ------------------------------------------------------------------------- */
/* {{{ PHP Yami Agent Resource */
typedef struct {
    HYAMIAGENT agent;
    struct yamipolicies policies;
} Y_AGENT_RSRC;
/* }}} */

/* {{{ PHP Yami Resource Constants */
#define Y_INBOUND_MESSAGE 1
#define Y_OUTBOUND_MESSAGE 2
#define Y_AGENT_R_NAME "YAMI Agent"
#define Y_MESSAGE_R_NAME "YAMI Message"
#define Y_OBJECT_R_NAME "YAMI Object"
/* }}} */

/* {{{ PHP Yami Message Resource */
typedef struct {
    union {
        HINCMSG  imessage;
        HMESSAGE omessage;
    };
    int mtype;
    int level;
    int options;
    HYAMIAGENT agent;
} Y_MESSAGE_RSRC;
/* }}} */

/* {{{ PHP Yami Object Resource */
typedef struct {
    zval *callback;
    zval *hint;
    HYAMIAGENT agent;
} Y_OBJECT_RSRC;
/* }}} */

/* {{{ PHP Yami Global Resource Identifiers (Zend List Entries)
    globals available across web server requests - according to some
    documentation it seems there is no need for thread safety here!
*/
static int le_y_agent, le_y_message, le_y_object;
/* }}} */

/* {{{ PHP Yami Extension Local Helper Functions */

/* {{{ y_create_object_rsrc - local function
    Creates object resource initialized with the agent,
    callback and hint.
*/
static Y_OBJECT_RSRC *y_create_object_rsrc(HYAMIAGENT yagent, zval *zcallback, zval *zhint)
{
    Y_OBJECT_RSRC *y_object_rsrc = emalloc(sizeof (Y_OBJECT_RSRC));
    y_object_rsrc->agent = 0;
    y_object_rsrc->callback = 0;
    y_object_rsrc->hint = 0;

    assert (yagent);
    /*assert (zcallback);*/

    if ( zcallback ) {
        MAKE_STD_ZVAL(y_object_rsrc->callback);
        *y_object_rsrc->callback = *zcallback;
        zval_copy_ctor(y_object_rsrc->callback);
    }

    if ( zhint ) {
        MAKE_STD_ZVAL(y_object_rsrc->hint);
        *y_object_rsrc->hint = *zhint;
        zval_copy_ctor(y_object_rsrc->hint);
    }

    y_object_rsrc->agent = yagent;

    return y_object_rsrc;
}
/* }}} */


/* {{{ y_destroy_object_rsrc - local function
    Destroys the object resource structure. It is an error to access this
    structure after this function call.
*/
static void y_destroy_object_rsrc(Y_OBJECT_RSRC *y_object_rsrc)
{
    assert(y_object_rsrc);

    if ( y_object_rsrc->callback ) {
        zval_dtor(y_object_rsrc->callback);
        y_object_rsrc->callback = 0;
    }

    if ( y_object_rsrc->hint ) {
        zval_dtor(y_object_rsrc->hint);
        y_object_rsrc->hint = 0;
    }

}
/* }}} */


/* {{{ y_strerror - local function */
static const char *y_strerror(int ec)
{
    const char *final_msg = "Unknown error";
    const char *msg = 0;
    if ( (msg = yamiErrorString(ec)) ) {
        final_msg = msg;
    }
    /* php_printf("y_strerror: ec = %d, em = %s\n", ec, final_msg); */
    return final_msg;
}
/* }}} */

/* {{{ y_message_status - local function */
static const char *y_message_status(int ec)
{
    const char *info = 0;

    switch (ec) {
    case ePosted:
        info = "posted"; break;
    case ePending:
        info = "pending";break;
    case eReplied:
        info = "replied";break;
    case eRejected:
        info = "rejected";break;
    case eUnknownObj:
        info = "unknown object";break;
    case eOverflow:
        info = "overflow";break;
    case eNetError:
        info = "network error";break;
    case eTimedOut:
        info = "time out";break;
    case eRejectByAgent:
        info = "rejected by agent";break;
    case eRejectByAgent + 1:
        info = "delivered";break;
    default:
        info = "bad error code";
    }

    return info;
}
/* }}} */


/* {{{ y_php_array_to_params - local function
    Returns HPARAMSET containing the PHP array elements.

    Currently only long, double and string parameters
    are supported. Strings are treated as binary data.

    The caller is responsible for deleting the return
    value using yamiDestroyParamSet() function.

    Note that the return value may be null;
*/
static HPARAMSET y_php_array_to_params(zval *array TSRMLS_DC)
{
    HPARAMSET yparams = 0;

    if ( array ) {
        HashTable *arr_hash     = HASH_OF(array);/*Z_ARRVAL_P(array);*/
        int arr_count           = zend_hash_num_elements(arr_hash);
        zval **data             = 0;
        HashPosition arr_pointer= 0;
        int ypp                 = 0; /* yami parameter position */

        /* php_printf("The array passed contains %d elements\n", arr_count); */
        if ( arr_count > 0 ) {
            yamiCreateParamSet(&yparams, arr_count);
            for (
                zend_hash_internal_pointer_reset_ex(arr_hash, &arr_pointer);
                zend_hash_get_current_data_ex(arr_hash, (void **) &data, &arr_pointer) == SUCCESS;
                zend_hash_move_forward_ex(arr_hash, &arr_pointer)
            ) {
                switch ( Z_TYPE_PP(data) ) {
                case IS_BOOL:
                    yamiSetByte(yparams, ypp, (char)Z_LVAL_PP(data));
                    break;
                case IS_LONG:
                    yamiSetInt(yparams, ypp, Z_LVAL_PP(data));
                    break;
                case IS_DOUBLE:
                    yamiSetDouble(yparams, ypp, Z_DVAL_PP(data));
                    break;
                case IS_STRING:
                    yamiSetBinary(yparams, ypp, Z_STRVAL_PP(data), Z_STRLEN_PP(data));
                    /*
                    yamiSetString(yparams, ypp, Z_STRVAL_PP(data));
                    */
                    break;
                default: /*case IS_NULL: case IS_RESOURCE: case IS_ARRAY: case IS_OBJECT:*/
                    {
                        /* this is only meant for trouble shooting */
                        zval temp;
                        temp = **data;
                        zval_copy_ctor(&temp);
                        convert_to_string(&temp);
                        yamiSetBinary(yparams, ypp, Z_STRVAL(temp), Z_STRLEN(temp));
                        /*
                        yamiSetString(yparams, ypp, Z_STRVAL(temp));
                        */
                        zval_dtor(&temp);
                    }
                }

                ++ypp;
            }
        }
    }

    return yparams;
}
/* }}} */

/* {{{ y_params_to_php_array - local function
    Copys the contents of YAMI parameter set to a PHP Array.

    Currently only long, double and string parameters
    are supported. Strings are treated as binary data.

    The caller is responsible for deleting the return
    value using yamiDestroyParamSet() function.

    Returns the number of elements copied to the PHP array.
*/
static int y_params_to_php_array(HPARAMSET ymparams, zval *zmparams TSRMLS_DC)
{
    /*assert(yparams);*/
    int ypi = 0, ypc = 0;

    if ( zmparams ) {
        YAMI_G(last_error) = yamiGetParamCount(ymparams, &ypc);

        if ( YAMI_G(last_error) == YAMIOK) {
            array_init(zmparams);
            /** @todo: Since we already know the type, I think we can skip
                error checking in the for loop below. Question for Maciej...
            */
            for (ypi = 0; ypi < ypc; ++ypi) {
                enum paramType ypt;
                YAMI_G(last_error) = yamiGetType(ymparams, ypi, &ypt);

                if ( YAMI_G(last_error) != YAMIOK)  break;

                switch (ypt) {
                case eByte:
                    {
                        char ybyte; /* is bool in php, well in my books! */
                        YAMI_G(last_error) = yamiGetByte(ymparams, ypi, &ybyte);
                        if ( YAMI_G(last_error) != YAMIOK)  break;
                        add_index_bool(zmparams, ypi, ybyte);
                    }
                    break;
                case eInt:
                    {
                        int yint;
                        YAMI_G(last_error) = yamiGetInt(ymparams, ypi, &yint);
                        if ( YAMI_G(last_error) != YAMIOK)  break;
                        add_index_long(zmparams, ypi, yint);
                    }
                    break;
                case eDouble:
                    {
                        double ydouble;
                        YAMI_G(last_error) = yamiGetDouble(ymparams, ypi, &ydouble);
                        if ( YAMI_G(last_error) != YAMIOK)  break;
                        add_index_double(zmparams, ypi, ydouble);
                    }
                    break;
                case eString:
                    {
                        const char *ystring;
                        YAMI_G(last_error) = yamiGetStringBuffer(ymparams, ypi, &ystring);
                        if ( YAMI_G(last_error) != YAMIOK)  break;
                        add_index_string(zmparams, ypi, (char *)ystring, 1);
                    }
                    break;
                case eBinary: default:
                    {
                        const char *ybinary = 0; int ybsize;

                        YAMI_G(last_error) = yamiGetBinarySize(ymparams, ypi, &ybsize);
                        if ( YAMI_G(last_error) != YAMIOK)  break;

                        YAMI_G(last_error) = yamiGetBinaryBuffer(ymparams, ypi, &ybinary);
                        if ( YAMI_G(last_error) != YAMIOK)  break;

                        add_index_stringl(zmparams, ypi, (char *)ybinary, ybsize, 1);
                    }
                    break;
                }
            }
        }
    }

    return ypi;
}
/* }}} */

/* {{{ y_free_agent - zend callback function to destroy Agent resource */
static void y_free_agent(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
    Y_AGENT_RSRC *y_agent_rsrc = (Y_AGENT_RSRC *)rsrc->ptr;
    if ( y_agent_rsrc->agent ) {
        /* php_printf("YAMI Agent is going bye-bye."); */
        yamiDestroyAgent(y_agent_rsrc->agent);
    }
    efree(y_agent_rsrc);
}
/* }}} */

/* {{{ y_free_message - zend callback function to destroy Message resource */
static void y_free_message(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
    Y_MESSAGE_RSRC *y_message_rsrc = (Y_MESSAGE_RSRC *)rsrc->ptr;
    /*if ( y_message_rsrc->agent ) {*/
        if ( y_message_rsrc->mtype == Y_OUTBOUND_MESSAGE ) {
            /* php_printf("YAMI Outbound Message is going bye-bye."); */
            yamiAgentMsgDestroy(y_message_rsrc->omessage);
            y_message_rsrc->omessage = 0;
        }

        if ( y_message_rsrc->mtype == Y_INBOUND_MESSAGE ) {
            /* php_printf("YAMI Inbound Message is going bye-bye."); */
            yamiAgentDestroyIncomingMsg(y_message_rsrc->imessage);
            y_message_rsrc->imessage = 0;
        }

    /*}*/
    efree(y_message_rsrc);
}
/* }}} */

/* {{{ y_free_object - zend callback function to destroy Object resource */
static void y_free_object(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
    Y_OBJECT_RSRC *y_object_rsrc = (Y_OBJECT_RSRC *)rsrc->ptr;

    /* php_printf("YAMI Object is going bye-bye."); */
    y_destroy_object_rsrc(y_object_rsrc);
}
/* }}} */


/* {{{ y_retrieve_policies - local function
    Retrieves policies from the user land array and merges
    them with the default policies of YAMI. The dispatchers
    policy is overridden in all cases to 0; this implies that
    user land is responsible for driving the event loop.
*/
static void y_retrieve_policies(struct yamipolicies *ypolicies, zval *zpolicies TSRMLS_DC)
{
    assert(ypolicies != 0);
    yamiGetDefaultPolicies(ypolicies);
    ypolicies->dispatchers = 0;

    if ( zpolicies ) {
        zval          **value;

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "agentlevel", sizeof("agentlevel"), (void **)&value) != FAILURE ) {
            ypolicies->agentlevel = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "allowduplex", sizeof("allowduplex"), (void **)&value) != FAILURE ) {
            ypolicies->allowduplex = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "connpoolsize", sizeof("connpoolsize"), (void **)&value) != FAILURE ) {
            ypolicies->connpoolsize = Z_LVAL_PP(value);
        }

        /*
        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "dispatchers", sizeof("dispatchers"), (void **)&value) != FAILURE ) {
            ypolicies->dispatchers = Z_LVAL_PP(value);
        }
        */

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "hasreceiver", sizeof("hasreceiver"), (void **)&value) != FAILURE ) {
            ypolicies->hasreceiver = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "hassender", sizeof("hassender"), (void **)&value) != FAILURE ) {
            ypolicies->hassender = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "hassocket", sizeof("hassocket"), (void **)&value) != FAILURE ) {
            ypolicies->hassocket = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "haswaker", sizeof("haswaker"), (void **)&value) != FAILURE ) {
            ypolicies->haswaker = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "maxdplxconns", sizeof("maxdplxconns"), (void **)&value) != FAILURE ) {
            ypolicies->maxdplxconns = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "mtuser", sizeof("mtuser"), (void **)&value) != FAILURE ) {
            ypolicies->mtuser = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "objqmaxlength", sizeof("objqmaxlength"), (void **)&value) != FAILURE ) {
            ypolicies->objqmaxlength = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "objqmaxsize", sizeof("objqmaxsize"), (void **)&value) != FAILURE ) {
            ypolicies->objqmaxsize = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "receiveridle", sizeof("receiveridle"), (void **)&value) != FAILURE ) {
            ypolicies->receiveridle = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "reuseaddr", sizeof("reuseaddr"), (void **)&value) != FAILURE ) {
            ypolicies->reuseaddr = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "sendtries", sizeof("sendtries"), (void **)&value) != FAILURE ) {
            ypolicies->sendtries = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "sqmaxlength", sizeof("sqmaxlength"), (void **)&value) != FAILURE ) {
            ypolicies->sqmaxlength = Z_LVAL_PP(value);
        }

        if ( zend_hash_find(Z_ARRVAL_P(zpolicies), "sqmaxsize", sizeof("sqmaxsize"), (void **)&value) != FAILURE ) {
            ypolicies->sqmaxsize = Z_LVAL_PP(value);
        }
    }

}
/* }}} */

/* {{{ y_fill_policies - local function
    Fills userland array with default policies of YAMI
*/
static void y_fill_policies(struct yamipolicies *ypolicies, zval *zpolicies TSRMLS_DC)
{
    struct yamipolicies default_ypolicies;
    assert(zpolicies != 0);

    if ( ypolicies == 0 ) {
        yamiGetDefaultPolicies(&default_ypolicies);
        ypolicies = &default_ypolicies;
    }

    add_assoc_long(zpolicies, "agentlevel", ypolicies->agentlevel);
    add_assoc_bool(zpolicies, "allowduplex", ypolicies->allowduplex);
    add_assoc_long(zpolicies, "connpoolsize", ypolicies->connpoolsize);

    add_assoc_long(zpolicies, "dispatchers", 0/*ypolicies->dispatchers*/);

    add_assoc_bool(zpolicies, "hasreceiver", ypolicies->hasreceiver);
    add_assoc_bool(zpolicies, "hassender", ypolicies->hassender);
    add_assoc_bool(zpolicies, "hassocket", ypolicies->hassocket);
    add_assoc_bool(zpolicies, "haswaker", ypolicies->haswaker);
    add_assoc_long(zpolicies, "maxdplxconns", ypolicies->maxdplxconns);
    add_assoc_bool(zpolicies, "mtuser", ypolicies->mtuser);
    add_assoc_long(zpolicies, "objqmaxlength", ypolicies->objqmaxlength);
    add_assoc_long(zpolicies, "objqmaxsize", ypolicies->objqmaxsize);
    add_assoc_long(zpolicies, "receiveridle", ypolicies->receiveridle);
    add_assoc_bool(zpolicies, "reuseaddr", ypolicies->reuseaddr);
    add_assoc_long(zpolicies, "sendtries", ypolicies->sendtries);
    add_assoc_long(zpolicies, "sqmaxlength", ypolicies->sqmaxlength);
    add_assoc_long(zpolicies, "sqmaxsize", ypolicies->sqmaxsize);
}
/* }}} */

/* {{{ y_fill_agent_info - local function
    Fills userland array with agent information
*/
static void y_fill_agent_info(Y_AGENT_RSRC *y_agent_rsrc, zval *return_value, zend_bool zdetailed TSRMLS_DC)
{
    zval *zpolicies = 0;
    char host[64];
    int port;
    if ( yamiAgentGetLocalAddress(y_agent_rsrc->agent, host) != YAMIOK ) {
        strncpy(host, "bad local host", sizeof("bad local host"));
    }

    yamiAgentGetLocalPort(y_agent_rsrc->agent, &port);

    add_assoc_string(return_value, "host", host, 1);
    add_assoc_long(return_value, "port", port);

    if ( zdetailed ) {
        ALLOC_INIT_ZVAL(zpolicies);
        array_init(zpolicies);
        y_fill_policies(&y_agent_rsrc->policies, zpolicies TSRMLS_CC);
        add_assoc_zval(return_value, "policies", zpolicies);
    }
}
/* }}} */

/* {{{ y_passive_object_callback - local function
    Adapter to make an upcall into user land.

    The signature for this function matches:
        typedef void (*DISPFUNCTION)(HINCMSG inc);

    This function is called for a given object (if registered as passive)
    when an incoming message arrives.

    The message will be considered rejected, unless the dispatch function
    has processed the message in some way. It means, that if the dispatch
    function did not process the message, it will be automatically rejected
    and the client (calling object) will receive appropriate notification.

    The incoming message structure (handled by inc) is owned by the Agent,
    so the dispatch function should not release it by itself.
*/
static void y_passive_object_callback(HINCMSG inc)
{
   /** @todo PHP Yami servers cannot be created in this version */
}
/* }}} */

/* }}} */

/* ------------------------------------------------------------------------- */

/* {{{ PHP Yami Userland Function Entry Points */
static function_entry yami_functions[] = {
    PHP_FE(yami_get_last_error, NULL)
    PHP_FE(yami_get_error_string, NULL)
    PHP_FE(yami_send_to, NULL)
    PHP_FE(yami_get_default_policies, NULL)
    PHP_FE(yami_agent_create, NULL)
    PHP_FE(yami_agent_destroy, NULL)
    PHP_FE(yami_agent_get_info, NULL)
    PHP_FE(yami_agent_domain_register, NULL)
    PHP_FE(yami_agent_domain_unregister, NULL)
    PHP_FE(yami_agent_object_register, NULL)
    PHP_FE(yami_agent_object_unregister, NULL)
    PHP_FE(yami_agent_send, NULL)
    PHP_FE(yami_agent_send_to, NULL)
    PHP_FE(yami_message_setup_timeout, NULL)
    PHP_FE(yami_message_wait, NULL)
    PHP_FE(yami_message_get_response, NULL)
    PHP_FE(yami_agent_receive, NULL)
    PHP_FE(yami_message_consume, NULL)
    PHP_FE(yami_message_reject, NULL)
    PHP_FE(yami_message_reply, NULL)
    PHP_FE(yami_message_forward, NULL)
    PHP_FE(yami_message_forward_to, NULL)
    PHP_FE(yami_message_destroy, NULL)
    PHP_FE(yami_message_get_request, NULL)
    PHP_FE(yami_message_get_type, NULL)
    PHP_FE(yami_message_get_info, NULL)
    PHP_FE(yami_message_get_status, NULL)
    PHP_FE(yami_message_get_status_string, NULL)

    {NULL, NULL, NULL}
};
/* }}} */

/* {{{ PHP Yami Module Entry */
zend_module_entry yami_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
    STANDARD_MODULE_HEADER,
#endif
    PHP_YAMI_EXT_NAME,
    yami_functions,
    PHP_MINIT(yami),
    PHP_MSHUTDOWN(yami),
    PHP_RINIT(yami),
    PHP_RSHUTDOWN(yami),
    PHP_MINFO(yami),
#if ZEND_MODULE_API_NO >= 20010901
    PHP_YAMI_EXT_VERSION,
#endif
    STANDARD_MODULE_PROPERTIES
};
/* }}} */


/* {{{ PHP_MINIT_FUNCTION - called when extension is loaded. */
PHP_MINIT_FUNCTION(yami)
{
    ZEND_INIT_MODULE_GLOBALS(yami, php_yami_init_globals, NULL);
    REGISTER_INI_ENTRIES();

    REGISTER_LONG_CONSTANT("YAMI_OK"                    ,YAMIOK                     ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_BADHANDLE"             ,YAMIBADHANDLE              ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_OUTOFRANGE"            ,YAMIOUTOFRANGE             ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_CANNOTINITIALIZETWICE" ,YAMICANNOTINITIALIZETWICE  ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_BADTYPE"               ,YAMIBADTYPE                ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_NOTINITIALIZED"        ,YAMINOTINITIALIZED         ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_ALREADYREGISTERED"     ,YAMIALREADYREGISTERED      ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_NOTFOUND"              ,YAMINOTFOUND               ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_QUEUEEMPTY"            ,YAMIQUEUEEMPTY             ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_PARSEERROR"            ,YAMIPARSEERROR             ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_CANNOTPROCESSTWICE"    ,YAMICANNOTPROCESSTWICE     ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_BADLENGTH"             ,YAMIBADLENGTH              ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_NOSUCHSERVICE"         ,YAMINOSUCHSERVICE          ,CONST_CS|CONST_PERSISTENT);

    REGISTER_LONG_CONSTANT("YAMI_NOTAVAILABLE"          ,YAMINOTAVAILABLE           ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_NOMEMORY"              ,YAMINOMEMORY               ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_CANNOTRESOLVEADDRESS"  ,YAMICANNOTRESOLVEADDRESS   ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_BADADDRTYPE"           ,YAMIBADADDRTYPE            ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_CANNOTSTARTTHREAD"     ,YAMICANNOTSTARTTHREAD      ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_THREADERROR"           ,YAMITHREADERROR            ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_SYNCHROERROR"          ,YAMISYNCHROERROR           ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_NETERROR"              ,YAMINETERROR               ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_OVERFLOW"              ,YAMIOVERFLOW               ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_INTERNAL"              ,YAMIINTERNAL               ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_NOHANDSHAKE"           ,YAMINOHANDSHAKE            ,CONST_CS|CONST_PERSISTENT);

    REGISTER_LONG_CONSTANT("YAMI_MSG_POSTED"            ,ePosted                    ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_PENDING"           ,ePending                   ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_REPLIED"           ,eReplied                   ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_REJECTED"          ,eRejected                  ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_UNKNOWNOBJECT"     ,eUnknownObj                ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_OVERFLOW"          ,eOverflow                  ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_NETERROR"          ,eNetError                  ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_TIMEDOUT"          ,eTimedOut                  ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_REJECTBYAGENT"     ,eRejectByAgent             ,CONST_CS|CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("YAMI_MSG_RECEIVED"          ,eRejectByAgent + 1         ,CONST_CS|CONST_PERSISTENT);

    le_y_agent      = zend_register_list_destructors_ex(y_free_agent, NULL, Y_AGENT_R_NAME, module_number);
    le_y_message    = zend_register_list_destructors_ex(y_free_message, NULL, Y_MESSAGE_R_NAME, module_number);
    le_y_object     = zend_register_list_destructors_ex(y_free_object, NULL, Y_OBJECT_R_NAME, module_number);
    yamiNetInitialize();

    return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION - called when extension is being unloaded. */
PHP_MSHUTDOWN_FUNCTION(yami)
{
    UNREGISTER_INI_ENTRIES();
    yamiNetCleanup();
    return SUCCESS;
}
/* }}} */

/* {{{ PHP_RINIT_FUNCTION - called when a request arrives, possibly concurrently */
PHP_RINIT_FUNCTION(yami)
{
    YAMI_G(last_error) = YAMIOK;
    return SUCCESS;
}
/* }}} */

/* {{{ PHP_RSHUTDOWN_FUNCTION - called when request processing is complete. */
PHP_RSHUTDOWN_FUNCTION(yami)
{
    return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION - typically called by phpinfo() */
PHP_MINFO_FUNCTION(yami)
{
    php_info_print_table_start();
    php_info_print_table_header(2, "YAMI support", "enabled");
    php_info_print_table_row(2, "YAMI (PHP Module) Version", PHP_YAMI_EXT_VERSION " - 200706260827");
    php_info_print_table_row(2, "YAMI (PHP Module) License", "http://www.php.net/license/3_0.txt");
    php_info_print_table_row(2, "YAMI (PHP Module) Copyright", "Copyright (C) 2005-2007 Vijay Aswadhati");        
    php_info_print_table_row(2, "YAMI Library Version", "3.1.0");
    php_info_print_table_row(2, "YAMI Library License", "http://www.msobczak.com/prog/yami/copyright.html");
    php_info_print_table_end();

    DISPLAY_INI_ENTRIES();
}
/* }}} */

/* {{{ PHP Yami Userland Functions */

/* {{{ proto long yami_get_last_error()
   Returns the last error that occured during an operation that typically
   returns a valid object/resource/array, but returned null value or false.
*/
PHP_FUNCTION(yami_get_last_error)
{
    RETURN_LONG(YAMI_G(last_error));
}
/* }}} */

/* {{{ proto string yami_get_error_string(int $ecode)
   Returns a string that contains human readable error message corresponding
   to the error code returned by a YAMI operation.
*/
PHP_FUNCTION(yami_get_error_string)
{
    long zerror;

    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &zerror
    );

    if ( rc == FAILURE ) {
        RETURN_FALSE;
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)zerror), 1);
}
/* }}} */

/* {{{ proto bool yami_send_to(string $host, int $port, string $oname, string $mname[, array $mparams])
   Sends a message to a remote object offering service on a port located
   at an address. The message can have optional parameters.
*/
PHP_FUNCTION(yami_send_to)
{
    char *ip, *oname, *mname;
    int ip_len, oname_len, mname_len;
    long port;
    zval *mparams = 0;

    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "slss|a!",
        &ip, &ip_len,
        &port,
        &oname, &oname_len,
        &mname, &mname_len,
        &mparams
    );

    if ( rc != FAILURE ) {
        if ( mparams && Z_TYPE_P(mparams) != IS_ARRAY) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING,"yami_send_to: fifth argument should be integer indexed array");
            YAMI_G(last_error) = YAMIBADTYPE;
        }
        else {
            HPARAMSET yparams = y_php_array_to_params(mparams TSRMLS_CC);

            /* See TODO for notes about PHP strings */
            YAMI_G(last_error) = yamiEasySend(ip, port, 2, oname, mname, yparams);

            if ( yparams ) {
                yamiDestroyParamSet(yparams);
            }
        }
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto array yami_get_default_policies()
   Returns default policies. Each policy is indexed by a name.
*/
PHP_FUNCTION(yami_get_default_policies)
{
    array_init(return_value);
    y_fill_policies(NULL, return_value TSRMLS_CC);
}
/* }}} */

/* {{{ proto resource yami_agent_create([int $port = 0[, array $policies = null]])
   Creates a YAMI agent with specified policies at specified service port.
   Both port and policies are optional. Note that in YAMI there is no
   notion of a server or client. An agent can be client or server or both;
   it is the objects that register with the agent that can claim as being
   client or server or both.
*/
PHP_FUNCTION(yami_agent_create)
{
    zval *zpolicies = 0;
    long zport = 0;

    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "|la!",
        &zport,
        &zpolicies
    );

    if ( rc != FAILURE ) {
        if ( zpolicies && Z_TYPE_P(zpolicies) != IS_ARRAY) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING,"yami_agent_create: second argument should be an associative array");
            rc = YAMIBADTYPE;
        }
        else {
            struct yamipolicies ypolicies;
            HYAMIAGENT yagent;

            y_retrieve_policies(&ypolicies, zpolicies TSRMLS_CC);

            YAMI_G(last_error) = yamiCreateAgent(&yagent, zport, &ypolicies);

            if ( YAMI_G(last_error) != YAMIOK ) {
                php_error_docref(NULL TSRMLS_CC, E_WARNING,"yami_agent_create: unable to create agent");
                RETURN_NULL();
            }
            else {
                Y_AGENT_RSRC *y_agent_rsrc = emalloc(sizeof (Y_AGENT_RSRC));
                y_agent_rsrc->agent = yagent;
                y_agent_rsrc->policies = ypolicies;
                ZEND_REGISTER_RESOURCE(return_value, y_agent_rsrc, le_y_agent);
            }
        }
    }
}
/* }}} */

/* {{{ proto bool yami_agent_destroy(resource $agent)
   Destroys the Agent resource. All messages queued to all servants are
   destroyed as well.
*/
PHP_FUNCTION(yami_agent_destroy)
{
    zval *zagent = 0;

    Y_AGENT_RSRC *y_agent_rsrc = 0;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "r",
        &zagent
    );

    if ( rc == FAILURE || Z_TYPE_P(zagent) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        ZEND_FETCH_RESOURCE(y_agent_rsrc, Y_AGENT_RSRC *, &zagent, -1, Y_AGENT_R_NAME, le_y_agent);
        ZEND_VERIFY_RESOURCE(y_agent_rsrc);
        YAMI_G(last_error) = yamiDestroyAgent(y_agent_rsrc->agent);
        y_agent_rsrc->agent = 0;
        zend_list_delete(Z_RESVAL_P(zagent));
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto array yami_agent_get_info(resource $agent[, bool $detailed = false])
   Returns information about the Agent resource. If detailed is true then the
   return value (an array) will also contain the policies that was used to
   create the Agent.
*/
PHP_FUNCTION(yami_agent_get_info)
{
    zend_bool zdetailed = 0;
    zval *zagent = 0;

    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "r|b",
        &zagent,
        &zdetailed
    );

    if ( rc == FAILURE || Z_TYPE_P(zagent) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        Y_AGENT_RSRC *y_agent_rsrc;
        ZEND_FETCH_RESOURCE(y_agent_rsrc, Y_AGENT_RSRC *, &zagent, -1, Y_AGENT_R_NAME, le_y_agent);
        ZEND_VERIFY_RESOURCE(y_agent_rsrc);
        array_init(return_value);
        y_fill_agent_info(y_agent_rsrc, return_value, zdetailed TSRMLS_CC);
    }
}
/* }}} */

/* {{{ proto bool yami_agent_domain_register(resource $agent, string $domain, string $host, int $port[, int $level=2])
   Associates the given host, service port and the receiving Agent's message
   specification level to a domain in the local Agent's white pages such that
   future correspondence can happen using a domain name.
*/
PHP_FUNCTION(yami_agent_domain_register)
{
    zval *zagent = 0;
    char *zdomain, *zhost;
    int zdomain_len, zhost_len;
    long zport, zlevel = 2;

    Y_AGENT_RSRC *y_agent_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "rssl|l",
        &zagent,
        &zdomain,
        &zdomain_len,
        &zhost,
        &zhost_len,
        &zport,
        &zlevel
    );

    if ( rc == FAILURE || Z_TYPE_P(zagent) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        ZEND_FETCH_RESOURCE(y_agent_rsrc, Y_AGENT_RSRC *, &zagent, -1, Y_AGENT_R_NAME, le_y_agent);
        ZEND_VERIFY_RESOURCE(y_agent_rsrc);
        YAMI_G(last_error) = yamiAgentDomainRegister(y_agent_rsrc->agent, zdomain, zhost, zport, zlevel);
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto bool yami_agent_domain_unregister(resource $agent, string $domain)
   Unregisters a domain from the internal white pages of the local Agent.
*/
PHP_FUNCTION(yami_agent_domain_unregister)
{
    zval *zagent = 0;
    char *zdomain;
    int zdomain_len;

    Y_AGENT_RSRC *y_agent_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "rs",
        &zagent,
        &zdomain,
        &zdomain_len
    );

    if ( rc == FAILURE || Z_TYPE_P(zagent) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        ZEND_FETCH_RESOURCE(y_agent_rsrc, Y_AGENT_RSRC *, &zagent, -1, Y_AGENT_R_NAME, le_y_agent);
        ZEND_VERIFY_RESOURCE(y_agent_rsrc);
        YAMI_G(last_error) = yamiAgentDomainUnregister(y_agent_rsrc->agent, zdomain);
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto bool yami_agent_object_register(resource $agent, string $oname[, mixed $callback = null[, mixed $hint = null]]])
   Registers an object by name. In general the local agent accepts only
   messages that have a destination object name that is registered with
   the agent. This method basically tells the agent to accept and queue
   all messages that are destined for an object with this name. Note
   that there can be several objects (with unique names) registered with
   the agent and each gets its own queue (sort of!).

   If callback is not null then the object is considered to be passive; the
   callback is invoked by the event loop for each inbound message. The
   callback gets the message and the hint parameter provided to the agent
   during registration. The hint is a opaque handle not inspected by the
   agent and may be null.

   If callback is null then the object is considered to be polling and is
   responsible for getting messages itself.
*/
PHP_FUNCTION(yami_agent_object_register)
{
    zval *zagent = 0, *zcallback = 0, *zhint = 0;
    char *zoname, *callback_name;
    int zoname_len;

    Y_AGENT_RSRC *y_agent_rsrc = 0;
    Y_OBJECT_RSRC *y_object_rsrc = 0;
    enum objectType y_object_type = polling;
    DISPFUNCTION y_dispatch_function = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "rs|z!z!",
        &zagent,
        &zoname,
        &zoname_len,
        &zcallback,
        &zhint
    );

    if ( rc == FAILURE || Z_TYPE_P(zagent) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        ZEND_FETCH_RESOURCE(y_agent_rsrc, Y_AGENT_RSRC *, &zagent, -1, Y_AGENT_R_NAME, le_y_agent);
        ZEND_VERIFY_RESOURCE(y_agent_rsrc);

        if ( !zcallback ) {
            y_object_rsrc       = y_create_object_rsrc(y_agent_rsrc->agent, NULL, zhint);
            y_object_type       = polling;
            y_dispatch_function = 0;
        }
        else {
            /* verify that the callback will work (if not null) */
            if ( !zend_is_callable(zcallback, 0, &callback_name) ) {
                php_error_docref1 (
                    NULL TSRMLS_CC,
                    callback_name,
                    E_WARNING,
                    "yami_agent_object_register: third argument is not a valid callback"
                );
                YAMI_G(last_error) = YAMIBADTYPE;
            }
            else {
                y_object_rsrc = y_create_object_rsrc(y_agent_rsrc->agent, zcallback, zhint);
                y_object_type = passive_singlethreaded;
                y_dispatch_function = y_passive_object_callback;
            }
            efree(callback_name);
        }

        YAMI_G(last_error) = yamiAgentObjectRegister(y_agent_rsrc->agent, zoname, y_object_type, y_dispatch_function, y_object_rsrc);

        if ( YAMI_G(last_error) != YAMIOK ) {
            php_error_docref1(NULL TSRMLS_CC, zoname, E_WARNING,"yami_agent_object_register: unable to register object");

            if ( y_object_rsrc ) {
                y_destroy_object_rsrc(y_object_rsrc);
                y_object_rsrc = 0;
            }
        }
        else {
            /* v.a:
            is this a hack? I don't know how other extensions handle this
            use case wherein the resource is not managed by the caller.
            */
            char soid[MAXNAMELEN + 64];
            zval *zoid = 0;
            int oid = ZEND_REGISTER_RESOURCE(NULL, y_object_rsrc, le_y_object);

            sprintf(soid, "php.yami.object:%s:%d", zoname, le_y_object);
            MAKE_STD_ZVAL(zoid);
            Z_RESVAL_P(zoid) = oid;
            Z_TYPE_P(zoid) = IS_RESOURCE;

            /*php_printf("registered oid %d as %s with ZEND. y_object_rsrc = %p\n", oid, soid, y_object_rsrc);*/
            zend_hash_update(&EG(symbol_table), soid, strlen(soid)+1, &zoid, sizeof(zval *), NULL);
        }
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto long yami_agent_object_unregister(resource $agent, string $oname)
   Unregisters an object from the local agent. All waiting messages and those
   yet to arrive addressed to this object will be discarded and
   'unknown object' reply message will be sent to the sender of the message.
*/
PHP_FUNCTION(yami_agent_object_unregister)
{
    zval *zagent = 0;
    char *zoname;
    int zoname_len;

    Y_AGENT_RSRC *y_agent_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "rs",
        &zagent,
        &zoname,
        &zoname_len
    );

    if ( rc == FAILURE || Z_TYPE_P(zagent) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        ZEND_FETCH_RESOURCE(y_agent_rsrc, Y_AGENT_RSRC *, &zagent, -1, Y_AGENT_R_NAME, le_y_agent);
        ZEND_VERIFY_RESOURCE(y_agent_rsrc);

        YAMI_G(last_error) = yamiAgentObjectUnregister(y_agent_rsrc->agent, zoname);

        if ( YAMI_G(last_error) != YAMIOK ) {
            php_error_docref1(NULL TSRMLS_CC, zoname, E_WARNING, "yami_agent_object_unregister: unable to register object");
        }
        else {
            /* v.a:
            is this a hack? I don't know how other extensions handle this
            use case wherein the resource is not managed by the caller.
            */
            char soid[MAXNAMELEN + 64];
            zval **zoid = 0;
            Y_OBJECT_RSRC *y_object_rsrc = 0;

            sprintf(soid, "php.yami.object:%s:%d", zoname, le_y_object);

            if ( zend_hash_find(&EG(symbol_table), soid, strlen(soid)+1, (void**)&zoid) == FAILURE ) {
                php_error_docref1 (
                    NULL TSRMLS_CC,
                    soid,
                    E_WARNING,
                    "yami_agent_object_unregister: unable to locate internal object resource id."
                );
                YAMI_G(last_error) = YAMIINTERNAL;
            }
            else {
                /* php_printf("unregistering oid %d identified by  %s from ZEND.\n", Z_RESVAL_PP(zoid), soid); */
                ZEND_FETCH_RESOURCE(y_object_rsrc, Y_OBJECT_RSRC *, zoid, -1, Y_OBJECT_R_NAME, le_y_object);
                ZEND_VERIFY_RESOURCE(y_object_rsrc);

                zend_list_delete(Z_RESVAL_PP(zoid));
            }
        }
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto mixed yami_agent_send(resource $agent, string $domain, string $oname, string $mname[, array $mparams = null[,bool oneway = false]])
    Sends a message to a remote object identified in the white pages of the
    agent by the domain name. The argument mparams can be null.

    If oneway is false then the message is considered to be of type one-way
    and no outbound message is created. Otherwise the outbound message
    that is created is returned.

    The return value is mixed:
        - for one way message (oneway = true), return value is true on success
          and false on error or failure.
        - for two way message (oneway = false), return value is message resource
          or null.
*/
PHP_FUNCTION(yami_agent_send)
{
    zval *zagent = 0, *zmparams = 0;
    zend_bool zoneway = 0;
    char *zdomain, *zoname, *zmname;
    int zdomain_len, zoname_len, zmname_len;

    Y_AGENT_RSRC *y_agent_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "rsss|a!b",
        &zagent,
        &zdomain,
        &zdomain_len,
        &zoname,
        &zoname_len,
        &zmname,
        &zmname_len,
        &zmparams,
        &zoneway
    );

    if ( rc != FAILURE ) {
        if ( zagent && Z_TYPE_P(zagent) != IS_RESOURCE ) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING,"yami_send: first argument should be valid agent resource");
            YAMI_G(last_error) = YAMIBADTYPE;
        }
        else if ( zmparams && Z_TYPE_P(zmparams) != IS_ARRAY) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING,"yami_send: fifth argument should be integer indexed array");
            YAMI_G(last_error) = YAMIBADTYPE;
        }
        else {
            /* See TODO for notes about PHP strings */
            HPARAMSET yparams = y_php_array_to_params(zmparams TSRMLS_CC);
            HMESSAGE ymessage = 0, *ymessage_ptr = 0;
            ZEND_FETCH_RESOURCE(y_agent_rsrc, Y_AGENT_RSRC *, &zagent, -1, Y_AGENT_R_NAME, le_y_agent);
            ZEND_VERIFY_RESOURCE(y_agent_rsrc);

            if ( !zoneway ) {
                ymessage_ptr = &ymessage;
            }

            YAMI_G(last_error) = yamiAgentMsgSend (
                y_agent_rsrc->agent,
                zdomain,
                zoname,
                zmname,
                yparams,
                ymessage_ptr
            );

            if ( yparams ) {
                yamiDestroyParamSet(yparams);
            }

            if ( YAMI_G(last_error) != YAMIOK ) {
                php_error_docref2(NULL TSRMLS_CC, zoname, zmname, E_WARNING, "yami_agent_send: error in send");
            }
            else {
                if ( !zoneway && ymessage != 0 ) {
                    Y_MESSAGE_RSRC *y_message_rsrc = emalloc(sizeof (Y_MESSAGE_RSRC));
                    y_message_rsrc->omessage = ymessage;
                    y_message_rsrc->mtype = Y_OUTBOUND_MESSAGE;
                    y_message_rsrc->level = y_agent_rsrc->policies.agentlevel;
                    y_message_rsrc->options = 0;
                    y_message_rsrc->agent = y_agent_rsrc->agent;
                    ZEND_REGISTER_RESOURCE(return_value, y_message_rsrc, le_y_message);
                }
            }
        }
    }

    if ( zoneway ) {
        RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
    }
    else {
        if ( YAMI_G(last_error) != YAMIOK ) {
            RETURN_NULL();
        }
    }
}
/* }}} */

/* {{{ proto mixed yami_agent_send_to(resource $agent, string $host, int $port, string $oname, string $mname[, array $mparams = null[, bool oneway = false[, bool $duplex = false[, int $mlevel = 2]]]])
    Sends a message to a remote object identified in the white pages of the
    agent by the domain name. The argument mparams can be null.

    The message argument receives a handle to the outbound message structure.
    The client code can use this handle to retrieve the message status and
    and any replies sent by the remote agent. If this argument is null,
    no message structure is created the outbound message is considered
    to be of type one-way.

    The message will be sent at message spec. level 2 by default and in
    simplex mode. Both can be overridden by the user land caller.

    The return value is mixed:
        - for one way message (oneway = true), return value is true on success
          and false on error or failure.
        - for two way message (oneway = false), return value is message resource
          or null.

*/
PHP_FUNCTION(yami_agent_send_to)
{
    zval *zagent = 0, *zmparams = 0;
    zend_bool zoneway = 0, zduplex = 0;
    char *zhost, *zoname, *zmname;
    int zhost_len, zoname_len, zmname_len;
    int zport, zslevel = 2;

    Y_AGENT_RSRC *y_agent_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "rslss|a!bbl",
        &zagent,
        &zhost,
        &zhost_len,
        &zport,
        &zoname,
        &zoname_len,
        &zmname,
        &zmname_len,
        &zmparams,
        &zoneway,
        &zduplex,
        &zslevel
    );

    if ( rc != FAILURE ) {
        if ( zagent && Z_TYPE_P(zagent) != IS_RESOURCE ) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING,"yami_send_to: first argument should be valid agent resource");
            YAMI_G(last_error) = YAMIBADTYPE;
        }
        else if ( zmparams && Z_TYPE_P(zmparams) != IS_ARRAY) {
            php_error_docref(NULL TSRMLS_CC, E_WARNING,"yami_send_to: sixth argument should be integer indexed array");
            YAMI_G(last_error) = YAMIBADTYPE;
        }
        else {
            /* See TODO for notes about PHP strings */
            HPARAMSET yparams = y_php_array_to_params(zmparams TSRMLS_CC);
            HMESSAGE ymessage = 0, *ymessage_ptr = 0;
            ZEND_FETCH_RESOURCE(y_agent_rsrc, Y_AGENT_RSRC *, &zagent, -1, Y_AGENT_R_NAME, le_y_agent);
            ZEND_VERIFY_RESOURCE(y_agent_rsrc);

            if ( !zoneway ) {
                ymessage_ptr = &ymessage;
            }

            YAMI_G(last_error) = yamiAgentMsgSendAddrEx (
                y_agent_rsrc->agent,
                zhost,
                zport,
                zslevel,
                zoname,
                zmname,
                yparams,
                ymessage_ptr,
                zduplex
            );

            if ( yparams ) {
                yamiDestroyParamSet(yparams);
            }

            if ( YAMI_G(last_error) != YAMIOK ) {
                php_error_docref2(NULL TSRMLS_CC, zoname, zmname, E_WARNING,"yami_agent_send_to: error in send");
            }
            else {
                if ( !zoneway && ymessage != 0 ) {
                    Y_MESSAGE_RSRC *y_message_rsrc = emalloc(sizeof (Y_MESSAGE_RSRC));
                    y_message_rsrc->omessage = ymessage;
                    y_message_rsrc->mtype = Y_OUTBOUND_MESSAGE;
                    y_message_rsrc->level = y_agent_rsrc->policies.agentlevel;
                    y_message_rsrc->options = 0;
                    y_message_rsrc->agent = y_agent_rsrc->agent;
                    ZEND_REGISTER_RESOURCE(return_value, y_message_rsrc, le_y_message);
                }
            }
        }
    }

    if ( zoneway ) {
        RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
    }
    else {
        if ( YAMI_G(last_error) != YAMIOK ) {
            RETURN_NULL();
        }
    }
}
/* }}} */

/* {{{ proto bool yami_message_setup_timeout(resource $message[, int $timeout=yami.default.response.timeout])
   Sets the timeout for a given outbound message. The timeout holds the number
   of seconds, after which the waker module will set the status of the
   message to YAMI_MSG_TIMEDOUT.

   The default timeout is picked up from the INI setting: yami.default.response.timeout
*/
PHP_FUNCTION(yami_message_setup_timeout)
{
    zval *zmessage = 0;
    int ztimeout = YAMI_G(default_response_timeout);

    Y_MESSAGE_RSRC *y_message_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "r|l",
        &zmessage,
        &ztimeout
    );

    if ( rc == FAILURE || Z_TYPE_P(zmessage) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        ZEND_FETCH_RESOURCE(y_message_rsrc, Y_MESSAGE_RSRC *, &zmessage, -1, Y_MESSAGE_R_NAME, le_y_message);
        ZEND_VERIFY_RESOURCE(y_message_rsrc);
        /* @todo the checks are not really needed because YAMI has a very good
            internal RTTI type checking system. Is this a good strategy?
        */
        /*
        if ( y_message_rsrc->mtype == Y_OUTBOUND_MESSAGE ) {
        */
            YAMI_G(last_error) = yamiAgentSetUpTimeOut(y_message_rsrc->omessage, ztimeout);
        /*
        }
        */
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto bool yami_message_wait(resource $message)
   Blocks the calling thread until the status of the message changes to
   one of: YAMI_MSG_REPLIED, YAMI_MSG_REJECTED, YAMI_MSG_UNKNOWN_OBJ,
   YAMI_MSG_OVERFLOW, YAMI_MSG_NET_ERROR, YAMI_MSG_TIMED_OUT,
   YAMI_MSG_REJECT_BY_AGENT.

   Clients calling this function should consider yami_message_setup_timeout()
   function, because without it, this function can block forever (which can
   happen when the remote object crashes in the middle of processing).
*/
PHP_FUNCTION(yami_message_wait)
{
    zval *zmessage = 0;

    Y_MESSAGE_RSRC *y_message_rsrc = 0;
    int rc;


    /* @todo is it better to pass an optional timeout parameter and
    call yamiAgentSetUpTimeOut implicitly instead of having the caller
    to explicitly set it up prior to this function call using
    yami_message_setup_timeout? Can the timeout parameter on a message
    be changed once it has been setup? question to Maciej
    */
    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "r",
        &zmessage
    );

    if ( rc == FAILURE || Z_TYPE_P(zmessage) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        ZEND_FETCH_RESOURCE(y_message_rsrc, Y_MESSAGE_RSRC *, &zmessage, -1, Y_MESSAGE_R_NAME, le_y_message);
        ZEND_VERIFY_RESOURCE(y_message_rsrc);
        /* @todo the checks are not really needed because YAMI has a very good
            internal RTTI type checking system. Is this a good strategy?
        */
        /*
        if ( y_message_rsrc->mtype == Y_OUTBOUND_MESSAGE ) {
        */
            YAMI_G(last_error) = yamiAgentMsgWait(y_message_rsrc->omessage);
        /*
        }
        */
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto bool yami_message_get_response_parameters(resource $message, array $mparams)
   Retrieves the parameters received as a response, if any, to an outbound message.
   for outbound. The the parameters constitute the response data from the remote object.

   Returns true if parameters were retrieved successfully; false otherwise. Note that
   the parameter count may be 0 if  the remote end completed the request with no
   parameters.
*/
PHP_FUNCTION(yami_message_get_response)
{
    zval *zmessage = 0;
    zval *zmparams = 0;

    Y_MESSAGE_RSRC *y_message_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "ra",
        &zmessage,
        &zmparams
    );

    if ( rc != FAILURE ) {
        if ( (Z_TYPE_P(zmessage) != IS_RESOURCE) || (Z_TYPE_P(zmparams) != IS_ARRAY) ) {
            YAMI_G(last_error) = YAMIBADTYPE;
        }
        else {
            ZEND_FETCH_RESOURCE(y_message_rsrc, Y_MESSAGE_RSRC *, &zmessage, -1, Y_MESSAGE_R_NAME, le_y_message);
            ZEND_VERIFY_RESOURCE(y_message_rsrc);

            if ( y_message_rsrc->mtype != Y_OUTBOUND_MESSAGE ) {
                YAMI_G(last_error) = YAMIBADTYPE;
            }
            else {
                HPARAMSET ymparams = 0;
                YAMI_G(last_error) = yamiAgentMsgGetResponse(y_message_rsrc->omessage, &ymparams);
                if ( YAMI_G(last_error) == YAMIOK ) {
                    y_params_to_php_array(ymparams, zmparams TSRMLS_CC);
                    yamiDestroyParamSet(ymparams);
                }
            }
        }
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */


/* {{{ proto long yami_agent_receive(resource $agent, string $oname, resource &$message[, int $timeout = 0])
   Receives incoming message for an object identified by oname.
   The message argument receives a handle to the inbound message, which
   contains the message name and parameters. If timeout is 0, this function
   will return immediately with the code YAMI_NOTAVAILABLE when there
   is no inbound message; otherwise it will wait for timeout seconds.
*/
PHP_FUNCTION(yami_agent_receive)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto long yami_message_consume(resource $message)
   Marks the inbound message as processed. The intent of this is to 'sink'
   the message in the Agent so that no response (neither reply nor reject)
   is possible. This function is mainly needed for 'passive' objects for
   implementing the one-way message semantics (so that the message is sent
   and received but no return information is sent to the originating object).
*/
PHP_FUNCTION(yami_message_consume)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto long yami_message_reject(resource $message)
   Marks the inbound message as rejected. No further processing (for
   example, sending reply) will be possible.
*/
PHP_FUNCTION(yami_message_reject)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto long yami_message_reply(resource $message [, array $mparams = null])
   Sends the reply, together with the parameters to the remote object.
   mparams can be null - in which case the reply is just an acknowledgment
   to the remote sender that the message was received.
*/
PHP_FUNCTION(yami_message_reply)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto long yami_message_forward(resource $message, string $domain, string $oname, string $mname[, array mparams = null])
   Forwards the inbound message to an object identified by $oname whose location
   can be looked up in the agent's white pages using $domain.

   It is possible to change the name and the parameters of the original inbound
   message while forwarding. The parameters can be null as well.

   This function can be useful for load-balancing, routers or adapters.
*/
PHP_FUNCTION(yami_message_forward)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto long yami_message_forward_to(resource $message, string $host, int $port, string $oname, string $mname[, array mparams = null[, $mlevel = 2]])
   Forwards the inbound message to an object identified by $oname whose location
   is given by the combination of $host and $port.

   It is possible to change the name and the parameters of the original inbound
   message while forwarding. The parameters can be null as well.

   This function can be useful for load-balancing, routers or adapters.
*/
PHP_FUNCTION(yami_message_forward_to)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto long yami_message_get_request(resource $message, array $mparams)
   Retrieves the parameters received with a request, if any, in an inbound message.
   The the parameters constitute the request data from the remote object.
*/
PHP_FUNCTION(yami_message_get_request)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto bool yami_message_destroy(resource $message)
   Destroys the inbound or outbound message. True if destroyed. False otherwise.
*/
PHP_FUNCTION(yami_message_destroy)
{
    zval *zmessage = 0;

    Y_MESSAGE_RSRC *y_message_rsrc = 0;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "r",
        &zmessage
    );

    if ( rc == FAILURE || Z_TYPE_P(zmessage) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        ZEND_FETCH_RESOURCE(y_message_rsrc, Y_MESSAGE_RSRC *, &zmessage, -1, Y_MESSAGE_R_NAME, le_y_message);
        ZEND_VERIFY_RESOURCE(y_message_rsrc);

        if ( y_message_rsrc->mtype == Y_OUTBOUND_MESSAGE ) {
            YAMI_G(last_error) = yamiAgentMsgDestroy(y_message_rsrc->omessage);
        }
        else {
            YAMI_G(last_error) = yamiAgentDestroyIncomingMsg(y_message_rsrc->imessage);
        }
        zend_list_delete(Z_RESVAL_P(zmessage));
    }

    RETURN_BOOL(YAMI_G(last_error) == YAMIOK);
}
/* }}} */

/* {{{ proto long yami_message_get_type(resource $message, string &$stype[, int &$itype = null])
   Returns the type of the message (i.e inbound or outbound)
*/
PHP_FUNCTION(yami_message_get_type)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto long yami_message_get_info(resource $message, array $info[, bool $detailed = false])
   Retrieves information about the message such as the message name, the
   recipient name (i.e the destination object name), message specification
   level.

   Currently the $detailed parameter is ignored; current thinking is to
   return the parameters of the message, if available .
*/
PHP_FUNCTION(yami_message_get_info)
{
    long ec;
    int rc;

    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "l",
        &ec
    );

    if ( rc == FAILURE ) {
        if ( ZEND_NUM_ARGS() != 1 ) {
            WRONG_PARAM_COUNT;
        }
        else {
            RETURN_FALSE;
        }
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_strerror((int)ec), 1);
}
/* }}} */

/* {{{ proto long yami_message_get_status(resource $message)
   Returns the status of a message.
*/
PHP_FUNCTION(yami_message_get_status)
{
    zval *zmessage = 0;
    zval *zsstatus = 0;
    int zistatus   = -1;

    Y_MESSAGE_RSRC *y_message_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "r",
        &zmessage
    );

    if ( rc == FAILURE || Z_TYPE_P(zmessage) != IS_RESOURCE ) {
        YAMI_G(last_error) = YAMIBADTYPE;
    }
    else {
        enum msgStatus ymstatus;
        ZEND_FETCH_RESOURCE(y_message_rsrc, Y_MESSAGE_RSRC *, &zmessage, -1, Y_MESSAGE_R_NAME, le_y_message);
        ZEND_VERIFY_RESOURCE(y_message_rsrc);

        if ( y_message_rsrc->mtype == Y_OUTBOUND_MESSAGE ) {
            YAMI_G(last_error) = yamiAgentMsgGetStatus(y_message_rsrc->omessage, &ymstatus);
        }
        else {
            ymstatus = (eRejectByAgent + 1);
        }

        zistatus = ymstatus;
    }

    RETURN_LONG(zistatus);
}
/* }}} */

/* {{{ proto string yami_message_get_status_string(mixed $message_or_message_status_code)
   Returns the string form of the status of a message.
*/
PHP_FUNCTION(yami_message_get_status_string)
{
    zval *zm_or_msc= 0;
    long zistatus  = -1;

    Y_MESSAGE_RSRC *y_message_rsrc = 0;
    int rc;


    rc = zend_parse_parameters (
        ZEND_NUM_ARGS() TSRMLS_CC,
        "z",
        &zm_or_msc
    );

    if ( zm_or_msc && Z_TYPE_P(zm_or_msc) == IS_RESOURCE ) {
        enum msgStatus ymstatus;
        ZEND_FETCH_RESOURCE(y_message_rsrc, Y_MESSAGE_RSRC *, &zm_or_msc, -1, Y_MESSAGE_R_NAME, le_y_message);
        ZEND_VERIFY_RESOURCE(y_message_rsrc);

        if ( ZEND_NUM_ARGS() == 1 ) {
            if ( y_message_rsrc->mtype == Y_OUTBOUND_MESSAGE ) {
                YAMI_G(last_error) = yamiAgentMsgGetStatus(y_message_rsrc->omessage, &ymstatus);
            }
            else {
                ymstatus = (eRejectByAgent + 1);
            }
            zistatus = ymstatus;
        }
    }
    else if ( zm_or_msc && Z_TYPE_P(zm_or_msc) == IS_LONG ) {
        zistatus = Z_LVAL_P(zm_or_msc);
    }
    else {
        YAMI_G(last_error) = YAMIBADTYPE;
    }

    /* See TODO about returning constant strings to avoid copying */
    RETURN_STRING( (char *)y_message_status(zistatus), 1);
}
/* }}} */



/* }}} */
