/*
 * yamiagent.c
 *
 * this is the main file that defines the agent's API
 *
 * Copyright (C) 2001-2006 Maciej Sobczak
 * http://www.msobczak.com/prog/yami/
 *
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose. 
 *
 */

#include "yamic.h"
#include "yamiparams.h"
#include "yamierrors.h"
#include "yamilimits.h"
#include "yamidomains.h"
#include "yamiqueue.h"
#include "yamiXDR.h"
#include "yamioslayer.h"
#include "yamitypes.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>


/* the following is for debugging */
/*#define YAMI_VERBOSE_MODE
*/

#ifdef YAMI_VERBOSE_MODE
#define VERBOSE
#define VERBSTREAM stderr
#define CRASH_AGENT { \
     pa->crashed = 1; \
     fprintf(VERBSTREAM, "CRASH_AGENT at %d\n", __LINE__); \
     fflush(VERBSTREAM); \
}
#else
#define CRASH_AGENT { pa->crashed = 1; }
#endif

/* this is for tracking memory leaks */
#ifdef YAMI_TRACK_MEMORY

#define MALLOC(p, s) { \
     p = malloc(s); \
     fprintf(stderr, "%s(%d) malloc %p\n", __FILE__, __LINE__, p); \
}
#define FREE(a) { \
     fprintf(stderr, "%s(%d) free %p\n", __FILE__, __LINE__, a); \
     free(a); \
}

#else

#define MALLOC(p, s) { p = malloc(s); }
#define FREE(a) free(a);

#endif

/********************************************************/

/*
 * here we have some static checks for this implementation's requirements
 * if the compiler chokes on some of the lines below, then some of the
 * assumptions commented are not met - you shouldn't even try to go on
 */

/* sizeof(int) == 4 */
typedef char _requirement1[sizeof(int) == 4 ? 1 : -1];

/* sizeof(double) == 8 */
typedef char _requirement2[sizeof(double) == 8 ? 1 : -1];

/********************************************************/

/* packet types */
enum packetType
{
     REQUEST = 0,
     RESPONSE = 1,
     REJECT = 2,
     UNKOBJECT = 3,
     OVERFLOW = 4,
     REJECTBYAGENT = 5
};

struct yamiagent;

/* this is hidden under the HMESSAGE handle           */
/* instances of this type are kept in a separate list */
struct msgwaitingforreply
{
     enum eYamiType rtti;
     
     int msgid;
     enum msgStatus status;
     void *data; /* NULL until the reply arrives */
     
     /* this indicates if the swap is needed (because the sender  */
     /* has different byte-ordering than the native machine)      */
     int swap;

     /* this is to ensure the integrity of this structure */
     HMUTEX msg_mutex;
     
     /* this is to wake up object waiting for reply             */
     /* there is one semaphore for each waiting call            */
     /* this semaphore gets increased when the receiving thread */
     /* gets the reply - this allows for blocking waits         */
     HSEMAPHORE msg_sem;
     
     /* this is to allow thread-safe deletion                   */
     /* this counter is set to 2 when the message is created    */
     /* it gets decreased when the user disposes it, when the   */
     /* sender sends the message (and the alarm comes for it)   */
     /* it plays the role of "intrusive reference counter"      */
     int threadson;
     
     /* up-link to the parent structure */
     struct yamiagent *agent;
     
     /* these are used while the structure is owned by the agent  */
     /* when the client gets rid of it, the structure is removed  */
     /* from the agent's collection and these two are meaningless */
     struct msgwaitingforreply *prev;
     struct msgwaitingforreply *next;
};

/* simple RTTI check */
static int _checkMESSAGE(HMESSAGE m)
{
     if (m == NULL)
          return YAMIBADHANDLE;
     if (((struct msgwaitingforreply*)m)->rtti != eSentMessage)
          return YAMIBADHANDLE;
     return YAMIOK;
}

/* destructor for the previous structure           */
/* doesn't crash with partially constructed object */
static void _destroy_msgwaitingforreply(struct msgwaitingforreply *mw)
{
     if (mw->data != NULL)
          FREE(mw->data)

     (void)yamiDestroyMutex(mw->msg_mutex);     
     (void)yamiDestroySemaphore(mw->msg_sem);
}

/* this is hidden under the HINCMSG handle                       */
/* and is supposed to encapsulate the incoming message           */
/* the user code uses it while replying or rejecting to messages */
struct msgincoming
{
     enum eYamiType rtti;
     
     char *msgname;
     char *objname;
     char *paramsetraw;
     void *hint;
     
     /* this indicates if the swap is needed (because the sender  */
     /* has different byte-ordering than the native machine)      */
     int swap;

     int levelno;       /* level of the sender     */
     int msgid;         /* generated by the sender */
     int retipaddress;  /* return address and port */
     int retipport;
     int origipaddress; /* source address and port                           */
     int origipport;    /* they are used when replying via duplex connection */
     int connoptions;   /* connection options from the originating request   */
     int processed;     /* bool variable - if the msg was already processed  */
                        /* note - all operations on this flag are performed  */
                        /* in the context of the same thread, so it does     */
                        /* not need to be separately synchronized            */
     
     /* up-link to the parent structure */
     struct yamiagent *agent;
};

/* simple RTTI check */
static int _checkINCMSG(HINCMSG inc)
{
     if (inc == NULL)
          return YAMIBADHANDLE;
     if (((struct msgincoming*)inc)->rtti != eIncomingMessage)
          return YAMIBADHANDLE;
     return YAMIOK;
}

/* destructor for the previous structure */
static void _destroy_msgincoming(struct msgincoming *inc)
{
     if (inc->msgname != NULL)
          FREE(inc->msgname)
     
     if (inc->objname != NULL)
          FREE(inc->objname)

     if (inc->paramsetraw != NULL)
          FREE(inc->paramsetraw)
}

/* this structure keeps everything that is associated with the object, */
/* for example its own message queue and the list of waiting calls     */
/* agent keeps a list of these, one for each object registered         */
struct object
{
     char *name;
     
     enum objectType type;
     
     void *hint;     /* this is set during registration */
     
     /* callback to the dispatch function */
     DISPFUNCTION    dispf;
     
     HMUTEX local_mutex;

     /* this semaphore is used by the polling object when it wants to wait */
     /* or by one of the dispatcher threads in case of passive object      */
     HSEMAPHORE incqueue_sem;
     
     /* queue for messages arriving from other agents                      */
     HQUEUE incqueue;
     
     /* these are used only for passive_singlethreaded object              */
     /* and when there is at least one dispatcher thread                   */
     /* receiver increases the counter when the request arrives and the    */
     /* flag 'processing' is non-zero, which means that one of the threads */
     /* is processing the message (otherwise, the receiver pushes up       */
     /* semaphores). When that thread returns, it pushes up semaphores and */
     /* decreases the counter below                                        */
     /* for passive_multithreaded and polling objects, these two have no   */
     /* meaning and the receiver unconditionally pushes incqueue_sem and   */
     /* (for passive_multithreaded) dispatch_sem in the agent              */
     /* in case where there is no dispatcher thread,                       */
     /* these two have no meaning, too                                     */
     int messages_waiting;
     int processing;
     
     /* the flag and counter below are only for safe object unregistration */
     /* when the object is marked to be unregistered, tobedeleted flag     */
     /* is set (and from that time, no requests are admitted),             */
     /* and when the last thread leaves the object, it makes the deletion  */
     int threadson;
     int tobedeleted;
     
     struct object *next;
};

/* destructor for the previous structure           */
/* doesn't crash with partially constructed object */
static void _destroy_object(struct object *o)
{
     if (o->name != NULL)
          FREE(o->name)

     (void)yamiDestroyQueue(o->incqueue);
     (void)yamiDestroySemaphore(o->incqueue_sem);
     (void)yamiDestroyMutex(o->local_mutex);
}

/* this is the alarm structure                         */
/* it's used for organizing timeouts for sent messages */
struct alarmobject
{
     int                       timeout;
     struct msgwaitingforreply *msg;
};

/* this structure keeps everything */
struct yamiagent
{
     enum eYamiType rtti;
     
     int ownlevelno;   /* own level number   */
     int ownipaddress; /* network-byte order */
     int ownipport;
     
     HDOMAINS domains; /* address-book       */
     
     /* one send queue for the agent */
     HQUEUE         sendqueue;
     HTHREAD        sender_th;
     HSEMAPHORE     sendqueue_sem;
     HSOCKETPOOL    sendpool;
     HSOCKETDYNPOOL senddplxpool;
     HMUTEX         sendpool_mutex; /* used only with synchro-sends */
     int            sendtries;

     /* receiving socket */
     HSOCKET     listensock;
     int         allowduplex;
     
     /* receiving thread */
     HTHREAD     receiver_th;
     HSOCKETPOOL receivepool;
     HSOCKETPOOL receivedplxpool;
     int         receiveridle;
     
     /* semaphore for dispatchers - it contains the number of         */
     /* incoming messages waiting for dispatching                     */
     /* it's not increased when there is a message for polling object */
     /* when this semaphore is available, there is at least one       */
     /* pending message in one of the objects' incoming queue         */
     HSEMAPHORE dispatch_sem;
     
     /* array of handlers to dispatcher threads */
     HTHREAD *dispatchers;
     int      dispatchers_number;

     /* list of objects' stuff                                        */
     /* this list is managed like a ring - agent turns it             */
     /* with every dispatch, thus giving all the objects equal chance */
     /* any_object is an object that matches all object names and     */
     /* can receive message when no object with the given name exists */
     /* in a ring; it's separate from the ring to facilitate its mgmt */
     struct object *objects_ring;
     struct object *any_object;
     HMUTEX objects_mutex;
     size_t objqmaxlength;  /* length and size of queue belonging     */
     size_t objqmaxsize;    /* to registered object                   */
     int    need_objectqueue_mtx;
     int    need_local_mtx;
     
     /* list of calls posted to this agent and waiting for replies    */
     struct msgwaitingforreply *msgwaiting_list;
     HMUTEX msgwaiting_mutex;
     int    need_msg_mtx;
     
     /* array of alarms set for messages sent                         */
     /* this is necessary: clients should have a chance to wake up    */
     /* when waiting for a response to the message and the server     */
     /* crashed or "forgot" to send a response packet                 */
     struct alarmobject *alarms;
     size_t             alarms_size;
     HMUTEX             alarms_mutex;
     HTHREAD            alarms_th;
     
     /* unique message ID for this agent */
     int lastID;
     
     /* if it's 0 - the machine is little-endian */
     /* it's non-zero otherwise (big-endian)     */
     int nativebyteorder;
     
     /* crash indicator (I would put it in planes and cars, too) */
     int crashed;
};

/* simple RTTI check */
static int _checkYAMIAGENT(HYAMIAGENT agent)
{
     if (agent == NULL)
          return YAMIBADHANDLE;
     if (((struct yamiagent*)agent)->rtti != eYamiAgent)
          return YAMIBADHANDLE;
     return YAMIOK;
}

/*
 * thread functions
 */

/* helper structure for sending packets */
struct sendhelp
{
     int ipaddr;
     int ipport;
     int connoptions;
     size_t size;
     struct msgwaitingforreply *msgwaiting;
     struct yamiagent *agent;
};

/* helper for sending REQUEST messages */
static int _send_request(void *data, struct sendhelp *help, int tries)
{
     int cc1, cc2;

#ifdef VERBOSE
     fprintf(VERBSTREAM, "s: sending request message\n");
     fflush(VERBSTREAM);
#endif
     
     /* it's necessary to synchronize access to "status" field */
     /* of the msgwaiting object, because it's possible that   */
     /* in the meantime, the response will arrive before       */
     /* this thread take its chance to set it properly         */
     /* this applies only to the case where the message being  */
     /* sent is a new message, not a forward                   */
     /* in the latter case, there is no msgwaiting structure   */

     if (help->msgwaiting != NULL)
     {
          cc2 = yamiMutexLock(help->msgwaiting->msg_mutex);
          if (cc2 != YAMIOK)
               return cc2;
     }

     if (help->connoptions == eFixedDuplex)
     {
          /* this is a duplex connection, use dedicated pool */
          cc1 = yamiNetDuplexPoolSend(help->agent->senddplxpool, help->ipaddr,
               help->ipport, (const char*)data, help->size);
     }
     else
     {
          /* this is a normal (non-duplex) connection */
          cc1 = yamiNetPoolSend(help->agent->sendpool, help->ipaddr,
               help->ipport, (const char*)data, help->size, tries);
     }

     cc2 = YAMIOK;
     
     if (help->msgwaiting != NULL)
     {
          if (cc1 == YAMIOK)
          {
               /* set the success status */
               help->msgwaiting->status = ePending;
          }
          else
          {
               /* set the status to error */
               help->msgwaiting->status = eNetError;
          }
     
          /* and push up the semaphore (in case of error)    */
          /* so that threads waiting on this token will have */
          /* a chance to wake up and examine the problem     */
          if (cc1 != YAMIOK)
               (void)yamiSemaphoreRelease(help->msgwaiting->msg_sem);

          if (--(help->msgwaiting->threadson) == 0)
          {
               /* this means that no thread is likely to */
               /* process this message - destroy         */
               cc2 = yamiMutexUnlock(help->msgwaiting->msg_mutex);
               _destroy_msgwaitingforreply(help->msgwaiting);
               FREE(help->msgwaiting)
          }
          else
          {
               /* anyway, unlock the mutex */
               cc2 = yamiMutexUnlock(help->msgwaiting->msg_mutex);
          }
     }
     
     return (cc1 != YAMIOK) ? cc1 : cc2;
}

/* helper for sending RESPONSE, UNKOBJECT, OVERFLOW or REJECT messages */
/* theses are all "returning" messages - if there is a duplex          */
/* connection with the given address and port, it is preferred         */
static int _send_generic(void *data, struct sendhelp *help, int tries)
{
     int cc;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "s: sending RESPONSE, UNKOBJECT, OVERFLOW, "
          "REJECT or REJECTBYAGENT message\n");
     fflush(VERBSTREAM);
#endif

     if (help->connoptions == eFixedDuplex)
     {
          /* send back through the same dulpex connections */
          cc = yamiNetPoolSendBack(help->agent->receivedplxpool, help->ipaddr,
               help->ipport, (const char*)data, help->size);
     }
     else
     {
          /* this was a simplex request, send back using normal pool */
          cc = yamiNetPoolSend(help->agent->sendpool, help->ipaddr,
               help->ipport, (const char*)data, help->size, tries);
     }
     
     return cc;
}

/* helper for queing messages in the sender queue  */
/* (if there is a sender thread) or for performing */
/* synchro send (if there is no sender thread)     */
/* in case of success, it takes care of the memory */
/* allocated for *data and *help                   */
static int _queue_or_send(struct yamiagent *pa, void *data,
          struct sendhelp *help, size_t size, enum packetType type)
{
     int cc, cc2;
     
     if (pa->sender_th)
     {
          /* there is a sender thread                          */
          /* do asynchronous send with the use of sender queue */
          
          /* the sender thread will release the memory         */
          /* for *data and *help                               */

#ifdef VERBOSE
          fprintf(VERBSTREAM, "s: queueing message for sending\n");
          fflush(VERBSTREAM);
#endif
          cc = yamiQueuePushBack(pa->sendqueue, data, help, size);
          if (cc != YAMIOK)
               return cc;
          return yamiSemaphoreRelease(pa->sendqueue_sem);
     }
     else
     {
          /* there is no sender thread */
          /* do synchronous send       */
#ifdef VERBOSE
          fprintf(VERBSTREAM, "s: sending message synchronously\n");
          fflush(VERBSTREAM);
#endif
          
          if ((cc = yamiMutexLock(pa->sendpool_mutex)) != YAMIOK)
               return cc;

          if (type == REQUEST)
          {
               cc = _send_request(data, help, pa->sendtries);
          }
          else
          {
               cc = _send_generic(data, help, pa->sendtries);
          }

          cc2 = yamiMutexUnlock(pa->sendpool_mutex);

          /* release the memory if OK */

          if (cc == YAMIOK && cc2 == YAMIOK)
          {
               FREE(data)
               FREE(help)
               
               return YAMIOK;
          }          

          return (cc != YAMIOK) ? cc : cc2;
     }
}

/* sender thread function */
static void sender_function(HYAMIAGENT agent,
     HSEMAPHORE start_sem, HFLAG stopflag)
{
#ifndef YAMI_NO_THREADS

     struct yamiagent* pa;
     void *praw1, *praw2;
     const char *p;
     size_t size;
     int msgtype;
     enum packetType etype;
     int stopvalue;
     int cc;
     
     pa = (struct yamiagent*)agent;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "s: sender thread started\n");
     fflush(VERBSTREAM);
#endif

     /* signal the parent that is ready to work */
     cc = yamiSemaphoreRelease(start_sem);
     if (cc != YAMIOK)
     {
          CRASH_AGENT
          return;
     }
     
     while (1 == 1)
     {
          /* wait for something to arrive to the queue */
          cc = yamiSemaphoreAcquire(pa->sendqueue_sem);
          if (cc != YAMIOK)
          {
               /* this is a serious error,        */
               /* which in fact crashes the agent */
               CRASH_AGENT
               break;
          }
          
          cc = yamiFlagGet(stopflag, &stopvalue);
          if (cc != YAMIOK)
          {
               /* again, crash */
               CRASH_AGENT
               break;
          }

          if (stopvalue != 0)
               break;
          
          /* get the next message data from the queue */
          cc = yamiQueuePopFront(pa->sendqueue, &praw1, &praw2, &size);
          if (cc != YAMIOK)
          {
               /* again, crash */
               CRASH_AGENT
               break;
          }
          
          /* then, extract the type of this message */
          
          p = (char*)praw1;
          
          p += 12;
          (void)yamiXDRReadInt(p, &msgtype, 0); /* no swap here */
          etype = (enum packetType) (msgtype & ((1 << 16) - 1));

          switch (etype)
          {
          case REQUEST:
               (void)_send_request(praw1, (struct sendhelp*)praw2,
                         pa->sendtries);
               break;
          case RESPONSE:
          case UNKOBJECT:
          case OVERFLOW:
          case REJECT:
          case REJECTBYAGENT:
               (void)_send_generic(praw1, (struct sendhelp*)praw2,
                         pa->sendtries);
               break;
          }          
          
          /* release the memory */
          FREE(praw1)
          FREE(praw2)
     }

#ifdef VERBOSE
     fprintf(VERBSTREAM, "s: sender thread quitting\n");
     fflush(VERBSTREAM);
#endif

#endif /* YAMI_NO_THREADS */
}

/* helper for reading strings from a socket            */
/* these strings are used for message and object names */
static int _receive_string(HSOCKET s, char **pbuf, size_t *size, int swap)
{
     char tmpbuf[sizeof(int)];
     const char *pc;
     size_t align, readn;
     int tmp;
     int cc;
     
     cc = yamiNetReadN(s, tmpbuf, sizeof(int), &readn);
     if (cc != YAMIOK || readn != sizeof(int))
          return YAMINETERROR;
     
     pc = tmpbuf;
     (void)yamiXDRReadInt(pc, &tmp, swap);
     
     if (tmp <= 0 || tmp > MAXNAMELEN)
          return YAMIPARSEERROR;
     
     *size = (size_t)tmp;
     MALLOC(*pbuf, *size + 1)

     if (*pbuf == NULL)
          return YAMINOMEMORY;

     cc = yamiNetReadN(s, *pbuf, *size, &readn);

     if (cc != YAMIOK || readn != *size)
     {
          FREE(*pbuf)
          return YAMINETERROR;
     }
     (*pbuf)[*size] = (char)0;
     
     /* because we use XDR as a career protocol, */
     /* we need to align to 4-bytes boundary     */
     if (*size % 4 != 0)
     {
          align = 4 - *size % 4;
          cc = yamiNetReadN(s, tmpbuf, align, &readn);
          if (cc != YAMIOK || readn != align)
               return YAMINETERROR;
     }

     return YAMIOK;
}

/* helper for setting the status of the message token */
static int _set_status(struct yamiagent *pa, int msgid, enum msgStatus status)
{
     struct msgwaitingforreply *msg;
     int quitloop;
     int cc;
     
     /* find the message token */
     
     cc = yamiMutexLock(pa->msgwaiting_mutex);
     if (cc != YAMIOK)
          return cc;
     
     msg = pa->msgwaiting_list;
     quitloop = 0;
     while (msg)
     {
          if (msg->msgid == msgid)
          {
               cc = yamiMutexLock(msg->msg_mutex);
               if (cc != YAMIOK)
               {
                    (void)yamiMutexUnlock(pa->msgwaiting_mutex);
                    return cc;
               }
               
               if (msg->status == ePending)
               {
                    msg->status = status;

                    cc = yamiSemaphoreRelease(msg->msg_sem);
                    if (cc != YAMIOK)
                    {
                         (void)yamiMutexUnlock(msg->msg_mutex);
                         (void)yamiMutexUnlock(pa->msgwaiting_mutex);
                         return cc;
                    }

                    quitloop = 1;
               }
               
               cc = yamiMutexUnlock(msg->msg_mutex);
               if (cc != YAMIOK)
               {
                    (void)yamiMutexUnlock(pa->msgwaiting_mutex);
                    return cc;
               }
               
               if (quitloop == 1)
                    break;
          }
          msg = msg->next;
     }
     
     /* if the token was not found, we don't care */
     /* - either the user has removed it himself  */
     /* or this is a message from another planet  */
     
     cc = yamiMutexUnlock(pa->msgwaiting_mutex);
     return cc;
}

/* helper for responding with UNKNOWN, OVERFLOW */
/* or REJECT for shortcut routing               */
static void _respond_with_uor_shortcut(struct yamiagent *pa, int msgid,
          enum packetType type)
{
     enum msgStatus status;
     
     switch (type)
     {
     case UNKOBJECT:
          status = eUnknownObj;
          break;
     case OVERFLOW:
          status = eOverflow;
          break;
     case REJECT:
          status = eRejected;
          break;
     /* no need for case REJECTBYAGENT                           */
     /* because there is no possibility to send the L2 message   */
     /* to the local agent, which is L1 - this is detected       */
     /* while the message is being sent (and logic error raised) */
     default:
          assert(0);
          status = 0; /* to shut up compiler warning */
     }
     
     (void)_set_status(pa, msgid, status);
}

/* helper for responding with error messages: */
/* UNKOBJECT, OVERFLOW, REJECT, REJECTBYAGENT */
static void _respond_with_unkobj_of_rej(struct yamiagent *pa,
          int msgid, enum packetType type, int ipaddr, int ipport)
{
     char *buf, *p;
     struct sendhelp *help;
     int cc;

     /* check the oportunity for shortcut routing */
     if (ipaddr == pa->ownipaddress && ipport == pa->ownipport)
     {
          _respond_with_uor_shortcut(pa, msgid, type);
          return;
     }
          
     /* prepare the UNKOBJECT, OVERFLOW or REJECT message */
          
     MALLOC(buf, 4 * sizeof(int))

     if (buf == NULL)
          return;

     MALLOC(help, sizeof(struct sendhelp))

     if (help == NULL)
     {
          FREE(buf)
          return;
     }

     p = buf;
     p = yamiXDRWriteInt(p, pa->nativebyteorder);
     p = yamiXDRWriteInt(p, pa->ownlevelno);
     p = yamiXDRWriteInt(p, msgid);
     (void)yamiXDRWriteInt(p, (int)type);
     
     help->ipaddr = ipaddr;
     help->ipport = ipport;
     help->size = 4 * sizeof(int);
     help->msgwaiting = NULL;
     help->agent = pa;

     /* here, we cheat telling that the message has 0 bytes, */
     /* but this msg is internal to the agent and MUST pass  */

     cc = _queue_or_send(pa, buf, help, 0, type);
     if (cc != YAMIOK)
     {
          FREE(buf)
          FREE(help)

          /* if the error is of network type, do nothing */
          /* it can be a fault of the remote part        */
          
          if (cc != YAMINETERROR)
          {
               /* crash */
               CRASH_AGENT
          }
     }
}

/* helper for finding object in a ring                 */
/* the call should be already serialized               */
/* if strict is nonzero, it will not return any-object */
/* (when registered) when no other object matches      */
static struct object * _find_object(struct yamiagent *pa,
          const char *objname, int strict)
{
     struct object *po1, *po2;
     int cc;

     cc = 0;
     
     if (objname == YAMI_ANY_OBJECT)
     {
          return pa->any_object;
     }
     
     if (pa->objects_ring != NULL)
     {
          po2 = NULL;
          po1 = pa->objects_ring;
          while (po1 != po2)
          {
               po2 = pa->objects_ring;
               
               if (strcmp(po1->name, objname) == 0)
               {
                    cc = 1;
                    break;
               }
               
               po1 = po1->next;
          }
     }
     
     if (cc == 1)
     {
          return po1;
     }
     else
     {
          return (strict != 0) ? NULL : pa->any_object;
     }
}

/* helper for queueing incoming messages in an object's queue */
static
int _queue_incoming(struct yamiagent *pa, struct object *po, char *pobjname,
          char *pmsgname, char *pparset, size_t parsetsize, int swap,
          int levelno, int msgid, int retipaddr, int retipport,
          int origipaddress, int origipport, int connoptions,
          int *queued, int *objsmtxunlocked)
{
     struct msgincoming *pincoming;
     int cc;
     
     *queued = 0;
     *objsmtxunlocked = 0;

     MALLOC(pincoming, sizeof(struct msgincoming))

     if (pincoming == NULL)
          return YAMINOMEMORY;

     /* prepare the msgincoming structure */

     pincoming->rtti = eIncomingMessage;
     
     pincoming->msgname = pmsgname;
     pincoming->objname = pobjname;
     pincoming->paramsetraw = pparset;
     pincoming->hint = po->hint;
     pincoming->swap = swap;
     pincoming->levelno = levelno;
     pincoming->msgid = msgid;
     pincoming->retipaddress = retipaddr;
     pincoming->retipport = retipport;
     pincoming->origipaddress = origipaddress;
     pincoming->origipport = origipport;
     pincoming->connoptions = connoptions;
     pincoming->processed = 0;
     pincoming->agent = pa;

     cc = yamiMutexLock(po->local_mutex);
     if (cc != YAMIOK)
     {
          FREE(pincoming)
          return cc;
     }
     
     if (po->type == polling)
     {
          cc = yamiQueuePushBack(po->incqueue, pincoming, NULL, parsetsize);
          if (cc != YAMIOK)     
          {
               FREE(pincoming)
               (void)yamiMutexUnlock(po->local_mutex);
               return cc;
          }
     
          *queued = 1;

          /* release semaphore so that any waiting thread */
          /* can retrieve the new message                 */
          cc = yamiSemaphoreRelease(po->incqueue_sem);
          if (cc != YAMIOK)
          {
               (void)yamiMutexUnlock(po->local_mutex);
               return cc;
          }
          
          return yamiMutexUnlock(po->local_mutex);
     }

     /* here is the case of passive object */
     
     if (pa->dispatchers != 0)
     {
          /* if there is at least one dispatcher thread, then */
          /* there is a sense in pushing the incoming message */
          /* into the object's private queue                  */
          
#ifdef VERBOSE
          fprintf(VERBSTREAM, "a: storing incoming message in the queue\n");
          fflush(VERBSTREAM);
#endif

          cc = yamiQueuePushBack(po->incqueue, pincoming, NULL, parsetsize);
          if (cc != YAMIOK)     
          {
               FREE(pincoming)
               (void)yamiMutexUnlock(po->local_mutex);
               return cc;
          }
     
          *queued = 1;
          
          if (po->type == passive_multithreaded)
          {
               /* release semaphore in the object and */
               /* the global semaphore in the agent   */
               cc = yamiSemaphoreRelease(po->incqueue_sem);
               if (cc != YAMIOK)
               {
                    (void)yamiMutexUnlock(po->local_mutex);
                    return cc;
               }
               cc = yamiSemaphoreRelease(pa->dispatch_sem);
               if (cc != YAMIOK)
               {
                    (void)yamiMutexUnlock(po->local_mutex);
                    return cc;
               }
          }
          else /* po->type == passive_singlethreaded */
          {
               /* if the flag 'processing' is set          */
               /* it means, that one of the dispatchers    */
               /* is still working on the previous message */
               /* so just adjust the message counter       */
               /* otherwise, push up the semaphore         */
               if (po->processing != 0)
               {
                    ++(po->messages_waiting);
               }
               else
               {
                    cc = yamiSemaphoreRelease(po->incqueue_sem);
                    if (cc != YAMIOK)
                    {
                         (void)yamiMutexUnlock(po->local_mutex);
                         return cc;
                    }
                    cc = yamiSemaphoreRelease(pa->dispatch_sem);
                    if (cc != YAMIOK)
                    {
                         (void)yamiMutexUnlock(po->local_mutex);
                         return cc;
                    }
               }
          }
          
          return yamiMutexUnlock(po->local_mutex);
     }
     else
     {
          /* there is no dispatcher thread and the object is passive */
          /* - the current thread has to dispatch the message itself */

          /* increase the counter of threads working on this object */
          ++(po->threadson);
          
          cc = yamiMutexUnlock(po->local_mutex);
          if (cc != YAMIOK)
          {
               FREE(pincoming)
               return cc;
          }
          
          cc = yamiMutexUnlock(pa->objects_mutex);
          if (cc != YAMIOK)
          {
               FREE(pincoming)
               return cc;
          }
          
          *objsmtxunlocked = 1;
          
          /* call the dispatch function */
          
#ifdef VERBOSE
          fprintf(VERBSTREAM, "a: calling the dispatch function directly\n");
          fflush(VERBSTREAM);
#endif

          (po->dispf)(pincoming);

          /* the message should be considered rejected   */
          /* but only, if it wasn't marked as processed  */
          /* which would mean that the dispatch function */
          /* did reply or rejected itself                */
               
          if (pincoming->processed == 0)
          {

#ifdef VERBOSE
               fprintf(VERBSTREAM, "a: direct dispatched unsuccessful - "
                    "rejecting this message\n");
               fflush(VERBSTREAM);
#endif

               _respond_with_unkobj_of_rej(pa, msgid, REJECT,
                    retipaddr, retipport);
          }

          /* whatever the outcome of the dispatch, */
          /* release the msgincoming structure     */
          /* (do *not* release the strings!)       */

          FREE(pincoming)

          cc = yamiMutexLock(po->local_mutex);
          if (cc != YAMIOK)
          {
               return cc;
          }

          if (--(po->threadson) == 0 && po->tobedeleted == 1)
          {
               /* this means, that the object structure is no longer */
               /* in the objects' collection and no thread is        */
               /* working on it (and no one will) - we should        */
               /* destroy the object structure right now             */

               cc = yamiMutexUnlock(po->local_mutex);
               if (cc != YAMIOK)
               {
                    return cc;
               }
               
               _destroy_object(po);
               FREE(po)
               
               return YAMIOK;
          }
          else
          {
               return yamiMutexUnlock(po->local_mutex);
          }
     }
}

/* helper for receiving requests */
static int _receive_request(HSOCKET s, struct yamiagent *pa,
               int levelno, int msgid, int swap,
               int origipaddress, int origipport, int connoptions)
{
     char buf[2 * sizeof(int)];
     const char *pc;
     int retipaddr, retipport;
     size_t objnamelen, msgnamelen;
     int parsetsize;
     char *pobjname, *pmsgname, *pparset;
     int queued, objsmtxunlocked;
     size_t readn;
     struct object *po;
     int cc;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiving request\n");
     fflush(VERBSTREAM);
#endif
     
     /* read the return IP address */

     cc = yamiNetReadN(s, buf, 2 * sizeof(int), &readn);
     if (cc != YAMIOK || readn != 2 * sizeof(int))
          return YAMINETERROR;
     
     pc = buf;
     retipaddr = *(int*)pc;     /* ipaddr is in net-order */
     pc += 4;
     pc = yamiXDRReadInt(pc, &retipport, swap);
     
     /* read the object name */

     cc = _receive_string(s, &pobjname, &objnamelen, swap);
     if (cc != YAMIOK)
          return cc;
     
     /* read the message name */

     cc = _receive_string(s, &pmsgname, &msgnamelen, swap);
     if (cc != YAMIOK)
     {
          FREE(pobjname)
          return cc;
     }
     
     /* read the size of the parameter set */
     
     cc = yamiNetReadN(s, buf, sizeof(int), &readn);
     if (cc != YAMIOK || readn != sizeof(int))
     {
          FREE(pobjname)
          FREE(pmsgname)
          return YAMINETERROR;
     }
     pc = buf;
     (void)yamiXDRReadInt(pc, &parsetsize, swap);
     
     if (parsetsize < 0 || parsetsize > MAXPARAMSETSIZE)
     {
          /* we consider this to be "denial of service attack" */
          /* this packet is invalid                            */
          return YAMIBADLENGTH;
     }
     
     /* read the parameter set */
     
     /* the handshake is performed immediately after reading */
     /* all of the packet - this is to prevent deadlock in   */
     /* agents that send and receive synchronously           */
     
     if (parsetsize != 0)
     {
          MALLOC(pparset, (size_t)parsetsize)
          if (pparset == NULL)
          {
               FREE(pobjname)
               FREE(pmsgname)
               return YAMINOMEMORY;
          }
          cc = yamiNetReadN(s, pparset, (size_t)parsetsize, &readn);
          if (cc != YAMIOK || readn != (size_t)parsetsize)
          {
               FREE(pobjname)
               FREE(pmsgname)
               FREE(pparset)
               return YAMINETERROR;
          }
          
          /* the handshake is performed only for simplex connection */
          if (connoptions == eNone)
               (void)yamiNetShakeHand(s);

          /* check if the parameter set complies to the local level */
          cc = yamiCheckRawBuffer(pa->ownlevelno, pparset, swap);
          if (cc != YAMIOK)
          {
               /* it's a message from the level higher than */
               /* the level of the agent: reject it         */

               FREE(pobjname)
               FREE(pmsgname)
               FREE(pparset)

#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: reject by agent, msg: %d\n", msgid);
     fflush(VERBSTREAM);
#endif

               _respond_with_unkobj_of_rej(pa, msgid, REJECTBYAGENT,
                    retipaddr, retipport);

               return YAMIOK;
          }
     }
     else
     {
          /* the handshake is performed only for simplex connection */
          if (connoptions == eNone)
               (void)yamiNetShakeHand(s);

          pparset = NULL;
     }

#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: received message for objname: %s, msg: %s, "
          "parsize: %d\n", pobjname, pmsgname, parsetsize);
     fflush(VERBSTREAM);
#endif

     /* find the object in a collection */
     
     cc = yamiMutexLock(pa->objects_mutex);
     if (cc != YAMIOK)
     {
          /* crash */
          FREE(pobjname)
          FREE(pmsgname)

          if (pparset)
               FREE(pparset)

          CRASH_AGENT
          return YAMISYNCHROERROR;
     }

     po = _find_object(pa, pobjname, 0);

     if (po != NULL)
     {
          /* object with a given name was found (po) */
          
#ifdef VERBOSE
          fprintf(VERBSTREAM, "r: found object with a given name - "
               "queueing message\n");
          fflush(VERBSTREAM);
#endif
          
          cc = _queue_incoming(pa, po, pobjname, pmsgname, pparset,
               (size_t)parsetsize, swap, levelno, msgid,
               retipaddr, retipport, origipaddress, origipport, connoptions,
               &queued, &objsmtxunlocked);

          if (queued == 0)
          {
               FREE(pobjname)
               FREE(pmsgname)
               FREE(pparset)
          }

          if (cc != YAMIOK)
          {
               if (cc == YAMIOVERFLOW)
               {
                    /* overflow in the object's queue */
                    /* send the overflow notification */
#ifdef VERBOSE
          fprintf(VERBSTREAM, "r: overflow while queueing message\n");
          fflush(VERBSTREAM);
#endif

                    _respond_with_unkobj_of_rej(pa, msgid,
                         OVERFLOW, retipaddr, retipport);
               }
               else
               {
                    /* crash */
                    CRASH_AGENT
               }
          }

          if (objsmtxunlocked == 0)
          {
               cc = yamiMutexUnlock(pa->objects_mutex);
               if (cc != YAMIOK)
               {
                    /* crash */
                    CRASH_AGENT
               }
          }
          
          if (pa->crashed != 0)
               return YAMIINTERNAL;

          return YAMIOK;
     }
     else
     {
          /* not found */
          /* in this case, we post UNKOBJECT message to the sender */
          
          cc = yamiMutexUnlock(pa->objects_mutex);
          if (cc != YAMIOK)
          {
               /* crash */
               CRASH_AGENT
          }

#ifdef VERBOSE
          fprintf(VERBSTREAM, "r: object with this name not found - "
               "pinging back\n");
          fflush(VERBSTREAM);
#endif

          FREE(pmsgname)
          FREE(pobjname)
          FREE(pparset)

          _respond_with_unkobj_of_rej(pa, msgid, UNKOBJECT,
               retipaddr, retipport);
     }
     
     if (pa->crashed != 0)
          return YAMIINTERNAL;
     
     return YAMIOK;
}

/* helper for receiving responses */
static int _receive_response(HSOCKET s, struct yamiagent *pa,
               int msgid, int swap)
{
     char buf[sizeof(int)];
     const char *pc;
     int parsetsize;
     char *pparset;
     size_t readn;
     struct msgwaitingforreply *waiting;
     int replied, cc;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiving response\n");
     fflush(VERBSTREAM);
#endif
     
     /* read the size of the parameter set */

     cc = yamiNetReadN(s, buf, sizeof(int), &readn);
     if (cc != YAMIOK || readn != sizeof(int))
     {
          return YAMINETERROR;
     }
     pc = buf;
     (void)yamiXDRReadInt(pc, &parsetsize, swap);
     
     if (parsetsize < 0 || parsetsize > MAXPARAMSETSIZE)
          return YAMIBADLENGTH;
     
     /* read the parameter set */
     
     if (parsetsize != 0)
     {
          MALLOC(pparset, (size_t)parsetsize)

          if (pparset == NULL)
          {
               return YAMINOMEMORY;
          }

          cc = yamiNetReadN(s, pparset, (size_t)parsetsize, &readn);

          if (cc != YAMIOK || readn != (size_t)parsetsize)
          {
               FREE(pparset)
               return YAMINETERROR;
          }

          /* check if the parameter set complies to the local level */
          cc = yamiCheckRawBuffer(pa->ownlevelno, pparset, swap);
          if (cc != YAMIOK)
          {
               /* this means that some non-compliant agent allowed */
               /* to reply with inappropriate level - ignore it    */
               FREE(pparset)
               return cc;
          }
     }
     else
     {
          pparset = NULL;
     }

#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: received response for msgid: %d, parsize: %d\n",
          msgid, parsetsize);
     fflush(VERBSTREAM);
#endif
     
     /* find the message token */

     cc = yamiMutexLock(pa->msgwaiting_mutex);
     if (cc != YAMIOK)
     {
          /* crash */
          if (pparset)
               FREE(pparset)

          CRASH_AGENT
          return YAMISYNCHROERROR;
     }
     
     waiting = pa->msgwaiting_list;
     replied = 0;
     while (waiting)
     {
          if (waiting->msgid == msgid)
          {
               /* message token was found */

               cc = yamiMutexLock(waiting->msg_mutex);
               if (cc != YAMIOK)
               {
                    /* crash */
                    CRASH_AGENT
                    break;
               }
                                        
               /* the response message can be valid only        */
               /* if the status of the message token is pending */
               /* otherwise, we just ignore it - it may be a junk */
               if (waiting->status == ePending)
               {
#ifdef VERBOSE
                    fprintf(VERBSTREAM, "r: writing data to the waiting "
                         "token for msgid = %d\n", msgid);
                    fflush(VERBSTREAM);
#endif

                    waiting->status = eReplied;
                    waiting->data = pparset;
                    waiting->swap = swap;
                    replied = 1;
                    
                    /* push up the semaphore for this token */
                    cc = yamiSemaphoreRelease(waiting->msg_sem);
                    if (cc != YAMIOK)
                    {
                         /* crash */
                         CRASH_AGENT
                    }
               }

               cc = yamiMutexUnlock(waiting->msg_mutex);
               if (cc != YAMIOK)
               {
                    /* crash */
                    CRASH_AGENT
               }
                    
               break;
          }
          
          waiting = waiting->next;
     }
     
     if (replied == 0)
     {
          /* there was a soft error (junk) */
          FREE(pparset)
     }

     cc = yamiMutexUnlock(pa->msgwaiting_mutex);
     if (cc != YAMIOK)
          CRASH_AGENT
     
     if (pa->crashed != 0)
          return YAMIINTERNAL;
     
     return YAMIOK;
}

/* helper for receiving rejections */
static int _receive_reject(struct yamiagent *pa, int msgid)
{
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiving reject for msgid: %d\n", msgid);
     fflush(VERBSTREAM);
#endif

     return _set_status(pa, msgid, eRejected);
}

/* helper for receiving rejections by agent */
static int _receive_reject_by_agent(struct yamiagent *pa, int msgid)
{
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiving reject by agent for msgid: %d\n",
          msgid);
     fflush(VERBSTREAM);
#endif

     return _set_status(pa, msgid, eRejectByAgent);
}

/* helper for receiving UNKOBJECT notifications */
static int _receive_unkobject(struct yamiagent *pa, int msgid)
{
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiving unkobject for msgid: %d\n", msgid);
     fflush(VERBSTREAM);
#endif

     return _set_status(pa, msgid, eUnknownObj);
}

/* helper for receiving OVERFLOW notifications */
static int _receive_overflow(struct yamiagent *pa, int msgid)
{
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiving overflow for msgid: %d\n", msgid);
     fflush(VERBSTREAM);
#endif

     return _set_status(pa, msgid, eOverflow);
}

/* helper for _process_connection */
/* for closing misbehaving socket */
static void _close_connection(struct yamiagent *pa, HSOCKET accepteds,
     int whatconn)
{
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: closing connection\n");
     fflush(VERBSTREAM);
#endif

     switch (whatconn)
     {
     case 0:
          /* socket accepted from the listening socket */
          (void)yamiNetCloseSocket(accepteds);
          break;
     case 1:
          /* socket kept in the simplex pool */
          (void)yamiNetPoolCloseOne(pa->receivepool, accepteds);
          (void)yamiNetDisposeButDontCloseSocket(accepteds);
          break;
     case 2:
          /* socket kept in the duplex pool */
          (void)yamiNetPoolCloseOne(pa->receivedplxpool, accepteds);
          (void)yamiNetDisposeButDontCloseSocket(accepteds);
          break;
     case 3:
          /* socket kept in the sending duplex pool */
          (void)yamiNetDynPoolClose(pa->senddplxpool, accepteds);
          (void)yamiNetDisposeButDontCloseSocket(accepteds);
          break;
     }
}

/* process the incoming connection */
static int _process_connection(struct yamiagent *pa, HSOCKET accepteds,
     int whatconn, int origipaddr, int origipport)
{
     char tmpbuf[4 * sizeof(int)];
     const char *pc;
     size_t readn;
     int levelno, msgid, msgtype;
     int endianness, swap;
     int connoptions;
     int cc;

     /* we just close the accepted socket              */
     /* if anything goes wrong (eg. ill msg format)    */
          
     /* read the beginning of the message          */
     /* its form is the same for all message types */

#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: reading from the socket, whatconn = %d\n", whatconn);
     fflush(VERBSTREAM);
#endif

     cc = yamiNetReadN(accepteds, tmpbuf, 4 * sizeof(int), &readn);
     if (cc != YAMIOK)
     {
          _close_connection(pa, accepteds, whatconn);
          return cc;
     }
     
     if (readn != 4 * sizeof(int))
     {
          /* soft error */
          _close_connection(pa, accepteds, whatconn);
          return YAMIOK;
     }

     pc = tmpbuf;
     pc = yamiXDRReadInt(pc, &endianness, 0);
          
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiving packet from a %s-endian machine\n",
          (endianness == 0) ? "little" : "big");
     fflush(VERBSTREAM);
#endif

     /* here we decide if the byte ordering of the sender    */
     /* is the same or different of the native byte ordering */
     /* of this agent and so if to do the swap or no         */
     if ((endianness == 0 && pa->nativebyteorder == 0) ||
          (endianness != 0 && pa->nativebyteorder != 0))
          swap = 0;
     else
          swap = 1;
     
     pc = yamiXDRReadInt(pc, &levelno, swap);
     pc = yamiXDRReadInt(pc, &msgid, swap);
     (void)yamiXDRReadInt(pc, &msgtype, swap);
     
     if (levelno != 1 && levelno != 2)
     {
          /* the packet level is invalid */
          /* this is soft error - it is not a fault of a receiver */
          _close_connection(pa, accepteds, whatconn);
          return YAMIOK;
     }

     /* extract the connection type (normal or duplex) */
     connoptions = msgtype >> 16;
     msgtype &= ((1 << 16) - 1);

     if (whatconn == 0)
     {
          if (connoptions == eNone)
          {
               /* this is a normal simplex connection */
               /* add it to the simplex pool          */
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: this is a new simplex connection, "
          "addr = %d, port = %d\n", origipaddr, origipport);
     fflush(VERBSTREAM);
#endif
               cc = yamiNetPoolAdd(pa->receivepool, accepteds, 1,
                    origipaddr, origipport);
               if (cc != YAMIOK)
               {
                    (void)yamiNetCloseSocket(accepteds);
                    return cc;
               }
          }
          else
          {
               /* this is a duplex connection */
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: this is a new duplex connection, "
          "addr = %d, port = %d\n", origipaddr, origipport);
     fflush(VERBSTREAM);
#endif
               if (pa->allowduplex == 0)
               {
                    /* incoming duplex connections not allowed */
                    (void)yamiNetCloseSocket(accepteds);
                    return YAMIOK;
               }

               /* add the new conncetion to the duplex pool   */
               cc = yamiNetPoolAdd(pa->receivedplxpool, accepteds, 0,
                    origipaddr, origipport);
               if (cc == YAMIOVERFLOW)
               {
                    /* there is more incoming duplex connections */
                    /* than allowed in the maxdplxconns policy   */
                    /* just discard this connection and continue */
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: too many duplex connections, closing\n");
     fflush(VERBSTREAM);
#endif

                    (void)yamiNetCloseSocket(accepteds);
                    return YAMIOK;
               }
               else if (cc != YAMIOK)
                    return cc;
          }
     }
     
     switch (msgtype)
     {
     case (int)REQUEST:
          cc = _receive_request(accepteds, pa, levelno,
               msgid, swap, origipaddr, origipport, connoptions);
          break;
     case (int)RESPONSE:
          cc = _receive_response(accepteds, pa, msgid, swap);
          break;
     case (int)REJECT:
          cc = _receive_reject(pa, msgid);
          break;
     case (int)REJECTBYAGENT:
          cc = _receive_reject_by_agent(pa, msgid);
          break;
     case (int)UNKOBJECT:
          cc = _receive_unkobject(pa, msgid);
          break;
     case (int)OVERFLOW:
          cc = _receive_overflow(pa, msgid);
          break;
     default:
          /* wrong packet type - just ignore it */
          /* no handshake for this message      */
          cc = YAMINETERROR;
#ifdef VERBOSE
          fprintf(VERBSTREAM, "r: received a packet of wrong type\n");
          fflush(VERBSTREAM);
#endif
     }

     if (cc != YAMIOK)
     {
#ifdef VERBOSE
          fprintf(VERBSTREAM, "r: there was some error: 0x08%x\n", cc);
          fflush(VERBSTREAM);
#endif
          _close_connection(pa, accepteds, whatconn);
     }
     else
     {
          /* in the case of REQUEST, the handshake is done     */
          /* in the _receive_request function                  */
          /* anyway, handshake is only for simplex connections */
          if (msgtype != (int)REQUEST && connoptions == eNone &&
               whatconn != 3)
               (void)yamiNetShakeHand(accepteds);
     
          (void)yamiNetDisposeButDontCloseSocket(accepteds);
     }
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: end of processing this connection\n");
     fflush(VERBSTREAM);
#endif

     return cc;
}

/* receiver thread function */
static void receiver_function(HYAMIAGENT agent,
     HSEMAPHORE start_sem, HFLAG stopflag)
{
#ifndef YAMI_NO_THREADS

     struct yamiagent* pa;
     HSOCKET accepteds;
     int stopvalue, whatconn, ipaddr, ipport;
     int cc, cc2;

     pa = (struct yamiagent*)agent;

#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiver thread started\n");
     fflush(VERBSTREAM);
#endif

     /* signal the parent that is ready to work */
     cc = yamiSemaphoreRelease(start_sem);
     if (cc != YAMIOK)
     {
          CRASH_AGENT
          return;
     }
               
     while (pa->crashed == 0)
     {
          accepteds = NULL;
          
          /* do blocking select/accept */
          
          cc = yamiNetPoolAccept(pa->listensock, pa->receivepool,
               pa->receivedplxpool, pa->senddplxpool,
               &accepteds, pa->receiveridle, &whatconn,
               &ipaddr, &ipport);
          
          /* if accept failed, then we have a problem */
          if (cc != YAMIOK && cc != YAMINOTAVAILABLE && cc != YAMINOTFOUND)
          {
               CRASH_AGENT
               break;
          }

          /* here, check if the stopevent is set */
          cc2 = yamiFlagGet(stopflag, &stopvalue);
          if (cc2 != YAMIOK)
          {
               CRASH_AGENT
               break;
          }

          if (stopvalue != 0)
          {
               if (cc == YAMIOK && whatconn == 0)
               {
                    /* close new connection from the listening socket */
                    (void)yamiNetCloseSocket(accepteds);
               }
               break;
          }

          if (cc == YAMIOK)
          {
#ifdef VERBOSE
               fprintf(VERBSTREAM, "r: accepted something\n");
               fflush(VERBSTREAM);
#endif
          
               /* process the connection, but here ignore the results */
               (void)_process_connection(pa, accepteds, whatconn,
                    ipaddr, ipport);
          }
          else if (cc == YAMINOTFOUND)
          {
               /* this return value means              */
               /* that there was no socket to wait for */
               /* wait for the configured time         */

               cc = yamiSleep(pa->receiveridle);
               if (cc != YAMIOK)
               {
                    CRASH_AGENT
                    break;
               }
          }
          
          /* the YAMINOTAVAILABLE return value means that during */
          /* the given time there was no incoming connection     */
     }
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "r: receiver thread quitting\n");
     fflush(VERBSTREAM);
#endif

#endif /* YAMI_NO_THREADS */
}

/* helper for replying to messages */
static int _reply_shortcut(struct msgincoming *pincoming,
               HPARAMSET paramset, size_t parsetsize)
{
     struct yamiagent *pa;
     struct msgwaitingforreply *waiting;
     char *pparset = NULL;
     int replied;
     int cc;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: shortcut routing: replying to the message\n");
     fflush(VERBSTREAM);
#endif

     pa = pincoming->agent;

     if (parsetsize != 0)
     {     
          MALLOC(pparset, parsetsize)

          if (pparset == NULL)
               return YAMINOMEMORY;
     
          cc = yamiGetPacked(paramset, pparset);
          if (cc != YAMIOK)
          {
               FREE(pparset)
               return cc;
          }
     }
     
     /* find the message token */
     cc = yamiMutexLock(pa->msgwaiting_mutex);
     if (cc != YAMIOK)
     {
          if (pparset)
               FREE(pparset)

          return cc;
     }
     
     waiting = pa->msgwaiting_list;
     replied = 0;
     while (waiting)
     {
          if (waiting->msgid == pincoming->msgid)
          {
               /* message token was found */
               cc = yamiMutexLock(waiting->msg_mutex);
               if (cc != YAMIOK)
               {
                    if (pparset)
                         FREE(pparset)

                    (void)yamiMutexUnlock(pa->msgwaiting_mutex);
                    return cc;
               }
                                        
               /* the response message can be valid only        */
               /* if the status of the message token is pending */
               /* otherwise, we just ignore it - it may be a junk */
               if (waiting->status == ePending)
               {
#ifdef VERBOSE
                    fprintf(VERBSTREAM, "a: shortcut routing: writing data "
                         "to the waiting token for msgid = %d\n",
                         waiting->msgid);
                    fflush(VERBSTREAM);
#endif

                    waiting->status = eReplied;
                    waiting->data = pparset;
                    waiting->swap = 0;
                    replied = 1;
                    
                    /* push up the semaphore for this token */
                    cc = yamiSemaphoreRelease(waiting->msg_sem);
                    if (cc != YAMIOK)
                    {
                         (void)yamiMutexUnlock(pa->msgwaiting_mutex);
                         return cc;
                    }
               }

               cc = yamiMutexUnlock(waiting->msg_mutex);
               if (cc != YAMIOK)
               {
                    (void)yamiMutexUnlock(pa->msgwaiting_mutex);
                    return cc;
               }
                    
               break;
          }
          
          waiting = waiting->next;
     }
     
     if (replied == 0)
     {
          /* there was a soft error                       */
          /* - it's possible that with shortcut routing   */
          /* there is a reply to the message that has no  */
          /* waiting token                                */
          /* if this happens, it means that the client in */
          /* the same domain has destroyed the message    */
          /* token before reply arrives                   */
          if (pparset)
               FREE(pparset)

          (void)yamiMutexUnlock(pa->msgwaiting_mutex);

#ifdef VERBOSE
          fprintf(VERBSTREAM, "a: reply to the message with no "
               "waiting token - soft error\n");
          fflush(VERBSTREAM);
#endif

          return YAMIOK;
     }

     cc = yamiMutexUnlock(pa->msgwaiting_mutex);
     return cc;
}

/* helper for replying to messages */
static int _reply(struct msgincoming *pincoming, HPARAMSET paramset)
{
     struct yamiagent *pa;
     int psparamcount;
     int parsetsize;
     enum paramType partype;
     size_t totalsize;
     char *pbuffer, *p;
     struct sendhelp *help;
     int i, cc;
     
     /* check if the parameter set complies to the level */
     /* of the destination domain                        */

     if (paramset == NULL)
     {
          psparamcount = 0;
          parsetsize = 0;
     }
     else
     {
          cc = yamiGetParamCount(paramset, &psparamcount);
          if (cc != YAMIOK)
               return cc;

          if (pincoming->levelno == 1)
          {
               for (i = 0; i < psparamcount; ++i)
               {
                    cc = yamiGetType(paramset, i, &partype);
                    if (cc != YAMIOK)
                         return cc;
                    if (partype != eString && partype != eWString)
                         return YAMIBADTYPE;
               }
          }
     
          /* find the length of the parameter set after packing */
          cc = yamiGetPackedSize(paramset, &parsetsize);
          if (cc != YAMIOK)
               return cc;
     }
     
     pa = pincoming->agent;

     if (pincoming->connoptions == eNone)
     {
          /* check the possibility of the shortcut routing */
          if (pincoming->retipaddress == pa->ownipaddress &&
               pincoming->retipport == pa->ownipport)
               return _reply_shortcut(pincoming, paramset,
                    (size_t)parsetsize);
     }
     
     /* find the total size of the message to be sent */

     totalsize = 4               /* endianness */
          +   4                  /* levelno */
          +   4                  /* msgid */
          +   4                  /* packettype */
          +   4                  /* size of the parameter set (packed) */
          + parsetsize;          /* parameter set itself */

     MALLOC(pbuffer, totalsize)

     if (pbuffer == NULL)
          return YAMINOMEMORY;

     MALLOC(help, sizeof(struct sendhelp))

     if (help == NULL)
     {
          FREE(pbuffer)
          return YAMINOMEMORY;
     }
     
     /* compose the message */

     memset(pbuffer, 0, totalsize);
     p = pbuffer;
     p = yamiXDRWriteInt(p, pa->nativebyteorder); /* endianness */
     p = yamiXDRWriteInt(p, pa->ownlevelno);      /* levelno of the sender */
     p = yamiXDRWriteInt(p, pincoming->msgid);    /* message id */
     p = yamiXDRWriteInt(p, (int)RESPONSE);       /* packet type */
     p = yamiXDRWriteInt(p, parsetsize);          /* size of the paramset */
     if (parsetsize != 0)
     {
          cc = yamiGetPacked(paramset, p);/* parameter set itself */
          if (cc != YAMIOK)
          {
               FREE(pbuffer)
               FREE(help)
               return cc;
          }
     }
     
     /* prepare the helper structure */

     help->connoptions = pincoming->connoptions;
     if (pincoming->connoptions == eNone)
     {
          /* this is a reply to the simplex message         */
          /* - route it according to the "reply to" address */
          help->ipaddr = pincoming->retipaddress;
          help->ipport = pincoming->retipport;
     }
     else
     {
          /* this is a reply to the duplex message     */
          /* - route it ALWAYS to the original address */
          help->ipaddr = pincoming->origipaddress;
          help->ipport = pincoming->origipport;
     }
     
     help->size = totalsize;
     help->msgwaiting = NULL;
     help->agent = pincoming->agent;
     
     /* put this message into the sender's queue */
     
     cc = _queue_or_send(pa, pbuffer, help, totalsize, RESPONSE);
     if (cc != YAMIOK)
     {
          FREE(pbuffer)
          FREE(help)
          return cc;
     }

     return YAMIOK;
}

/* dispatcher thread function */
static void dispatcher_function(HYAMIAGENT agent,
     HSEMAPHORE start_sem, HFLAG stopflag)
{
#ifndef YAMI_NO_THREADS

     struct yamiagent *pa;
     struct object *po, *pbeg;
     size_t parsetsize;
     void *praw, *praw2;
     struct msgincoming *pincoming;
     HINCMSG inc;
     int stopvalue;
     int found, cc;
     
     pa = (struct yamiagent*)agent;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "d: dispatcher thread started\n");
     fflush(VERBSTREAM);
#endif
     
     cc = yamiSemaphoreRelease(start_sem);
     if (cc != YAMIOK)
     {
          CRASH_AGENT
          return;
     }
     
     while (1 == 1)
     {
          /* wait for work */
          cc = yamiSemaphoreAcquire(pa->dispatch_sem);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               break;
          }

          cc = yamiFlagGet(stopflag, &stopvalue);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               break;
          }

          if (stopvalue != 0)
               break;
          
#ifdef VERBOSE
          fprintf(VERBSTREAM, "d: hunting for a message\n");
          fflush(VERBSTREAM);
#endif

          cc = yamiMutexLock(pa->objects_mutex);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               break;
          }
          
          /* find an object with a message waiting for processing */
          
          pbeg = po = pa->objects_ring;
          found = 0;
          if (pbeg != NULL)
          {
               while (1 == 1)
               {
                    if (po->type == passive_multithreaded ||
                         po->type == passive_singlethreaded)
                    {
                         cc = yamiSemaphoreTryAcquire(po->incqueue_sem);
                         if (cc == YAMIOK)
                         {
                              /* found passive object                  */
                              /* with a message waiting for processing */
                              found = 1;
                              break;
                              
                         }
                         else if (cc == YAMINOTAVAILABLE)
                         {
                              /* this object does not have any messages */
                              /* just cuntinue hunting                  */
                         }
                         else
                         {
                              CRASH_AGENT
                              break;
                         }
                    }
               
                    po = po->next;
               
                    if (po == pbeg)
                    {
                         /* we have made a round-up the circle */
                         break;
                    }
               }
          }
          
          if (found == 0)
          {
               /* last chance - any object (if registered)    */
               if (pa->any_object != NULL)
               {
                    po = pa->any_object;
                    if (po->type == passive_multithreaded ||
                         po->type == passive_singlethreaded)
                    {
                         cc = yamiSemaphoreTryAcquire(po->incqueue_sem);
                         if (cc == YAMIOK)
                         {
                              /* any_object is passive   */
                              /* and has message waiting */
                              found = 1;
                         }
                         else if (cc == YAMINOTAVAILABLE)
                         {
                              /* any_object does not have any messages */
                         }
                         else
                         {
                              CRASH_AGENT
                         }
                    }
               }
               else
               {
                    break;
               }
          }
          
          if (pa->crashed != 0)
          {
               (void)yamiMutexUnlock(pa->objects_mutex);
               break;
          }
          
          /* it's possible that this thread didn't found */
          /* any message waiting - it can happen when    */
          /* the object with a non-empty message queue   */
          /* gets unregistered - the semaphore is high,  */
          /* but no message can be found                 */
          if (found == 0)
               continue;
          
#ifdef VERBOSE
          fprintf(VERBSTREAM, "d: found an object with a message waiting\n");
          fflush(VERBSTREAM);
#endif

          /* po points to the object with a message waiting */
          /* take the data from the incoming message queue  */

          cc = yamiMutexLock(po->local_mutex);
          if (cc != YAMIOK)
          {
               (void)yamiMutexUnlock(pa->objects_mutex);
               CRASH_AGENT
               break;
          }

          cc = yamiQueuePopFront(po->incqueue, &praw, &praw2, &parsetsize);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               (void)yamiMutexUnlock(po->local_mutex);
               (void)yamiMutexUnlock(pa->objects_mutex);
               break;
          }
          
          pincoming = (struct msgincoming*)praw;
          inc = pincoming;
                    
          /* increase the counter of threads working on this object */
          ++(po->threadson);
          
          if (po->type == passive_singlethreaded)
               po->processing = 1;
     
          cc = yamiMutexUnlock(po->local_mutex);
          if (cc != YAMIOK)
          {
               (void)yamiMutexUnlock(pa->objects_mutex);
               CRASH_AGENT
               _destroy_msgincoming(pincoming);
               FREE(pincoming)
               break;
          }
          
          cc = yamiMutexUnlock(pa->objects_mutex);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               _destroy_msgincoming(pincoming);
               FREE(pincoming)
               break;
          }
          
          /* call the actual dispatch function */
          
#ifdef VERBOSE
          fprintf(VERBSTREAM, "d: calling the dispatch function\n");
          fflush(VERBSTREAM);
#endif

          (po->dispf)(inc);

          /* the message should be considered rejected   */
          /* but only, if it wasn't marked as processed  */
          /* which would mean that the dispatch function */
          /* did reply or rejected itself                */
               
          if (pincoming->processed == 0)
          {

#ifdef VERBOSE
               fprintf(VERBSTREAM, "d: not successfully dispatched - "
                    "rejecting this message\n");
               fflush(VERBSTREAM);
#endif

               _respond_with_unkobj_of_rej(pincoming->agent, pincoming->msgid,
                    REJECT, pincoming->retipaddress, pincoming->retipport);
          }
          
          /* whatever the outcome of the dispatch, */
          /* release the msgincoming structure     */

          _destroy_msgincoming(pincoming);
          FREE(pincoming)

          cc = yamiMutexLock(po->local_mutex);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               break;
          }
          
          if (--(po->threadson) == 0 && po->tobedeleted == 1)
          {
               /* this means, that the object structure is no longer */
               /* in the objects' collection and no thread is        */
               /* working on it (and no one will) - we should        */
               /* destroy the object structure right now             */

               cc = yamiMutexUnlock(po->local_mutex);
               if (cc != YAMIOK)
               {
                    CRASH_AGENT
                    break;
               }
               
               _destroy_object(po);
               FREE(po)
          }
          else
          {
               if (po->type == passive_singlethreaded)
               {
                    /* if the object is singlethreaded, we have to */
                    /* adjust the semaphores                       */
               
                    po->processing = 0;
               
                    if (po->messages_waiting > 0)
                    {
                         --(po->messages_waiting);
                         cc = yamiSemaphoreRelease(po->incqueue_sem);
                         if (cc != YAMIOK)
                         {
                              CRASH_AGENT
                              (void)yamiMutexUnlock(po->local_mutex);
                              break;
                         }
                         cc = yamiSemaphoreRelease(pa->dispatch_sem);
                         if (cc != YAMIOK)
                         {
                              CRASH_AGENT
                              (void)yamiMutexUnlock(po->local_mutex);
                              break;
                         }
                    }
               }
               
               cc = yamiMutexUnlock(po->local_mutex);
               if (cc != YAMIOK)
                    CRASH_AGENT
          }
          
          if (pa->crashed == 1)
               break;
          
#ifdef VERBOSE
          fprintf(VERBSTREAM, "d: end processing this message\n");
          fflush(VERBSTREAM);
#endif
     }
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "d: dispatcher thread quitting\n");
     fflush(VERBSTREAM);
#endif

#endif /* YAMI_NO_THREADS */
}

/* this is a thread that wakes up clients waiting for a time out */
static void waker_function(HYAMIAGENT agent,
     HSEMAPHORE start_sem, HFLAG stopflag)
{
#ifndef YAMI_NO_THREADS

     struct yamiagent *pa;
     struct msgwaitingforreply *msg;
     size_t i;
     int stopvalue;
     int threadson;
     int cc;
     
     pa = (struct yamiagent*)agent;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "w: waker thread started\n");
     fflush(VERBSTREAM);
#endif

     /* signal the parent that is ready to work */
     cc = yamiSemaphoreRelease(start_sem);
     if (cc != YAMIOK)
     {
          CRASH_AGENT
          return;
     }
     
     while (1 == 1)
     {
          cc = yamiFlagGet(stopflag, &stopvalue);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               break;
          }

          if (stopvalue != 0 || pa->crashed != 0)
               break;
          
          cc = yamiMutexLock(pa->alarms_mutex);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               break;
          }

          if (pa->alarms != NULL)
          {
               for (i = 0; i < pa->alarms_size; ++i)
               {
                    if (pa->alarms[i].msg != NULL)
                    {
                         msg = pa->alarms[i].msg;

                         /* as an optimization step unconditionally destroy */
                         /* messages that have reference count == 1         */
                         /* (this means that only waker operates on them)   */
          
                         cc = yamiMutexLock(msg->msg_mutex);
                         if (cc != YAMIOK)
                         {
                              CRASH_AGENT
                              break;
                         }
                         
                         threadson = msg->threadson;

                         cc = yamiMutexUnlock(msg->msg_mutex);
                         if (cc != YAMIOK)
                         {
                              CRASH_AGENT
                              break;
                         }

                         if (threadson == 1)
                         {
                              /* waker is the only "user" of the message */
                              /* - destroy it                            */
                              _destroy_msgwaitingforreply(msg);
                              FREE(msg)

                              pa->alarms[i].msg = NULL;
                         }
                    }

                    /* for each message handled by the waiter, */
                    /* tick down its "clock" counter           */

                    if (pa->alarms[i].msg != NULL)
                    {
                         msg = pa->alarms[i].msg;
                         if (--(pa->alarms[i].timeout) == 0)
                         {
                              msg = pa->alarms[i].msg;
                              pa->alarms[i].msg = NULL;
                              
                              cc = yamiMutexLock(msg->msg_mutex);
                              if (cc != YAMIOK)
                              {
                                   CRASH_AGENT
                                   break;
                              }
                              
                              if (msg->status == ePending ||
                                   msg->status == ePosted)
                              {
#ifdef VERBOSE
                                   fprintf(VERBSTREAM,
                                        "w: message waiting timed out\n");
                                   fflush(VERBSTREAM);
#endif

                                   msg->status = eTimedOut;
                                   /* push up the semaphore that is         */
                                   /* associated with the message structure */
                                   cc = yamiSemaphoreRelease(msg->msg_sem);
                                   if (cc != YAMIOK)
                                   {
                                        CRASH_AGENT
                                        (void)yamiMutexUnlock(msg->msg_mutex);
                                        break;
                                   }
                              }
                              
                              threadson = --(msg->threadson);

                              cc = yamiMutexUnlock(msg->msg_mutex);
                              if (cc != YAMIOK)
                              {
                                   CRASH_AGENT
                                   break;
                              }

                              if (threadson == 0)
                              {
                                   /* no other thread will use */
                                   /* this message structure   */
                                   /* (not even client code)   */
                                   /* - destroy it             */
                                   _destroy_msgwaitingforreply(msg);
                                   FREE(msg)
                              }
                         }
                    }
               }
          }
          
          cc = yamiMutexUnlock(pa->alarms_mutex);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               break;
          }
          
          /* wait approx. 1 second                  */
          /* the precision does not matter -        */
          /* this functionality is to allow clients */
          /* to recover from the server crashes     */
          cc = yamiSleep(1000);
          if (cc != YAMIOK)
          {
               CRASH_AGENT
               break;
          }
     }     

#ifdef VERBOSE
     fprintf(VERBSTREAM, "w: waker thread quitting\n");
     fflush(VERBSTREAM);
#endif

#endif /* YAMI_NO_THREADS */
}

/* fill the given structure with the default policies */
void yamiGetDefaultPolicies(struct yamipolicies *ppolicies)
{
     ppolicies->agentlevel    = YAMIDEFAULT_AGENTLEVEL;
     ppolicies->objqmaxlength = YAMIDEFAULT_OBJQMAXLENGTH;
     ppolicies->objqmaxsize   = YAMIDEFAULT_OBJQMAXSIZE;
     ppolicies->dispatchers   = YAMIDEFAULT_DISPATCHERS;
     ppolicies->hassender     = YAMIDEFAULT_HASSENDER;
     ppolicies->sqmaxlength   = YAMIDEFAULT_SQMAXLENGTH;
     ppolicies->sqmaxsize     = YAMIDEFAULT_SQMAXSIZE;
     ppolicies->connpoolsize  = YAMIDEFAULT_CONNPOOLSIZE;
     ppolicies->sendtries     = YAMIDEFAULT_SENDTRIES;
     ppolicies->hassocket     = YAMIDEFAULT_HASSOCKET;
     ppolicies->reuseaddr     = YAMIDEFAULT_REUSEADDR;
     ppolicies->hasreceiver   = YAMIDEFAULT_HASRECEIVER;
     ppolicies->haswaker      = YAMIDEFAULT_HASWAKER;
     ppolicies->mtuser        = YAMIDEFAULT_MTUSER;
     ppolicies->allowduplex   = YAMIDEFAULT_ALLOWDUPLEX;
     ppolicies->maxdplxconns  = YAMIDEFAULT_MAXDPLXCONNS;
     ppolicies->receiveridle  = YAMIDEFAULT_RECEIVERIDLE;
}

/* create the agent with all its stuff: threads, queues, etc */
int yamiCreateAgent(HYAMIAGENT *pagent, int port,
               struct yamipolicies *ppolicies)
{
     struct yamipolicies policies;
     struct yamiagent *pa;
     int need_objects_mtx, need_msgwaiting_mtx;
     int need_local_mtx, need_msg_mtx, need_domains_mtx;
     int need_senderqueue_mtx, need_objectqueue_mtx;
     int need_sendpool_mtx;
     int i;
     int cc;

     /* apply policies */
     if (ppolicies == NULL)
          yamiGetDefaultPolicies(&policies);
     else
     {
          /* the flag fields can have any value */
          /* - they are only tested against 0   */
          if ((ppolicies->agentlevel != 1 && ppolicies->agentlevel != 2) ||
               ppolicies->objqmaxlength <= 0 ||
               ppolicies->objqmaxsize <= 0 ||
               ppolicies->dispatchers < 0 ||
               ppolicies->sqmaxlength <= 0 ||
               ppolicies->sqmaxsize <= 0 ||
               ppolicies->connpoolsize <= 0 ||
               ppolicies->sendtries <= 0 ||
               ppolicies->maxdplxconns < 0 ||
               ppolicies->receiveridle < 0)
               return YAMIOUTOFRANGE;
          
          memcpy(&policies, ppolicies, sizeof(struct yamipolicies));
     }

#ifdef YAMI_NO_THREADS
     /* simple check against YAMI_NO_THREADS compile flag             */
     /* if there is no thread support, all these policies should be 0 */
     /* which means that the everything runs in a single thread       */

     if (policies.dispatchers != 0 ||
         policies.hassender != 0 ||
         policies.hasreceiver != 0 ||
         policies.haswaker != 0 ||
         policies.mtuser != 0)
         return YAMIOUTOFRANGE;
#endif
     
     /* figure out what is needed in this configuration */

     /* collection of objects */
     need_objects_mtx     = (policies.hasreceiver != 0 ||
                             policies.dispatchers != 0 ||
                             policies.mtuser != 0);
     
     /* list of messages sent and waiting for responses */
     need_msgwaiting_mtx  = (policies.hasreceiver != 0 ||
                             policies.mtuser != 0);
     
     /* object's own mutex */
     need_local_mtx       = need_objects_mtx;
     
     /* message's own mutex */
     need_msg_mtx         = (policies.hassender != 0 ||
                             policies.hasreceiver != 0 ||
                             policies.mtuser != 0);
     
     /* domain set internal mutex */
     need_domains_mtx     = policies.mtuser;
     
     /* mutex for sender's queue */
     need_senderqueue_mtx = policies.hassender;
     
     /* mutex for sender pool if no sender thread */
     need_sendpool_mtx    = (policies.hassender == 0 &&
                            (policies.hasreceiver != 0 ||
                             policies.dispatchers != 0 ||
                             policies.mtuser != 0));
     
     /* mutex for object's own queue of incoming messages */
     need_objectqueue_mtx = policies.mtuser;
     

     /* create the agent object */

#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: creating agent\n");
#endif

     MALLOC(pa, sizeof(struct yamiagent))

     if (pa == NULL)
          return YAMINOMEMORY;
     
     memset(pa, 0, sizeof(struct yamiagent));

     pa->rtti = eYamiAgent;
          
     pa->crashed = 0;
     pa->ownlevelno = policies.agentlevel;
     pa->nativebyteorder = yamiXDRGetNativeByteOrder();
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: the native byte ordering is %s-endian\n",
          (pa->nativebyteorder == 0) ? "little" : "big");
     fflush(VERBSTREAM);
#endif
     
     /* get local host address */
     if ((cc = yamiNetGetLocalHost(&(pa->ownipaddress))) != YAMIOK)
     {
          FREE(pa)
          return cc;
     }
     
     pa->ownipport = port;

     /* create the domain set */

     if ((cc = yamiCreateDomainSet(
          &(pa->domains), need_domains_mtx)) != YAMIOK)
     {
          FREE(pa)
          return cc;
     }
     
     /* create the objects collection */

     if (need_objects_mtx != 0)
     {
          if ((cc = yamiCreateMutex(&(pa->objects_mutex))) != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }
     }
     else
     {
          pa->objects_mutex = NULL;
     }

     pa->objects_ring = NULL;
     
     pa->msgwaiting_list = NULL;

     if (need_msgwaiting_mtx != 0)
     {
          if ((cc = yamiCreateMutex(&(pa->msgwaiting_mutex))) != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }
     }
     else
     {
          pa->msgwaiting_mutex = NULL;
     }
     
     /* create the sender subsystem */
          
     pa->sendtries = policies.sendtries;
     
     /* the sending socket pool                                */
     /* this pool is created even when policies.hassender == 0 */
     /* because then the direct send will be performed         */

     if ((cc = yamiNetPoolCreate(&(pa->sendpool),
          policies.connpoolsize)) != YAMIOK)
     {
          (void)yamiDestroyAgent(pa);
          return cc;
     }
     
     /* the sending duplex socket dynamic pool */
     
     if ((cc = yamiNetDynPoolCreate(&(pa->senddplxpool), 0)) != YAMIOK)
     {
          (void)yamiDestroyAgent(pa);
          return cc;
     }
     
     /* and the queue, semaphore and the thread itself */
     
     if (policies.hassender != 0)
     {
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: creating the sender module\n");
     fflush(VERBSTREAM);
#endif

          if ((cc = yamiCreateQueue(&(pa->sendqueue),
               (size_t)policies.sqmaxlength, (size_t)policies.sqmaxsize,
               need_senderqueue_mtx)) != YAMIOK ||
              (cc = yamiCreateSemaphore(&(pa->sendqueue_sem), 0)) != YAMIOK ||
              (cc = yamiStartNewThread(&(pa->sender_th),
               sender_function, pa)) != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }
     }
     else
     {
          pa->sendqueue = NULL;
          pa->sendqueue_sem = NULL;
          pa->sender_th = NULL;
          
          /* but when there is no sender thread, the access to the */
          /* sending pool needs to be synchronized                 */
          
          if (need_sendpool_mtx != 0)
          {
               if ((cc = yamiCreateMutex(&(pa->sendpool_mutex))) != YAMIOK)
               {
                    (void)yamiDestroyAgent(pa);
                    return cc;
               }
          }
     }

     /* create the listening socket */
     
     if (policies.hassocket != 0)
     {
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: creating the listening pool\n");
     fflush(VERBSTREAM);
#endif
          cc = yamiNetListen(port, policies.reuseaddr, &(pa->listensock));
          if (cc != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }
          
          if (port == 0)
          {
               /* the listening socket was bound to the default port */
               /* read it and store "for further reference"          */
               
               cc = yamiNetGetPort(pa->listensock, &pa->ownipport);
               if (cc != YAMIOK)
               {
                    (void)yamiDestroyAgent(pa);
                    return cc;
               }
               
#ifdef VERBOSE
               fprintf(VERBSTREAM, "a: default port requested, using %d\n",
                    pa->ownipport);
               fflush(VERBSTREAM);
#endif
          }

          if ((cc = yamiNetPoolCreate(&(pa->receivepool),
               policies.connpoolsize)) != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }
          
          pa->allowduplex = policies.allowduplex;
          if (pa->allowduplex != 0)
          {
               /* create the pool of incoming duplex connections */
               if ((cc = yamiNetPoolCreate(&(pa->receivedplxpool),
                    policies.maxdplxconns)) != YAMIOK)
               {
                    (void)yamiDestroyAgent(pa);
                    return cc;
               }
          }
     }
     else
     {
          pa->listensock = NULL;
          pa->receivepool = NULL;
          pa->receivedplxpool = NULL;
     }
     
     /* create the receiving thread */
     
     if (policies.hasreceiver != 0)
     {
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: creating the receiver\n");
     fflush(VERBSTREAM);
#endif

          pa->receiveridle = policies.receiveridle;

          cc = yamiStartNewThread(&(pa->receiver_th), receiver_function, pa);
          if (cc != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }
     }
     else
     {
          pa->receiver_th = NULL;
     }

     /* create the dispatcher threads */

     if (policies.dispatchers != 0)
     {
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: creating the dispatchers\n");
     fflush(VERBSTREAM);
#endif

          cc = yamiCreateSemaphore(&(pa->dispatch_sem), 0);
          if (cc != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }

          MALLOC(pa->dispatchers, sizeof(HTHREAD) * policies.dispatchers)

          if (pa->dispatchers == NULL)          
          {
               (void)yamiDestroyAgent(pa);
               return YAMINOMEMORY;
          }
          pa->dispatchers_number = policies.dispatchers;
          for (i = 0; i < policies.dispatchers; ++i)
          {
               cc = yamiStartNewThread(&(pa->dispatchers[i]),
                    dispatcher_function, pa);
               if (cc != YAMIOK)
               {
                    (void)yamiDestroyAgent(pa);
                    return cc;
               }
               if (pa->crashed != 0)
                    return YAMIINTERNAL;
          }
     }
     else
     {
          pa->dispatch_sem = NULL;
          pa->dispatchers = NULL;
     }
     
     /* create the alarm system */
     
     if (policies.haswaker != 0)
     {
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: creating the waker subsystem\n");
     fflush(VERBSTREAM);
#endif

          cc = yamiCreateMutex(&(pa->alarms_mutex));
          if (cc != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }

          cc = yamiStartNewThread(&(pa->alarms_th), waker_function, pa);
          if (cc != YAMIOK)
          {
               (void)yamiDestroyAgent(pa);
               return cc;
          }
     }
     else
     {
          pa->alarms_th = NULL;
          pa->alarms_mutex = NULL;
     }
     
     /* some of the configuratioin options will be needed later */

     pa->need_objectqueue_mtx = need_objectqueue_mtx;
     pa->need_local_mtx = need_local_mtx;
     pa->need_msg_mtx = need_msg_mtx;
     pa->objqmaxlength = (size_t)policies.objqmaxlength;
     pa->objqmaxsize = (size_t)policies.objqmaxsize;

     *pagent = pa;
     return YAMIOK;
}

/* destroy the agent */
int yamiDestroyAgent(HYAMIAGENT agent)
{
     struct yamiagent *pa;
     struct object *o, *otmp;
     struct msgwaitingforreply *mw, *mwtmp;
     int cc;

     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     pa = (struct yamiagent*)agent;
     
     /* here, I don't care about the return codes */
     
     /* destroy the domain set */
     (void)yamiDestroyDomainSet(pa->domains);
     
     /* stop the receiver thread and its socket pools */
     (void)yamiStopReceiverThread(pa->receiver_th, pa->ownipaddress,
               pa->ownipport);
     (void)yamiNetCloseSocket(pa->listensock);
     (void)yamiNetPoolDestroy(pa->receivepool);
     (void)yamiNetPoolDestroy(pa->receivedplxpool);
     
     /* stop the sender thread and destroy it's queue and the socket pools */
     (void)yamiStopThread(pa->sender_th, pa->sendqueue_sem);
     (void)yamiDestroySemaphore(pa->sendqueue_sem);
     (void)yamiDestroyQueue(pa->sendqueue);
     (void)yamiNetPoolDestroy(pa->sendpool);
     (void)yamiNetDynPoolDestroy(pa->senddplxpool);
     (void)yamiDestroyMutex(pa->sendpool_mutex);
     
     /* stop the waker thread */
     (void)yamiStopThread(pa->alarms_th, NULL);
     (void)yamiDestroyMutex(pa->alarms_mutex);

     /* stop the dispatchers */
     if (pa->dispatchers != NULL)
     {
          (void)yamiStopManyThreads(pa->dispatchers, pa->dispatch_sem,
               pa->dispatchers_number);
          FREE(pa->dispatchers)
     }
     
     (void)yamiDestroySemaphore(pa->dispatch_sem);
     
     /* destroy the objects collection */
     if (pa->objects_ring != NULL)
     {
          o = pa->objects_ring->next;
          pa->objects_ring->next = NULL;
          while (o)
          {
               otmp = o->next;
               _destroy_object(o);
               FREE(o)
               o = otmp;
          }
     }
     (void)yamiDestroyMutex(pa->objects_mutex);

     /* destroy the collection of waiting mesages */
     mw = pa->msgwaiting_list;
     while (mw)
     {
          mwtmp = mw->next;
          _destroy_msgwaitingforreply(mw);
          FREE(mw)
          mw = mwtmp;
     }
     (void)yamiDestroyMutex(pa->msgwaiting_mutex);

     FREE(pa)
     
     return YAMIOK;
}

/*
 * stuff for domain set operation
 */

/* register new domain in a set */
int yamiAgentDomainRegister(HYAMIAGENT agent, const char *name,
     const char *addr, int port, int level)
{
     return yamiAgentDomainRegisterEx(agent, name,
          addr, port, level, eNone);
}

/* register new domain in a set, allow specifying connection options */
int yamiAgentDomainRegisterEx(HYAMIAGENT agent, const char *name,
     const char *addr, int port, int level, int options)
{
     struct yamiagent *pa;
     int cc;
     
     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     pa = (struct yamiagent*)agent;
     
     if (pa->crashed == 1)
          return YAMIINTERNAL;
     
     return yamiDomainSetRegister(pa->domains, name,
          addr, port, level, options);
}

/* unregister a domain */
int yamiAgentDomainUnregister(HYAMIAGENT agent, const char *name)
{
     struct yamiagent *pa;
     int cc;
     
     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     pa = (struct yamiagent*)agent;
     
     if (pa->crashed == 1)
          return YAMIINTERNAL;
     
     return yamiDomainSetUnregister(pa->domains, name);
}

/*
 * stuff for objects operation
 */

/* register object */
int yamiAgentObjectRegister(HYAMIAGENT agent, const char *name,
               enum objectType otype, DISPFUNCTION df,
               void *hint)
{
     struct yamiagent *pa;
     struct object *po, *po2, *newobject;
     int namelen = 0; /* to shut up compiler warning */
     int cc;
     
     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     if (name != YAMI_ANY_OBJECT)     
          if ((namelen = (int)strlen(name)) > MAXNAMELEN)
               return YAMIBADLENGTH;
     
     pa = (struct yamiagent*)agent;
     
     if (pa->crashed == 1)
          return YAMIINTERNAL;
     
     MALLOC(newobject, sizeof(struct object))

     if (newobject == NULL)
          return YAMINOMEMORY;
     
     memset(newobject, 0, sizeof(struct object));
     
     newobject->type = otype;
     newobject->hint = hint;
     if (otype == passive_singlethreaded || otype == passive_multithreaded)
          newobject->dispf = df;

     if (name != YAMI_ANY_OBJECT)
     {     
          MALLOC(newobject->name, (size_t)(namelen + 1))

          if (newobject->name == NULL)
          {
               _destroy_object(newobject);
               FREE(newobject)
               return YAMINOMEMORY;
          }
          strcpy(newobject->name, name);
     }
     
     /* the mutex in this queue is created only    */
     /* when the user code declares multithreading */
     
     cc = yamiCreateQueue(&(newobject->incqueue),
          pa->objqmaxlength, pa->objqmaxsize, pa->need_objectqueue_mtx);
     if (cc != YAMIOK)
     {
          _destroy_object(newobject);
          FREE(newobject)
          return cc;
     }
     
     /* but this semaphore is created always             */
     /* note: this semaphore works like a plain counter  */
     /* when YAMI_NO_THREADS flag is set for compiletion */
     
     cc = yamiCreateSemaphore(&(newobject->incqueue_sem), 0);
     if (cc != YAMIOK)
     {
          _destroy_object(newobject);
          FREE(newobject)
          return cc;
     }
     
     /* this mutex is not always necessary */
     
     if (pa->need_local_mtx != 0)
     {
          cc = yamiCreateMutex(&(newobject->local_mutex));
          if (cc != YAMIOK)
          {
               _destroy_object(newobject);
               FREE(newobject)
               return cc;
          }
     }
     else
     {
          newobject->local_mutex = NULL;
     }
     
     /* add this new object structure to the collection */
     
     cc = yamiMutexLock(pa->objects_mutex);
     if (cc != YAMIOK)
     {
          _destroy_object(newobject);
          FREE(newobject)
          return cc;
     }

     /* check if such an object is already registered */
     
     if (name == YAMI_ANY_OBJECT)
     {
          if (pa->any_object != NULL)
          {
               _destroy_object(newobject);
               FREE(newobject)
               (void)yamiMutexUnlock(pa->objects_mutex);
               return YAMIALREADYREGISTERED;
          }

          pa->any_object = newobject;

          cc = yamiMutexUnlock(pa->objects_mutex);
          return cc;
     }

     /* normal object - check the list of objects */
     
     po2 = NULL;
     po = pa->objects_ring;
     if (po)
          while (po != po2)
          {
               po2 = pa->objects_ring;

               if (strcmp(po->name, name) == 0)
               {
                    _destroy_object(newobject);
                    FREE(newobject)
                    (void)yamiMutexUnlock(pa->objects_mutex);
                    return YAMIALREADYREGISTERED;
               }
               po = po->next;
          }

     /* add it */
     
     if (pa->objects_ring == NULL)
     {
          newobject->next = newobject;
          pa->objects_ring = newobject;
     }
     else
     {
          newobject->next = pa->objects_ring->next;
          pa->objects_ring->next = newobject;
     }
     
     cc = yamiMutexUnlock(pa->objects_mutex);
     return cc;
}

/* helper for removing objects */
static int _markfordeletion(struct object *po)
{
     void *praw, *praw2;
     size_t parsetsize;
     struct msgincoming *pincoming;
     int cc;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: marking the object for deletion\n");
     fflush(VERBSTREAM);
#endif

     /* lock the object */
     cc = yamiMutexLock(po->local_mutex);
     if (cc != YAMIOK)
          return cc;
     
     /* now, discard all the messages that are   */
     /* in the incoming message queue with the   */
     /* proper notification to the remote object */
     while (1 == 1)
     {
          cc = yamiQueuePopFront(po->incqueue, &praw, &praw2, &parsetsize);
          if (cc == YAMIQUEUEEMPTY)
               break;
          if (cc != YAMIOK)
          {
               (void)yamiMutexUnlock(po->local_mutex);
               return cc;
          }
     
          pincoming = (struct msgincoming*)praw;
#ifdef VERBOSE
          fprintf(VERBSTREAM, "a: discarding an incoming message\n");
          fflush(VERBSTREAM);
#endif

          _respond_with_unkobj_of_rej(pincoming->agent, pincoming->msgid,
               UNKOBJECT, pincoming->retipaddress, pincoming->retipport);

          _destroy_msgincoming(pincoming);
          FREE(pincoming)
     }
     
     if (po->threadson == 0)
     {
          /* this means, that no dispatcher thread */
          /* is making any operation on the object */
          cc = yamiMutexUnlock(po->local_mutex);

          if (cc != YAMIOK)
               return cc;

          _destroy_object(po);
          FREE(po)
     }
     else
     {
          /* this means, that one or more dispatchers */
          /* are making operations on the object      */
          po->tobedeleted = 1;
                         
          cc = yamiMutexUnlock(po->local_mutex);
          if (cc != YAMIOK)
               return cc;
     }
     
     return YAMIOK;
}

/* unregister object */
int yamiAgentObjectUnregister(HYAMIAGENT agent, const char *name)
{
     struct yamiagent *pa;
     struct object *po, **ppo;
     int found, firstturn;
     int cc;
     
     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     pa = (struct yamiagent*)agent;

     if (pa->crashed == 1)
          return YAMIINTERNAL;
     
     cc = yamiMutexLock(pa->objects_mutex);
     if (cc != YAMIOK)
          return cc;

     /* find this object in a collection */

     found = 0;     
     if (name == YAMI_ANY_OBJECT)
     {
          if (pa->any_object != NULL)
          {
               po = pa->any_object;
               pa->any_object = NULL;

               /* unlock the whole collection                  */
               /* now, the object structure is no longer in it */
               cc = yamiMutexUnlock(pa->objects_mutex);
               if (cc != YAMIOK)
                    return cc;
                    
               /* mark the object for deletion */
               cc = _markfordeletion(po);
               if (cc != YAMIOK)
                    return cc;
               
               found = 1;
          }
     }
     else
     {
          if (pa->objects_ring != NULL)
          {
               po = pa->objects_ring->next;
               ppo = &(pa->objects_ring->next);
               firstturn = 1;
               while (firstturn != 0)
               {
                    if (strcmp(po->name, name) == 0)
                    {
                         if (po == po->next)
                         {
                              /* only one object in a collection */
                              pa->objects_ring = NULL;
                         }
                         else
                         {
                              *ppo = po->next;
                              if (po == pa->objects_ring)
                                   pa->objects_ring = po->next;
                         }
                    
                         /* unlock the whole collection                  */
                         /* now, the object structure is no longer in it */
                         cc = yamiMutexUnlock(pa->objects_mutex);
                         if (cc != YAMIOK)
                              return cc;
                    
                         /* mark the object for deletion */
                         cc = _markfordeletion(po);
                         if (cc != YAMIOK)
                              return cc;

                         found = 1;
                    
                         break;
                    }
               
                    if (po == pa->objects_ring)
                         firstturn = 0;
               
                    ppo = &(po->next);
                    po = po->next;
               }
          }
     }
     
     if (found == 0)
     {
          cc = yamiMutexUnlock(pa->objects_mutex);
          if (cc != YAMIOK)
               return cc;
     }
     
     return (found != 0) ? YAMIOK : YAMINOTFOUND;
}

/*
 * stuff for message handling
 */

/* generate new ID */
static int _getNewID(struct yamiagent *pa)
{
     return ++(pa->lastID);
}

/* stuff for new messages */

/* helper for creating waiting message tokens */
static int _create_waiting(struct yamiagent *pa, int msgid,
     enum msgStatus status, int initthreadcount,
     HMESSAGE *phm)
{
     struct msgwaitingforreply *waiting;
     int cc;

     MALLOC(waiting, sizeof(struct msgwaitingforreply))

     if (waiting == NULL)
          return YAMINOMEMORY;
          
     waiting->rtti = eSentMessage;
          
     waiting->threadson = initthreadcount;
     
     waiting->msgid = msgid;
     waiting->status = status;
     waiting->data = NULL;
     waiting->agent = pa;
     waiting->next = waiting->prev = NULL;

     /* this mutex is not always needed */
     
     if (pa->need_msg_mtx != 0)
     {
          cc = yamiCreateMutex(&(waiting->msg_mutex));
          if (cc != YAMIOK)
          {
               FREE(waiting)
               return cc;
          }
     }
     else
     {
          waiting->msg_mutex = NULL;
     }

     /* but this semaphore is always created */
     
     cc = yamiCreateSemaphore(&(waiting->msg_sem), 0);
     if (cc != YAMIOK)
     {
          (void)yamiDestroyMutex(waiting->msg_mutex);
          FREE(waiting)
          return cc;
     }
     
     /* connect it to the list of tokens */
     
     cc = yamiMutexLock(pa->msgwaiting_mutex);
     if (cc != YAMIOK)
     {
          (void)yamiDestroySemaphore(waiting->msg_sem);
          (void)yamiDestroyMutex(waiting->msg_mutex);
          FREE(waiting)
          return cc;
     }

     if (pa->msgwaiting_list == NULL)
     {
          pa->msgwaiting_list = waiting;
     }
     else
     {
          pa->msgwaiting_list->prev = waiting;
          waiting->next = pa->msgwaiting_list;
          pa->msgwaiting_list = waiting;
     }
     
     *phm = waiting;
     
     cc = yamiMutexUnlock(pa->msgwaiting_mutex);
     return cc;
}

/* send or forward the message using          */
/* shortcut routing                           */
/* sendforward == 0 means sending new message */
/* (phm receives the structure, unless NULL), */
/* otherwise it means forwarding a message    */
/* (phm not used)                             */
static int _sendforward_shortcut(struct yamiagent *pa,
          const char *objectname, const char *messagename,
          HPARAMSET paramset, int msgid,
          int retipaddr, int retipport, int originlevel,
          int sendforward, HMESSAGE *phm)
{
     int psparamcount, parsetsize;
     enum paramType partype;
     char *pobjname, *pmsgname, *pparset;
     struct object *po;
     int queued, objsmtxunlocked;
     int objectnamelen, messagenamelen;
     
     int i, cc;

#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: shortcut routing to object %s\n", objectname);
     fflush(VERBSTREAM);
#endif

     /* find the lengths of the object and message names */
     objectnamelen = (int)strlen(objectname);
     messagenamelen = (int)strlen(messagename);
     if (messagenamelen > MAXNAMELEN || objectnamelen > MAXNAMELEN)
          return YAMIBADLENGTH;
     
     /* check if the parameter set complies to the level of this domain */
     if (paramset == NULL)
     {
          psparamcount = 0;
          parsetsize = 0;
     }
     else
     {
          cc = yamiGetParamCount(paramset, &psparamcount);
          if (cc != YAMIOK)
               return cc;

          if (pa->ownlevelno == 1)
          {
               for (i = 0; i < psparamcount; ++i)
               {
                    cc = yamiGetType(paramset, i, &partype);
                    if (cc != YAMIOK)
                         return cc;
                    if (partype != eString && partype != eWString)
                         return YAMIBADTYPE;
               }
          }
     
          /* find the length of the parameter set after packing */
          cc = yamiGetPackedSize(paramset, &parsetsize);
          if (cc != YAMIOK)
               return cc;
     }
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: there is %d parameters in a set, "
          "total size of the set is %d\n",
          psparamcount, parsetsize);
     fflush(VERBSTREAM);
#endif
     
     if (sendforward == 0)
     {
          /* we are sending new message, not forwarding it */     
          /* prepare the message token */

          if (phm != NULL)
          {
               /* with shortcut routing, the sended thread is not */
               /* used, so the only thread that will use the new  */
               /* object is the client thread - the initial       */
               /* thread count should be 1                        */

               cc = _create_waiting(pa, msgid, ePending, 1, phm);
               if (cc != YAMIOK)
                    return cc;
          }
          else
          {
               /* no waiting structure - this new message is one-way */
          }
     }
     else
     {
          /* we are forwarding a message - no waiting structure */
          /* needed here, it's already somewhere                */
     }

     /* find the object in a collection */
     
     cc = yamiMutexLock(pa->objects_mutex);
     if (cc != YAMIOK)
          return cc;
     
     po = _find_object(pa, objectname, 0);
     
     if (po != NULL)
     {
          /* object with a given name was found (po) */
          
          /* prepare the data to push to the object's queue */
          
          MALLOC(pmsgname, (size_t)(messagenamelen + 1))
          
          if (pmsgname == NULL)
          {
               (void)yamiMutexUnlock(pa->objects_mutex);
               return YAMINOMEMORY;
          }
          
          MALLOC(pobjname, (size_t)(objectnamelen + 1))
          
          if (pobjname == NULL)
          {
               FREE(pmsgname)
               (void)yamiMutexUnlock(pa->objects_mutex);
               return YAMINOMEMORY;
          }
          
          if (parsetsize != 0)
          {
               MALLOC(pparset, (size_t)parsetsize)

               if (pparset == NULL)
               {
                    FREE(pmsgname)
                    FREE(pobjname)
                    (void)yamiMutexUnlock(pa->objects_mutex);
                    return YAMINOMEMORY;
               }
                    
               cc = yamiGetPacked(paramset, pparset);
               if (cc != YAMIOK)
               {
                    FREE(pmsgname)
                    FREE(pobjname)
                    FREE(pparset)
                    (void)yamiMutexUnlock(pa->objects_mutex);
                    return cc;
               }
          }
          else
          {
               pparset = NULL;
          }

          strcpy(pmsgname, messagename);
          strcpy(pobjname, objectname);
          
          cc = _queue_incoming(pa, po, pobjname, pmsgname, pparset,
               (size_t)parsetsize, 0, originlevel, msgid,
               retipaddr, retipport, 0, 0, eNone, &queued, &objsmtxunlocked);
          
          if (queued == 0)
          {
               FREE(pmsgname)
               FREE(pobjname)
               FREE(pparset)
          }
          
          if (cc != YAMIOK)
          {
               if (cc == YAMIOVERFLOW)
               {
                    /* overflow in the object's queue */
                    /* send the overflow notification */
                    _respond_with_unkobj_of_rej(pa, msgid,
                         OVERFLOW, retipaddr, retipport);
               }
               else
                    return cc;
          }

          if (objsmtxunlocked == 0)
          {
               return yamiMutexUnlock(pa->objects_mutex);
          }
          else
          {
               return YAMIOK;
          }
     }
     else
     {
          /* object not found */

          cc = yamiMutexUnlock(pa->objects_mutex);
          if (cc != YAMIOK)
               return cc;

#ifdef VERBOSE
          fprintf(VERBSTREAM, "a: shortcut routing: object with this "
               "name not found\n");
          fflush(VERBSTREAM);
#endif

          _respond_with_unkobj_of_rej(pa, msgid,
               UNKOBJECT, retipaddr, retipport);
     }

     return YAMIOK;
}

/* send or forward the message to the given address */
/* sendforward == 0 means sending new message       */
/* (phm receives the structure, unless NULL),       */
/* otherwise it means forwarding a message          */
/* (then - phm is not used)                         */
static int _sendforward_addr(struct yamiagent *pa, int domainipaddr,
          int domainipport, int domainlevel, int connoptions,
          const char *objectname, const char *messagename,
          HPARAMSET paramset, int msgid, int retipaddr, int retipport,
          int originlevel, int sendforward, HMESSAGE *phm)
{
     int psparamcount, parsetsize;
     enum paramType partype;
     int objectnamelen;
     int messagenamelen;
     size_t totalsize;
     char *pbuffer, *p;
     struct sendhelp *help;
     struct msgwaitingforreply *waiting;
     int i, cc;

     /* resolve the string name into the IP address */

     /* check if the message is addressed to the local domain */
     /* and perform the shortcut routing                      */
     if (domainipaddr == pa->ownipaddress && domainipport == pa->ownipport)
     {
          return _sendforward_shortcut(pa, objectname, messagename,
                    paramset, msgid, retipaddr, retipport, originlevel,
                    sendforward, phm);
     }

     /* check if the parameter set complies to the level */
     /* of the destination domain                        */
     if (paramset == NULL)
     {
          psparamcount = 0;
          parsetsize = 0;
     }
     else
     {
          cc = yamiGetParamCount(paramset, &psparamcount);
          if (cc != YAMIOK)
               return cc;

          if (domainlevel == 1)
          {
               for (i = 0; i < psparamcount; ++i)
               {
                    cc = yamiGetType(paramset, i, &partype);
                    if (cc != YAMIOK)
                         return cc;
                    if (partype != eString && partype != eWString)
                         return YAMIBADTYPE;
               }
          }
     
          /* find the length of the parameter set after packing */
          cc = yamiGetPackedSize(paramset, &parsetsize);
          if (cc != YAMIOK)
               return cc;
     }
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: there is %d parameters in a set, "
          "total size of the set is %d\n",
          psparamcount, parsetsize);
     fflush(VERBSTREAM);
#endif
     
     /* find the lengths of the object and message names */
     objectnamelen = (int)strlen(objectname);
     messagenamelen = (int)strlen(messagename);
     if (messagenamelen > MAXNAMELEN || objectnamelen > MAXNAMELEN)
          return YAMIBADLENGTH;

     /* find the total size of the message to be sent */
     totalsize = 4                      /* endianness */
          +   4                         /* levelno */
          +   4                         /* msgid */
          +   4                         /* packettype */
          +   8                         /* retaddr == 2 * int */
          +   4                         /* length of object name */
          + yamiXDRRoundTo4(objectnamelen)  /* object name itself */
          +   4                         /* length of message name */
          + yamiXDRRoundTo4(messagenamelen) /* messagename itself */
          +   4                         /* size of the paramset (packed) */
          + parsetsize;                 /* parameter set itself */

     MALLOC(pbuffer, totalsize)

     if (pbuffer == NULL)
          return YAMINOMEMORY;

     MALLOC(help, sizeof(struct sendhelp))

     if (help == NULL)
     {
          FREE(pbuffer)
          return YAMINOMEMORY;
     }
     
     /* compose the message */
     
     if (sendforward != 0)
     {
          /* this is a forward         */
          /* ignore connection options */
          connoptions = 0;
     }

#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: composing message\n");
     fflush(VERBSTREAM);
#endif
     
     memset(pbuffer, 0, totalsize);
     p = pbuffer;
     p = yamiXDRWriteInt(p, pa->nativebyteorder); /* endianness */
     p = yamiXDRWriteInt(p, originlevel);    /* levelno of the orig. sender */
     p = yamiXDRWriteInt(p, msgid);          /* message id */
     p = yamiXDRWriteInt(p, (int)REQUEST
                | (int)connoptions << 16);   /* packet type and conn options*/
     *(int*)p = retipaddr;                   /* IP return address, */
                                             /* already in net-byte order */
     p += sizeof(int);
     p = yamiXDRWriteInt(p, retipport);      /* IP return port */
     p = yamiXDRWriteInt(p, objectnamelen);  /* object name length */
     strcpy(p, objectname);                  /* object name itself */
     p += yamiXDRRoundTo4(objectnamelen);
     p = yamiXDRWriteInt(p, messagenamelen); /* message name length */
     strcpy(p, messagename);                 /* message name itself */
     p += yamiXDRRoundTo4(messagenamelen);
     p = yamiXDRWriteInt(p, parsetsize);     /* size of the parameter set */
     if (parsetsize != 0)
     {
          cc = yamiGetPacked(paramset, p);   /* parameter set itself */
          if (cc != YAMIOK)
          {
               FREE(pbuffer)
               FREE(help)
               return cc;
          }
     }

     if (sendforward == 0)
     {
          /* we are sending new message, not forwarding it */     
          /* prepare the message token */

          if (phm != NULL)
          {
               /* initially there are two threads that can use */
               /* the new message object - the client thread   */
               /* and the sender thread; the initial thread    */
               /* count should be therefore 2                  */

               cc = _create_waiting(pa, msgid, ePosted, 2, phm);
               if (cc != YAMIOK)
               {
                    FREE(pbuffer)
                    FREE(help)
                    return cc;
               }
               waiting = (struct msgwaitingforreply*)(*phm);
          }
          else
          {
               /* it is a new message, but the user does not want to */
               /* create the message token - the message is one-way  */
               waiting = NULL;
          }
     }
     else
     {
          /* we are forwarding a message - no waiting structure */
          /* needed here, it's already somewhere else           */
          /* (in the originating Agent)                         */
          waiting = NULL;
     }
     
     /* prepare the helper structure */

     help->ipaddr = domainipaddr;
     help->ipport = domainipport;
     help->connoptions = connoptions;
     help->msgwaiting = waiting;
     help->size = totalsize;
     help->agent = pa;
     
     /* put this message into the sender's queue */

     cc = _queue_or_send(pa, pbuffer, help, totalsize, REQUEST);
     if (cc != YAMIOK)
     {
          FREE(pbuffer)
          FREE(help)
          return cc;
     }
     
     return YAMIOK;
}

/* send or forward the message to the given domain */
/* sendforward == 0 means sending new message      */
/* (phm receives the structure, unless NULL),      */
/* otherwise it means forwarding a message         */
/* (then - phm is not used)                        */
/* this function resolves the domain name and      */
/* calls the _sendforward_addr or *_shortcut       */
static int _sendforward_domain(struct yamiagent *pa, const char *domainname,
          const char *objectname, const char *messagename,
          HPARAMSET paramset, int msgid, int retipaddr, int retipport,
          int originlevel, int sendforward, HMESSAGE *phm)
{
     int domainipaddr;
     int domainipport;
     int domainlevel;
     int domainoptions;
     int cc;

     /* look up the domain */
     
     if (domainname == YAMI_THIS_DOMAIN)
     {
          return _sendforward_shortcut(pa, objectname, messagename,
                    paramset, msgid, retipaddr, retipport,
                    originlevel, sendforward, phm);
     }
     
     cc = yamiDomainSetLookUp(pa->domains, domainname, &domainipaddr,
          &domainipport, &domainlevel, &domainoptions);
     if (cc != YAMIOK)
          return cc;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM,
          "a: found domain %s: ip = %u, port = %d, level = %d, opts = %d\n",
          domainname, domainipaddr, domainipport,
          domainlevel, domainoptions);
     fflush(VERBSTREAM);
#endif

     /* once the destination domain details are know, send the message */

     return _sendforward_addr(pa, domainipaddr, domainipport, domainlevel,
          domainoptions, objectname, messagename, paramset, msgid, retipaddr,
          retipport, originlevel, sendforward, phm);
}

/* send a new message to the given domain */
int yamiAgentMsgSend(HYAMIAGENT agent, const char *domainname,
          const char *objectname, const char *messagename,
          HPARAMSET paramset, HMESSAGE *phm)
{
     struct yamiagent *pa;
     int msgid;
     int cc;

     /* paramset CAN be NULL, which means that the paramset is empty */

     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     if (objectname == YAMI_ANY_OBJECT)
          return YAMIBADHANDLE;
     
     pa = (struct yamiagent *)agent;
     if (pa->crashed == 1)
          return YAMIINTERNAL;

     /* generate new msgid */
     msgid = _getNewID(pa);
     
     cc = _sendforward_domain(pa, domainname, objectname, messagename,
          paramset, msgid, pa->ownipaddress, pa->ownipport,
          pa->ownlevelno, 0, phm);
     return cc;
}

/* send a new message with the explicit address */
int yamiAgentMsgSendAddr(HYAMIAGENT agent, const char *domainaddr,
          int domainport, int domainlevel,
          const char *objectname, const char *messagename,
          HPARAMSET paramset, HMESSAGE *phm)
{
     return yamiAgentMsgSendAddrEx(agent, domainaddr, domainport,
          domainlevel, objectname, messagename, paramset, phm, eNone);
}

/* send a new message with the explicit address */
/* allow to specify the connection options      */
int yamiAgentMsgSendAddrEx(HYAMIAGENT agent, const char *domainaddr,
          int domainport, int domainlevel,
          const char *objectname, const char *messagename,
          HPARAMSET paramset, HMESSAGE *phm, int options)
{
     struct yamiagent *pa;
     int domainipaddr;
     int msgid;
     int cc;

     /* paramset CAN be NULL, which means that the paramset is empty */

     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     if (objectname == YAMI_ANY_OBJECT)
          return YAMIBADHANDLE;
     
     pa = (struct yamiagent *)agent;
     if (pa->crashed == 1)
          return YAMIINTERNAL;

     /* resolve the string address into the IP address */
     cc = yamiNetResolveIPAddress(domainaddr, &domainipaddr);
     if (cc != YAMIOK)
          return cc;

     /* generate new msgid */
     msgid = _getNewID(pa);
     
     cc = _sendforward_addr(pa, domainipaddr, domainport, domainlevel,
          options, objectname, messagename,
          paramset, msgid, pa->ownipaddress, pa->ownipport,
          pa->ownlevelno, 0, phm);
     return cc;
}

/* set up a timeout alarm for a sent message */
/* the timeout is in seconds (not precise)   */
int yamiAgentSetUpTimeOut(HMESSAGE message, int timeout)
{
     struct yamiagent *pa;
     struct msgwaitingforreply *msg;
     struct alarmobject *newalarm;
     size_t newsize, newsizebytes;
     size_t i;
     int cc;
     
     if ((cc = _checkMESSAGE(message)) != YAMIOK)
          return cc;
     
     if (timeout <= 0)
          return YAMIOUTOFRANGE;
     
     msg = (struct msgwaitingforreply*)message;
     pa = msg->agent;

     if (pa->alarms_th == NULL)
     {
          /* the waker thread was not created */
          /* (policies.haswaker == 0)         */

          return YAMINOSUCHSERVICE;
     }
          
     if ((cc = yamiMutexLock(pa->alarms_mutex)) != YAMIOK)
          return cc;
     
     if (pa->alarms == NULL)
     {
          newsizebytes = 5 * sizeof(struct alarmobject);

          MALLOC(newalarm, newsizebytes)

          if (newalarm == NULL)
          {
               (void)yamiMutexUnlock(pa->alarms_mutex);
               return YAMINOMEMORY;
          }

          memset(newalarm, 0, newsizebytes);
          pa->alarms = newalarm;
          pa->alarms_size = 5;
     }
     else
     {
          for (i = 0; i < pa->alarms_size; ++i)
               if (pa->alarms[i].msg == NULL)
                    break;
          if (i == pa->alarms_size)
          {
               /* there is no space in the alarm table */
               /* - resize it                          */
               newsize = pa->alarms_size * 3 / 2;
               newsizebytes = newsize * sizeof(struct alarmobject);

               MALLOC(newalarm, newsizebytes)

               if (newalarm == NULL)
               {
                    (void)yamiMutexUnlock(pa->alarms_mutex);
                    return YAMINOMEMORY;
               }

               memset(newalarm, 0, newsizebytes);
               memcpy(newalarm, pa->alarms,
                    pa->alarms_size * sizeof(struct alarmobject));
               FREE(pa->alarms)
               pa->alarms = newalarm;
               pa->alarms_size = newsize;
          }
          newalarm = &(pa->alarms[i]);
     }
     
     newalarm->timeout = timeout;
     newalarm->msg = msg;

     /* lazy creation of the msg's mutex, if not yet created         */
     /* the reason is that the user can (not very likely, but let's  */
     /* make it complete) send a message with the agent where the    */
     /* optimization prevented from normal creation of this mutex    */
     /* see also yamiCreateAgent for those optimizations             */
     /* If fact, this mutex is needed only when more than one thread */
     /* can perform operations on the message object. Registering    */
     /* the message in the waker thread makes it true,               */
     /* no matter what are the policies.                             */
     if (msg->msg_mutex == NULL)
     {
          if ((cc = yamiCreateMutex(&(msg->msg_mutex))) != YAMIOK)
          {
               (void)yamiMutexUnlock(pa->alarms_mutex);
               return cc;
          }
     }
     
     if ((cc = yamiMutexLock(msg->msg_mutex)) != YAMIOK)
     {
          (void)yamiMutexUnlock(pa->alarms_mutex);
          return cc;
     }
     
     /* push up the reference counter on the message structure */
     /* so that it will not vanish until the waker thread      */
     /* comes to process the alarm                             */
     ++(msg->threadson);
     
     if ((cc = yamiMutexUnlock(msg->msg_mutex)) != YAMIOK)
     {
          (void)yamiMutexUnlock(pa->alarms_mutex);
          return cc;
     }
     
     return yamiMutexUnlock(pa->alarms_mutex);
}

/* get the status of the message sent previously */
int yamiAgentMsgGetStatus(HMESSAGE hm, enum msgStatus *status)
{
     struct msgwaitingforreply *msg;
     int cc;
     
     if ((cc = _checkMESSAGE(hm)) != YAMIOK)
          return cc;
     
     msg = (struct msgwaitingforreply*)hm;
     
     cc = yamiMutexLock(msg->msg_mutex);
     if (cc != YAMIOK)
          return cc;
     
     *status = msg->status;
     
     cc = yamiMutexUnlock(msg->msg_mutex);
     return cc;
}

/* this function is a helper for clients to wait for responses */
int yamiAgentMsgWait(HMESSAGE hm)
{
     struct msgwaitingforreply *msg;
     int cc;
     
     if ((cc = _checkMESSAGE(hm)) != YAMIOK)
          return cc;
     
     msg = (struct msgwaitingforreply*)hm;
     
     cc = yamiSemaphoreAcquire(msg->msg_sem);
     
     return cc;
}

/* get the response from the server */
int yamiAgentMsgGetResponse(HMESSAGE hm, HPARAMSET *pps)
{
     struct msgwaitingforreply *msg;
     int cc1, cc2;
     
     if ((cc1 = _checkMESSAGE(hm)) != YAMIOK)
          return cc1;
     
     msg = (struct msgwaitingforreply*)hm;
     
     cc1 = yamiMutexLock(msg->msg_mutex);
     if (cc1 != YAMIOK)
          return cc1;

     if (msg->status != eReplied)
     {
          /* no response available */
          cc2 = YAMINOTAVAILABLE;
     }
     else
     {
          /* retrieve the parameter set */
          if (msg->data == NULL)
          {
               *pps = NULL;
               cc2 = YAMINOTAVAILABLE;
          }
          else
               cc2 = yamiCreatePSFromPacked(pps, msg->agent->ownlevelno,
                    msg->data, msg->swap);
     }
     
     cc1 = yamiMutexUnlock(msg->msg_mutex);

     return (cc1 != YAMIOK) ? cc1 : cc2;     
}

/* destroy sent message token */
int yamiAgentMsgDestroy(HMESSAGE hm)
{
     struct msgwaitingforreply *msg;
     struct yamiagent *pa;
     int cc;

     if ((cc = _checkMESSAGE(hm)) != YAMIOK)
          return cc;
     
     msg = (struct msgwaitingforreply*)hm;
     pa = msg->agent;
     
     /* remove this structure from the collection */
     
     cc = yamiMutexLock(pa->msgwaiting_mutex);
     if (cc != YAMIOK)
          return cc;
     
     if (msg->prev == NULL)
     {
          /* this is the very first structure */
          pa->msgwaiting_list = msg->next;
          if (msg->next)
          {
               msg->next->prev = NULL;
          }
     }
     else if (msg->next == NULL)
     {
          /* last structure, but not first */
          msg->prev->next = NULL;
     }
     else
     {
          /* somewhere in the middle */
          msg->prev->next = msg->next;
          msg->next->prev = msg->prev;
     }
     
     /* now, the message structure is no longer in the collection */
     cc = yamiMutexUnlock(pa->msgwaiting_mutex);
     if (cc != YAMIOK)
          return cc;
     
     /* check if the message was already sent */
     cc = yamiMutexLock(msg->msg_mutex);
     if (cc != YAMIOK)
          return cc;
     
     if (--(msg->threadson) == 0)
     {
          /* this means, that no more threads are likely */
          /* to process this message, it's safe to       */
          /* destroy it                                  */
          cc = yamiMutexUnlock(msg->msg_mutex);

          if (cc != YAMIOK)
               return cc;

          _destroy_msgwaitingforreply(msg);
          FREE(msg)
          return YAMIOK;
     }

     cc = yamiMutexUnlock(msg->msg_mutex);
     return cc;
}

/* stuff for incoming messages */

/* ask the agent for any incoming messages      */
/* if wait is non-zero, this function can block */
int yamiAgentIncomingMsgGetNext(HYAMIAGENT agent,
          const char *objname, HINCMSG *pinc, int wait)
{
     struct yamiagent *pa;
     struct object *po;
     size_t parsetsize;
     int cc;
     void *praw, *praw2;
     struct msgincoming *pincoming;
     int skipqueue;
     
     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     pa = (struct yamiagent*)agent;
     
     if (pa->crashed == 1)
          return YAMIINTERNAL;
     
#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: retrieving a new message\n");
     fflush(VERBSTREAM);
#endif

     cc = yamiMutexLock(pa->objects_mutex);
     if (cc != YAMIOK)
          return cc;
     
     po = _find_object(pa, objname, 1);
     if (po == NULL)
     {
          /* no such object registered */
          cc = yamiMutexUnlock(pa->objects_mutex);
          if (cc != YAMIOK)
               return cc;

          return YAMINOTFOUND;
     }
     
     if (po->type != polling)
     {
          /* bad object type - it's an error to poll for        */
          /* a message when the object is registered as passive */
          cc = yamiMutexUnlock(pa->objects_mutex);
          if (cc != YAMIOK)
               return cc;

          return YAMIBADTYPE;
     }

     cc = yamiMutexLock(po->local_mutex);
     if (cc != YAMIOK)
     {
          yamiMutexUnlock(pa->objects_mutex);
          return cc;
     }
     
     ++(po->threadson);
     
     cc = yamiMutexUnlock(po->local_mutex);
     if (cc != YAMIOK)
     {
          yamiMutexUnlock(pa->objects_mutex);
          return cc;
     }
     
     /* unlock the collection */
     
     cc = yamiMutexUnlock(pa->objects_mutex);
     if (cc != YAMIOK)
          return cc;

     skipqueue = 0;
     if (wait == 0)     
     {
          /* non blocking poll */
          cc = yamiSemaphoreTryAcquire(po->incqueue_sem);
          if (cc == YAMINOTAVAILABLE)
          {
               *pinc = NULL;
               skipqueue = 1;
          }
          else if (cc != YAMIOK)
               return cc;
     }
     else
     {
          /* do blocking wait */
          cc = yamiSemaphoreAcquire(po->incqueue_sem);
          if (cc != YAMIOK)
               return cc;
     }

     /* there is something in the object's incoming queue */

     if (skipqueue == 0)
     {
          cc = yamiQueuePopFront(po->incqueue, &praw, &praw2, &parsetsize);
          if (cc != YAMIOK)
               return cc;
     
          pincoming = (struct msgincoming*)praw;
          *pinc = pincoming;

#ifdef VERBOSE
          fprintf(VERBSTREAM, "a: new message has been retrieved\n");
          fflush(VERBSTREAM);
#endif
     }
     
     cc = yamiMutexLock(po->local_mutex);
     if (cc != YAMIOK)
          return cc;
     
     if (--(po->threadson) == 0 && po->tobedeleted == 1)
     {
          cc = yamiMutexUnlock(po->local_mutex);
          if (cc != YAMIOK)
               return cc;
          _destroy_object(po);
          FREE(po)
          return skipqueue ? YAMINOTAVAILABLE : YAMIOK;
     }
     
     cc = yamiMutexUnlock(po->local_mutex);
     if (cc != YAMIOK)
          return cc;

     return skipqueue ? YAMINOTAVAILABLE : YAMIOK;
}

/* get the name of the incoming message */
int yamiAgentIncomingMsgGetMsgName(HINCMSG inc, const char **pname)
{
     struct msgincoming *pincoming;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     
     *pname = pincoming->msgname;
     
     return YAMIOK;
}

/* get the name of the destination object of incoming message */
/* this can be helpful when retrieving messages in ANY object */
int yamiAgentIncomingMsgGetObjectName(HINCMSG inc, const char **pname)
{
     struct msgincoming *pincoming;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     
     *pname = pincoming->objname;
     
     return YAMIOK;
}

/* get the parameter set from the incomming message */
int yamiAgentIncomingMsgGetParameters(HINCMSG inc, HPARAMSET *pps)
{
     struct msgincoming *pincoming;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     
     if (pincoming->paramsetraw == NULL)
     {
          *pps = NULL;
          return YAMINOTAVAILABLE;
     }
     else
          /* here, the parameter set is constructed with the level complying */
          /* to the level of the agent - this can fail if the message was    */
          /* sent with the parameter set of level2 and the destination agent */
          /* is of level1                                                    */
          /* this situation can happen only if the local (destination)       */
          /* domain was improperly registered at the sender (with a higher   */
          /* level that the actual) or when the message was sent by a poorly */
          /* designed agent                                                  */
          return yamiCreatePSFromPacked(pps, pincoming->agent->ownlevelno,
               pincoming->paramsetraw, pincoming->swap);
}

/* get the level of the sender */
int yamiAgentIncomingMsgGetLevel(HINCMSG inc, int *levelno)
{
     struct msgincoming *pincoming;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     
     *levelno = pincoming->levelno;
     
     return YAMIOK;
}

/* get the hint parameter that was set during object registration */
int yamiAgentIncomingMsgGetHint(HINCMSG inc, void **phint)
{
     struct msgincoming *pincoming;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     
     *phint = pincoming->hint;
     
     return YAMIOK;
}

/* reject the message */
int yamiAgentIncomingMsgReject(HINCMSG inc)
{
     struct msgincoming *pincoming;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     
     if (pincoming->processed == 1)
          return YAMICANNOTPROCESSTWICE;

     _respond_with_unkobj_of_rej(pincoming->agent, pincoming->msgid,
          REJECT, pincoming->retipaddress, pincoming->retipport);

     pincoming->processed = 1;     
     return YAMIOK;
}

/* eat the message - just mark it as processed           */
/* so that no further action (reply, reject) is possible */
int yamiAgentIncomingMsgEat(HINCMSG inc)
{
     struct msgincoming *pincoming;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     
     if (pincoming->processed == 1)
          return YAMICANNOTPROCESSTWICE;

     pincoming->processed = 1;     
     return YAMIOK;
}

/* reply to the incoming message */
int yamiAgentIncomingMsgReply(HINCMSG inc, HPARAMSET paramset)
{
     struct yamiagent *pa;
     struct msgincoming* pincoming;
     int cc;
     
     /* paramset CAN be NULL, which means that the paramset is empty */
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     pa = pincoming->agent;

     if (pincoming->processed == 1)
          return YAMICANNOTPROCESSTWICE;

     if (pa->crashed == 1)
          return YAMIINTERNAL;

     cc = _reply(pincoming, paramset);
     if (cc == YAMIOK)
          pincoming->processed = 1;
     
     return cc;
}

/* forward the message to another object, without changes */
int yamiAgentIncomingMsgForward(HINCMSG inc, const char *domainname,
          const char *objectname, const char *messagename,
          HPARAMSET paramset)
{
     struct msgincoming *pincoming;
     struct yamiagent *pa;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     if (objectname == YAMI_ANY_OBJECT)
          return YAMIBADHANDLE;
     
     pincoming = (struct msgincoming*)inc;
     pa = pincoming->agent;
     
     if (pincoming->processed == 1)
          return YAMICANNOTPROCESSTWICE;

#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: forwarding a message\n");
     fflush(VERBSTREAM);
#endif

     cc = _sendforward_domain(pa, domainname, objectname, messagename,
          paramset, pincoming->msgid, pincoming->retipaddress,
          pincoming->retipport, pincoming->levelno, 1, NULL);
     if (cc == YAMIOK)
          pincoming->processed = 1;

     return cc;
}

/* forward the message to another object, without changes, */
/* to explicitly given address                             */
int yamiAgentIncomingMsgForwardAddr(HINCMSG inc,
          const char *domainaddr, int domainport, int domainlevel,
          const char *objectname, const char *messagename,
          HPARAMSET paramset)
{
     struct msgincoming *pincoming;
     struct yamiagent *pa;
     int domainipaddr;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     if (objectname == YAMI_ANY_OBJECT)
          return YAMIBADHANDLE;
     
     pincoming = (struct msgincoming*)inc;
     pa = pincoming->agent;
     
     if (pincoming->processed == 1)
          return YAMICANNOTPROCESSTWICE;

     /* resolve the string address into the IP address */
     cc = yamiNetResolveIPAddress(domainaddr, &domainipaddr);
     if (cc != YAMIOK)
          return cc;

#ifdef VERBOSE
     fprintf(VERBSTREAM, "a: forwarding a message\n");
     fflush(VERBSTREAM);
#endif

     cc = _sendforward_addr(pa, domainipaddr, domainport, domainlevel,
          (int)eNone,
          objectname, messagename,
          paramset, pincoming->msgid, pincoming->retipaddress,
          pincoming->retipport, pincoming->levelno, 1, NULL);
     if (cc == YAMIOK)
          pincoming->processed = 1;

     return cc;
}

/* destroy the incomming message token */
int yamiAgentDestroyIncomingMsg(HINCMSG inc)
{
     struct msgincoming *pincoming;
     int cc;
     
     if ((cc = _checkINCMSG(inc)) != YAMIOK)
          return cc;
     
     pincoming = (struct msgincoming*)inc;
     
     _destroy_msgincoming(pincoming);
     FREE(pincoming)
     
     return YAMIOK;
}

/* creates a minimalistic ad-hoc agent and sends   */
/* a one-way message to the given address          */
int yamiEasySend(const char *address, int port, int level,
          const char *objectname, const char *messagename,
          HPARAMSET paramset)
{
     struct yamipolicies policies;
     HYAMIAGENT agent;
     int cc;
     
     /* prepare the minimalistic configuration */
     
     yamiGetDefaultPolicies(&policies);
     policies.dispatchers = 0; /* no passive objects here        */
     policies.hassender   = 0; /* only synchro send              */
     policies.hassocket   = 0; /* only one-way message           */
     policies.hasreceiver = 0; /* only one-way message           */
     policies.haswaker    = 0; /* no timeouts here               */
     policies.mtuser      = 0; /* only one thread                */
     policies.allowduplex = 0; /* no incoming duplex connections */
     
     /* port does not matter, since there is no receiver                 */
     /* (if the remote object tries to respond, it will be surprised...) */
     
     cc = yamiCreateAgent(&agent, 0, &policies);
     if (cc != YAMIOK)
          return cc;
     
     /* send one-way message with explicit address */
     
     cc = yamiAgentMsgSendAddr(agent, address, port, level,
          objectname, messagename, paramset, NULL);
     if (cc != YAMIOK)
     {
          yamiDestroyAgent(agent);
          return cc;
     }
     
     /* that's all, folks! */
     
     return yamiDestroyAgent(agent);
}

/* prints in the provided buffer the local address */
/* in the xxx.xxx.xxx.xxx form                     */
int yamiAgentGetLocalAddress(HYAMIAGENT agent, char *pbuf)
{
     struct yamiagent *pa;
     const unsigned char *p;
     int cc;

     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     pa = (struct yamiagent*)agent;
     p = (const unsigned char*)&(pa->ownipaddress);
     
     sprintf(pbuf, "%d.%d.%d.%d", (int)p[0], (int)p[1], (int)p[2], (int)p[3]);

     return YAMIOK;
}

/* retrieves the actual port number                */
/* used by the listening socket                    */
int yamiAgentGetLocalPort(HYAMIAGENT agent, int* pport)
{
     struct yamiagent *pa;
     int cc;

     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;

     pa = (struct yamiagent*)agent;

     if (pa->listensock == NULL)
          return YAMINOSUCHSERVICE;
     
     *pport = pa->ownipport;
     
     return YAMIOK;
}

/* this is for the explicit event loop               */
/* normally, the responsibilities of this function   */
/* are taken by the receiver thread                  */
/* this function is for those who like experimenting */
/* and for platforms where no threading is possible  */
int yamiAgentProcessEvent(HYAMIAGENT agent, int timeout)
{
     struct yamiagent *pa;
     HSOCKET accepteds;
     int whatconn, ipaddr, ipport;
     int cc;

     if ((cc = _checkYAMIAGENT(agent)) != YAMIOK)
          return cc;
     
     pa = (struct yamiagent*)agent;
     
     if (pa->listensock == NULL || pa->receivepool == NULL)
          return YAMINOSUCHSERVICE;
          
     if (pa->receiver_th != NULL)
          return YAMINOSUCHSERVICE;

     cc = yamiNetPoolAccept(pa->listensock, pa->receivepool,
          pa->receivedplxpool, pa->senddplxpool, &accepteds,
          timeout, &whatconn, &ipaddr, &ipport);
     if (cc != YAMIOK)
     {
          return cc;
     }
     
     return _process_connection(pa, accepteds, whatconn, ipaddr, ipport);
}
