/*
 * yamidomains.c
 *
 * this is an implementation of the domain set
 * (agent's own address book)
 *
 * Copyright (C) 2001-2005 Maciej Sobczak
 * http://www.msobczak.com/prog/yami/
 * 
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 */

#include "yamierrors.h"
#include "yamilimits.h"
#include "yamidomains.h"
#include "yamioslayer.h"
#include "yamitypes.h"
#include <stdlib.h>
#include <string.h>


/* this is for tracking memory leaks */
#ifdef YAMI_TRACK_MEMORY

#include <stdio.h>

#define MALLOC(p, s) { \
     p = malloc(s); \
     fprintf(stderr, "%s(%d) malloc %p\n", __FILE__, __LINE__, p); \
}
#define FREE(a) { \
     fprintf(stderr, "%s(%d) free %p\n", __FILE__, __LINE__, a); \
     free(a); \
}

#else

#define MALLOC(p, s) { p = malloc(s); }
#define FREE(a) free(a);

#endif


struct domainnode
{
     char              *name;
     int               ipaddr;
     int               ipport;
     int               level;
     int               options;
     struct domainnode *next;
};

struct domains
{
     enum eYamiType    rtti;
     HMUTEX            mutex;
     struct domainnode *first;
};

static int _checkDOMAINS(HDOMAINS ds)
{
     if (ds == NULL)
          return YAMIBADHANDLE;
     if (((struct domains*)ds)->rtti != eDomainSet)
          return YAMIBADHANDLE;
     return YAMIOK;
}

int yamiCreateDomainSet(HDOMAINS *pds, int synchro)
{
     struct domains *p;
     int cc;
     
     MALLOC(p, sizeof(struct domains))
     
     if (p == NULL)
          return YAMINOMEMORY;

     p->rtti = eDomainSet;          
     p->first = NULL;
     
     if (synchro != 0)
     {
          cc = yamiCreateMutex(&(p->mutex));
          if (cc != YAMIOK)
          {
               FREE(p)
               return cc;
          }
     }
     else
     {
          p->mutex = NULL;
     }
     
     *pds = p;
     return YAMIOK;
}

int yamiDestroyDomainSet(HDOMAINS ds)
{
     struct domains *pd;
     struct domainnode *p1, *p2;
     int cc;

     if (ds == NULL)
          return YAMIOK;     
     if ((cc = _checkDOMAINS(ds)) != YAMIOK)
          return cc;
     
     pd = (struct domains*)ds;
     
     p1 = pd->first;
     
     while (p1)
     {
          FREE(p1->name)
          p2 = p1->next;
          FREE(p1)
          p1 = p2;
     }

     cc = yamiDestroyMutex(pd->mutex);
     FREE(ds)
     return cc;
}

int yamiDomainSetRegister(HDOMAINS ds, const char *name,
     const char *addr, int port, int level, int options)
{
     struct domains *pdomains;
     struct domainnode *pnode, *pnew, **ptochange;
     int ipaddr;
     int namelen;
     int cc;

     if ((cc = _checkDOMAINS(ds)) != YAMIOK)
          return cc;
          
     if ((namelen = (int)strlen(name)) > MAXNAMELEN)
          return YAMIBADLENGTH;
     
     pdomains = (struct domains*)ds;
          
     if ((cc = yamiNetResolveIPAddress(addr, &ipaddr)) != 0)
          return cc;
     
     cc = yamiMutexLock(pdomains->mutex);
     if (cc != YAMIOK)
          return cc;
     
     pnode = pdomains->first;
     ptochange = &(pdomains->first);
     while (pnode)
     {
          if (strcmp(pnode->name, name) == 0)
          {
               /* change in-place */
               pnode->ipaddr = ipaddr;
               pnode->ipport = port;
               pnode->level = level;
               pnode->options = options;
               
               cc = yamiMutexUnlock(pdomains->mutex);
               return cc;
          }

          ptochange = &(pnode->next);
          pnode = pnode->next;
     }
     
     MALLOC(pnew, sizeof(struct domainnode))
     
     if (pnew == NULL)
     {
          cc = yamiMutexUnlock(pdomains->mutex);
          if (cc != YAMIOK)
               return cc;
          return YAMINOMEMORY;
     }
     
     MALLOC(pnew->name, (size_t)(namelen + 1))
     
     if (pnew->name == NULL)
     {
          FREE(pnew)
          cc = yamiMutexUnlock(pdomains->mutex);
          if (cc != YAMIOK)
               return cc;
          return YAMINOMEMORY;
     }
     strcpy(pnew->name, name);
     pnew->ipaddr = ipaddr;
     pnew->ipport = port;
     pnew->level = level;
     pnew->options = options;
     pnew->next = NULL;
     
     *ptochange = pnew;

     cc = yamiMutexUnlock(pdomains->mutex);
     return cc;
}

int yamiDomainSetUnregister(HDOMAINS ds, const char *name)
{
     struct domains *pdomains;
     struct domainnode *pnode, **pprev;
     int cc;
     
     if ((cc = _checkDOMAINS(ds)) != YAMIOK)
          return cc;
          
     pdomains = (struct domains*)ds;
     
     cc = yamiMutexLock(pdomains->mutex);
     if (cc != YAMIOK)
          return cc;
          
     pnode = pdomains->first;
     pprev = &(pdomains->first);
     
     while (pnode)
     {
          if (strcmp(pnode->name, name) == 0)
          {
               *pprev = pnode->next;
               FREE(pnode->name)
               FREE(pnode)
               
               cc = yamiMutexUnlock(pdomains->mutex);
               return cc;
          }
          pprev = &(pnode->next);
          pnode = pnode->next;
     }

     cc = yamiMutexUnlock(pdomains->mutex);
     if (cc != YAMIOK)
          return cc;          
     return YAMINOTFOUND;
}

int yamiDomainSetLookUp(HDOMAINS ds, const char *name,
     int *ipaddr, int *port, int *level, int *options)
{
     struct domains *pdomains;
     struct domainnode *pnode;
     int cc;
     
     if ((cc = _checkDOMAINS(ds)) != YAMIOK)
          return cc;
     
     pdomains = (struct domains*)ds;
     pnode = pdomains->first;

     cc = yamiMutexLock(pdomains->mutex);
     if (cc != YAMIOK)
          return cc;
     
     while (pnode)
     {
          if (strcmp(pnode->name, name) == 0)
          {
               *ipaddr = pnode->ipaddr;
               *port = pnode->ipport;
               *level = pnode->level;
               *options = pnode->options;
               
               cc = yamiMutexUnlock(pdomains->mutex);
               return cc;
          }
          pnode = pnode->next;
     }
     
     cc = yamiMutexUnlock(pdomains->mutex);
     if (cc != YAMIOK)
          return cc;          
     return YAMINOTFOUND;
}
