/*
 * yamiparams.c
 *
 * this is an implementation of the YAMI parameter set
 *
 * Copyright (C) 2001-2005 Maciej Sobczak
 * http://www.msobczak.com/prog/yami/
 *
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 */

#include "yamiparams.h"
#include "yamierrors.h"
#include "yamilimits.h"
#include "yamitypes.h"
#include "yamiXDR.h"
#include <string.h>
#include <stdlib.h>
#include <wchar.h>


/* this is for tracking memory leaks */
#ifdef YAMI_TRACK_MEMORY

#include <stdio.h>

#define MALLOC(p, s) { \
     p = malloc(s); \
     fprintf(stderr, "%s(%d) malloc %p\n", __FILE__, __LINE__, p); \
}
#define FREE(a) { \
     fprintf(stderr, "%s(%d) free %p\n", __FILE__, __LINE__, a); \
     free(a); \
}

#else

#define MALLOC(p, s) { p = malloc(s); }
#define FREE(a) free(a);

#endif


/*
 * HPARAMSET definition
 */

struct PARAMSETentry
{
     enum paramType partype;
     
     /* size of this field after packing (with the type field) */
     int            parsize;

     /* this is a variant discriminated by the partype field above */
     union
     {     
          int       ivalue;
          double    dvalue;
          char      cvalue;

          struct
          {
               /* size of the binary data stored          */
               /* this includes terminating 0 for strings */
               int     parlength;
               
               /* the flag if the entry is an owner of the buffer */
               int     owner;
               
               void    *pointer;
          } binvalue;
     } variant;
};

struct PARAMSETimpl
{
     enum eYamiType       rtti;
     int                  paramcount;
     int                  totalsize;     /* after packing */
     struct PARAMSETentry entries[1];
};

/* simple, but effective RTTI check */
static int _checkPARAMSET(HPARAMSET ps)
{
     if (ps == NULL)
          return YAMIBADHANDLE;
     if (((struct PARAMSETimpl*)ps)->rtti != eParamSet)
          return YAMIBADHANDLE;
     return YAMIOK;
}

int yamiCreateParamSet(HPARAMSET *pps, int parcount)
{
     struct PARAMSETimpl *pimpl;
     size_t          paramsetsize;
     int          i;
     
     if (parcount <= 0 || parcount > MAXPARAMS)
          return YAMIOUTOFRANGE;
     
     paramsetsize = sizeof(struct PARAMSETimpl)
          + (parcount - 1) * sizeof(struct PARAMSETentry);
          
     MALLOC(pimpl, paramsetsize)
     
     if (pimpl == NULL)
          return YAMINOMEMORY;

     memset(pimpl, 0, paramsetsize);
     
     pimpl->rtti = eParamSet;
     pimpl->paramcount = parcount;
     
     /* number of params is always included */
     /* no matter how many parameters       */
     pimpl->totalsize = (int)sizeof(int);
     
     for (i = 0; i < parcount; ++i)
     {
          pimpl->entries[i].partype = eNoType;
          pimpl->entries[i].parsize = 0;
     }
     
     *pps = pimpl;
     return YAMIOK;
}

static int _yamiCopyEntry(struct PARAMSETentry *pentrydst,
     struct PARAMSETentry *pentrysrc)
{
     enum paramType partype;
     size_t parlength;
     int owner;
     void *pointer;

     partype = pentrydst->partype = pentrysrc->partype;
     pentrydst->parsize = pentrysrc->parsize;
     switch (partype)
     {
     case eNoType:
          break;
     case eString:
     case eWString:
     case eBinary:
          parlength = (size_t)(pentrydst->variant.binvalue.parlength =
               pentrysrc->variant.binvalue.parlength);
          owner = pentrydst->variant.binvalue.owner =
               pentrysrc->variant.binvalue.owner;
          if (owner != 0)
          {
               MALLOC(pointer, parlength)
               
               if (pointer == NULL)
                    return YAMINOMEMORY;
               memcpy(pointer, pentrysrc->variant.binvalue.pointer,
                    parlength);
               pentrydst->variant.binvalue.pointer = pointer;
          }
          else
               pentrydst->variant.binvalue.pointer =
                    pentrysrc->variant.binvalue.pointer;
          break;
     case eInt:
          pentrydst->variant.ivalue = pentrysrc->variant.ivalue;
          break;
     case eDouble:
          pentrydst->variant.dvalue = pentrysrc->variant.dvalue;
          break;
     case eByte:
          pentrydst->variant.cvalue = pentrysrc->variant.cvalue;
          break;
     }
     
     return YAMIOK;
}

static
void _yamiFreeEntry(struct PARAMSETimpl *pimpl, struct PARAMSETentry *pentry)
{
     enum paramType partype;
     void *pointer;
     
     if ((partype = pentry->partype) == eNoType)
          return;
     
     switch (partype)
     {
     case eString:
     case eWString:
     case eBinary:
          if (pentry->variant.binvalue.owner != 0)
               if ((pointer = pentry->variant.binvalue.pointer) != NULL)
                    FREE(pointer)
          break;
     case eInt:
     case eDouble:
     case eByte:
     case eNoType: /* to shut up compiler's paranoic warnings */
          break;
     }
     
     pimpl->totalsize -= pentry->parsize;
     pentry->parsize = 0;
     pentry->partype = eNoType;
}

int yamiCopyParamSet(HPARAMSET *ppsdst, HPARAMSET pssrc)
{
     struct PARAMSETimpl *pimpldst, *pimplsrc;
     struct PARAMSETentry *pentrydst, *pentrysrc;
     int parcount;
     size_t paramsetsize;
     int cc, i;
     
     if ((cc = _checkPARAMSET(pssrc)) != YAMIOK)
          return cc;     
     
     pimplsrc = (struct PARAMSETimpl*)pssrc;
     
     parcount = pimplsrc->paramcount;
     paramsetsize = sizeof(struct PARAMSETimpl)
          + (parcount - 1) * sizeof(struct PARAMSETentry);
          
     MALLOC(pimpldst, paramsetsize)
     
     if (pimpldst == NULL)
          return YAMINOMEMORY;

     memset(pimpldst, 0, paramsetsize);
     
     pimpldst->rtti = eParamSet;
     pimpldst->paramcount = parcount;
     
     pimpldst->totalsize = pimplsrc->totalsize;
     
     for (i = 0; i < parcount; ++i)
     {
          pentrysrc = &(pimplsrc->entries[i]);
          pentrydst = &(pimpldst->entries[i]);
          
          cc = _yamiCopyEntry(pentrydst, pentrysrc);
          if (cc != YAMIOK)
          {
               while (--i >= 0)
               {
                    _yamiFreeEntry(pimpldst, &(pimpldst->entries[i]));
               }
               
               FREE(pimpldst)
               return cc;
          }
     }
     
     *ppsdst = pimpldst;
     return YAMIOK;
}

static 
int _yamiCheckArgs(HPARAMSET paramset, int position,
     struct PARAMSETimpl **ppimpl)
{
     struct PARAMSETimpl *pimpl;
     int cc;

     if ((cc = _checkPARAMSET(paramset)) != YAMIOK)
          return cc;     
     
     pimpl = (struct PARAMSETimpl*)paramset;
     
     if (position < 0 || position >= pimpl->paramcount)
          return YAMIOUTOFRANGE;

     *ppimpl = pimpl;
     return YAMIOK;
}

int yamiCopyParameter(HPARAMSET psdst, int positiondst,
     HPARAMSET pssrc, int positionsrc)
{
     struct PARAMSETimpl *pimpldst, *pimplsrc;
     struct PARAMSETentry *pentrydst, *pentrysrc;
     int cc;
     
     cc = _yamiCheckArgs(psdst, positiondst, &pimpldst);
     if (cc != YAMIOK)
          return cc;
     cc = _yamiCheckArgs(pssrc, positionsrc, &pimplsrc);
     if (cc != YAMIOK)
          return cc;
     
     pentrydst = &(pimpldst->entries[positiondst]);
     pentrysrc = &(pimplsrc->entries[positionsrc]);
     
     _yamiFreeEntry(pimpldst, pentrydst);
     
     cc = _yamiCopyEntry(pentrydst, pentrysrc);
     if (cc != YAMIOK)
          return cc; /* I know, the destination entry has been freed... */
     
     pimpldst->totalsize += pentrydst->parsize;

     return YAMIOK;
}

static int _yamiIsParamString(struct PARAMSETimpl *pimpl, int position)
{
     struct PARAMSETentry *pentry = &(pimpl->entries[position]);
     wchar_t *pwc;
     int i, length;
     
     if (pentry->partype == eString)
          return 1;
     if (pentry->partype != eWString)
          return 0;
     
     pwc = (wchar_t*)pentry->variant.binvalue.pointer;
     length = pentry->variant.binvalue.parlength / sizeof(wchar_t) - 1;
     for (i = 0; i < length; ++i)
          if (((unsigned int)(pwc[i]) & 0xffffff00) != 0)
               return 0;
     
     return 1;
}

int yamiIsASCIIString(HPARAMSET paramset, int position, int *isascii)
{
     struct PARAMSETimpl *pimpl;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     *isascii = _yamiIsParamString(pimpl, position);

     return YAMIOK;
}

int yamiSetInt(HPARAMSET paramset, int position, int ivalue)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;

     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;

     pentry = &(pimpl->entries[position]);
     _yamiFreeEntry(pimpl, pentry);
     
     /* int adds 2 * sizeof(int) to the total size */
     if (pimpl->totalsize + 2 * sizeof(int) > MAXPARAMSETSIZE)
          return YAMIBADLENGTH;
     
     pentry->partype = eInt;
     pentry->parsize = 2 * sizeof(int);
     pentry->variant.ivalue = ivalue;

     pimpl->totalsize += (2 * sizeof(int));
     
     return YAMIOK;
}

int yamiSetByte(HPARAMSET paramset, int position, char cvalue)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     pentry = &(pimpl->entries[position]);
     _yamiFreeEntry(pimpl, pentry);

     /* byte adds 2 * sizeof(int) to the total size */
     if (pimpl->totalsize + 2 * sizeof(int) > MAXPARAMSETSIZE)
          return YAMIBADLENGTH;
     
     pentry->partype = eByte;
     pentry->parsize = 2 * sizeof(int);
     pentry->variant.cvalue = cvalue;

     pimpl->totalsize += (2 * sizeof(int));
     
     return YAMIOK;
}

int yamiSetDouble(HPARAMSET paramset, int position, double dvalue)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     pentry = &(pimpl->entries[position]);
     _yamiFreeEntry(pimpl, pentry);
     
     /* double adds sizeof(int) + sizeof(double) to the total size */
     if (pimpl->totalsize + sizeof(int) + sizeof(double) > MAXPARAMSETSIZE)
          return YAMIBADLENGTH;
     
     pentry->partype = eDouble;
     pentry->parsize = (int)(sizeof(int) + sizeof(double));
     pentry->variant.dvalue = dvalue;

     pimpl->totalsize += sizeof(int) + sizeof(double);
     
     return YAMIOK;
}

static int _yamiSetBinary(struct PARAMSETimpl *pimpl, int position,
     const char *pbuf, int bufsize, int owner)
{
     struct PARAMSETentry *pentry;
     char *praw;
     
     pentry = &(pimpl->entries[position]);
     _yamiFreeEntry(pimpl, pentry);

     if (bufsize > 0)
     {
          if (owner != 0)
          {
               MALLOC(praw, (size_t)bufsize)
               
               if (praw == NULL)
                    return YAMINOMEMORY;

               memcpy(praw, pbuf, (size_t)bufsize);
          }
          else
               praw = (char*)pbuf;
     }
     else
          praw = NULL;
     
     pentry->variant.binvalue.pointer = praw;
     pentry->variant.binvalue.parlength = bufsize;
     pentry->variant.binvalue.owner = owner;
     
     return YAMIOK;
}

static int _yamiSetBinaryOwn(HPARAMSET paramset, int position,
     const char *pbuf, int bufsize, int owner)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;

     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;

     if (bufsize <= 0 || bufsize > MAXPARAMSIZE)
          return YAMIBADLENGTH;
     
     /* binary adds 2 * sizeof(int) + bufsize to the total size */
     if (pimpl->totalsize + 2 * sizeof(int) + bufsize > MAXPARAMSETSIZE)
          return YAMIBADLENGTH;
     
     cc = _yamiSetBinary(paramset, position, pbuf, bufsize, owner);
     if (cc == YAMIOK)
     {
          pentry = &(pimpl->entries[position]);
          pentry->partype = eBinary;
          pentry->parsize =
               2 * sizeof(int) + yamiXDRRoundTo4(bufsize);
          pimpl->totalsize += pentry->parsize;
     }
     
     return YAMIOK;
}

int yamiSetBinary(HPARAMSET paramset, int position,
     const char *pbuf, int bufsize)
{
     return _yamiSetBinaryOwn(paramset, position, pbuf, bufsize, 1);
}

int yamiSetBinaryShallow(HPARAMSET paramset, int position,
     const char *pbuf, int bufsize)
{
     return _yamiSetBinaryOwn(paramset, position, pbuf, bufsize, 0);
}

static int _yamiSetStringOwn(HPARAMSET paramset, int position,
     const char *pstr, int owner)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;
     int length = (int)strlen(pstr);
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     if (length > MAXPARAMSIZE)
          return YAMIBADLENGTH;

     /* string adds 2 * sizeof(int) + length to the total size */
     if (pimpl->totalsize + 2 * sizeof(int) + length > MAXPARAMSETSIZE)
          return YAMIBADLENGTH;
     
     /* store also the terminating (char)0 */
     cc = _yamiSetBinary(paramset, position,
          (const char*)pstr, length + 1, owner);
     if (cc == YAMIOK)
     {
          pentry = &(pimpl->entries[position]);
          pentry->partype = eString;
          pentry->parsize =
               2 * sizeof(int) + yamiXDRRoundTo4(length);
          pimpl->totalsize += pentry->parsize;
     }

     return cc;
}

int yamiSetString(HPARAMSET paramset, int position, const char *pstr)
{
     return _yamiSetStringOwn(paramset, position, pstr, 1);
}

int yamiSetStringShallow(HPARAMSET paramset, int position, const char *pstr)
{
     return _yamiSetStringOwn(paramset, position, pstr, 0);
}

static int _yamiSetWStringOwn(HPARAMSET paramset, int position,
     const wchar_t *pwstr, int owner)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;
     int length = (int)wcslen(pwstr);
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     if (length * sizeof(int) > MAXPARAMSIZE)
          return YAMIBADLENGTH;

     /* wide string adds 2 * sizeof(int) + sizeof(int) * length */
     /* to the total size                                       */
     if (pimpl->totalsize + 2 * sizeof(int)
          + sizeof(int) * length > MAXPARAMSETSIZE)
          return YAMIBADLENGTH;

     /* store also the terminating (wchar_t)0 */          
     cc = _yamiSetBinary(paramset, position,     (const char*)pwstr,
          (length + 1) * sizeof(wchar_t), owner);
     if (cc == YAMIOK)
     {
          pentry = &(pimpl->entries[position]);
          pentry->partype = eWString;
          pentry->parsize =
               2 * sizeof(int) + sizeof(int) * length;
          pimpl->totalsize += pentry->parsize;
     }
     
     return cc;
}

int yamiSetWString(HPARAMSET paramset, int position, const wchar_t *pwstr)
{
     return _yamiSetWStringOwn(paramset, position, pwstr, 1);
}

int yamiSetWStringShallow(HPARAMSET paramset, int position,
     const wchar_t *pwstr)
{
     return _yamiSetWStringOwn(paramset, position, pwstr, 0);
}

int yamiGetType(HPARAMSET paramset, int position, enum paramType *ptype)
{
     struct PARAMSETimpl *pimpl;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     *ptype = pimpl->entries[position].partype;
     
     return YAMIOK;
}

int yamiGetInt(HPARAMSET paramset, int position, int *pivalue)
{
     struct PARAMSETimpl *pimpl;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     if (pimpl->entries[position].partype != eInt)
          return YAMIBADTYPE;
     
     *pivalue = pimpl->entries[position].variant.ivalue;
     
     return YAMIOK;
}

int yamiGetByte(HPARAMSET paramset, int position, char *pcvalue)
{
     struct PARAMSETimpl *pimpl;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     if (pimpl->entries[position].partype != eByte)
          return YAMIBADTYPE;
     
     *pcvalue = pimpl->entries[position].variant.cvalue;

     return YAMIOK;
}

int yamiGetDouble(HPARAMSET paramset, int position, double *pdvalue)
{
     struct PARAMSETimpl *pimpl;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     if (pimpl->entries[position].partype != eDouble)
          return YAMIBADTYPE;
     
     *pdvalue = pimpl->entries[position].variant.dvalue;

     return YAMIOK;
}

int yamiGetBinarySize(HPARAMSET paramset, int position, int *psize)
{
     struct PARAMSETimpl *pimpl;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     if (pimpl->entries[position].partype != eBinary)
          return YAMIBADTYPE;
     
     *psize = pimpl->entries[position].variant.binvalue.parlength;
     
     return YAMIOK;
}

int yamiGetBinaryValue(HPARAMSET paramset, int position, char *pbuf)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     pentry = &(pimpl->entries[position]);
     if (pentry->partype != eBinary)
          return YAMIBADTYPE;

     if (pentry->variant.binvalue.parlength > 0)
          memcpy(pbuf, pentry->variant.binvalue.pointer,
               (size_t)pentry->variant.binvalue.parlength);
     
     return YAMIOK;
}

int yamiGetBinaryBuffer(HPARAMSET paramset, int position, const char **ppbuf)
{
     struct PARAMSETimpl *pimpl;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     if (pimpl->entries[position].partype != eBinary)
          return YAMIBADTYPE;
     
     *ppbuf = (const char*)pimpl->entries[position].variant.binvalue.pointer;
     
     return YAMIOK;
}

int yamiGetStringLength(HPARAMSET paramset, int position, int *plength)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;
     
     if (_yamiIsParamString(pimpl, position) == 0)
          return YAMIBADTYPE;

     pentry = &(pimpl->entries[position]);
     if (pentry->partype == eString)
          *plength = pentry->variant.binvalue.parlength - 1;
     else
          *plength = pentry->variant.binvalue.parlength / sizeof(wchar_t) - 1;
     
     return YAMIOK;
}

int yamiGetStringValue(HPARAMSET paramset, int position, char *pbuf)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc, i;
     int length;
     wchar_t *pwc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;

     if (_yamiIsParamString(pimpl, position) == 0)
          return YAMIBADTYPE;

     pentry = &(pimpl->entries[position]);
     length = pentry->variant.binvalue.parlength;     /* including 0 */
     if (length > 0)
     {
          if (pentry->partype == eString)
          {
               memcpy(pbuf, pentry->variant.binvalue.pointer, (size_t)length);
          }
          else
          {
               length = length / sizeof(wchar_t);
               pwc = (wchar_t*)(pentry->variant.binvalue.pointer);
               for (i = 0; i < length; ++i)
                    pbuf[i] = (char)((int)(pwc[i]) & 0xff);
          }
     }
     
     return YAMIOK;
}

int yamiGetStringBuffer(HPARAMSET paramset, int position, const char **ppbuf)
{
     struct PARAMSETimpl *pimpl;
     int cc;

     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;

     if (pimpl->entries[position].partype != eString)
          return YAMIBADTYPE;
     
     *ppbuf = (const char*)pimpl->entries[position].variant.binvalue.pointer;
     
     return YAMIOK;
}

int yamiGetWStringLength(HPARAMSET paramset, int position, int *plength)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;

     pentry = &(pimpl->entries[position]);
     if (pentry->partype != eString && pentry->partype != eWString)
          return YAMIBADTYPE;

     if (pentry->partype == eWString)
          *plength = pentry->variant.binvalue.parlength / sizeof(wchar_t) - 1;
     else
          *plength = pentry->variant.binvalue.parlength - 1;
     
     return YAMIOK;
}

int yamiGetWStringValue(HPARAMSET paramset, int position, wchar_t *pbuf)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int cc, i;
     int length;
     char *pc;
     
     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;

     pentry = &(pimpl->entries[position]);
     if (pentry->partype != eWString && pentry->partype != eString)
          return YAMIBADTYPE;

     length = pentry->variant.binvalue.parlength; /* including 0 */
     if (length > 0)
     {
          if (pentry->partype == eWString)
          {
               memcpy(pbuf, pentry->variant.binvalue.pointer, (size_t)length);
          }
          else
          {
               pc = (char*)pentry->variant.binvalue.pointer;
               for (i = 0; i < length; ++i)
               {
                    pbuf[i] = (wchar_t)(pc[i]);
               }
          }
     }
     
     return YAMIOK;
}

int yamiGetWStringBuffer(HPARAMSET paramset, int position,
     const wchar_t **ppbuf)
{
     struct PARAMSETimpl *pimpl;
     int cc;

     cc = _yamiCheckArgs(paramset, position, &pimpl);
     if (cc != YAMIOK)
          return cc;

     if (pimpl->entries[position].partype != eWString)
          return YAMIBADTYPE;
     
     *ppbuf =
          (const wchar_t*)pimpl->entries[position].variant.binvalue.pointer;
     
     return YAMIOK;
}

int yamiDestroyParamSet(HPARAMSET paramset)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     void *p;
     enum paramType type;
     int cc, i;

     if ((cc = _checkPARAMSET(paramset)) != YAMIOK)
          return cc;

     pimpl = (struct PARAMSETimpl*)paramset;

     for (i = 0; i < pimpl->paramcount; ++i)
     {
          pentry = &(pimpl->entries[i]);
          type = pentry->partype;
          if (type == eString || type == eWString || type == eBinary)
               if (pentry->variant.binvalue.owner != 0)
                    if ((p = pentry->variant.binvalue.pointer) != NULL)
                         FREE(p)
     }

     FREE(pimpl)
     
     return YAMIOK;
}

int yamiGetParamCount(HPARAMSET paramset, int *pparcount)
{
     struct PARAMSETimpl *pimpl;
     int cc;

     if ((cc = _checkPARAMSET(paramset)) != YAMIOK)
          return cc;

     pimpl = (struct PARAMSETimpl*)paramset;

     *pparcount = pimpl->paramcount;
     
     return YAMIOK;
}

int yamiGetPackedSize(HPARAMSET paramset, int *packedsize)
{
     struct PARAMSETimpl *pimpl;
     int cc, i;

     if ((cc = _checkPARAMSET(paramset)) != YAMIOK)
          return cc;

     pimpl = (struct PARAMSETimpl*)paramset;
     
     for (i = 0; i < pimpl->paramcount; ++i)
     {
          if (pimpl->entries[i].partype == eNoType)
               return YAMINOTINITIALIZED;
     }
     
     *packedsize = pimpl->totalsize;
     
     return YAMIOK;
}

static char * _yamiPackInt(char *pbuf, struct PARAMSETentry *pentry)
{
     return yamiXDRWriteInt(pbuf, pentry->variant.ivalue);
}

static char * _yamiPackString(char *pbuf, struct PARAMSETentry *pentry)
{
     char *pb = pbuf;
     int size = pentry->variant.binvalue.parlength - 1; /* do not store 0 */

     /* length of the string first */
     pb = yamiXDRWriteInt(pb, size);

     /* then, the string's data */
     memcpy(pb, pentry->variant.binvalue.pointer, (size_t)size);

     /* type field was already stored */
     return pbuf + pentry->parsize - sizeof(int);
}

static char * _yamiPackWString(char *pbuf, struct PARAMSETentry *pentry)
{
     wchar_t *pwc1 = (wchar_t*)pentry->variant.binvalue.pointer;
     char *pc2;
     int i, length;

     /* length of the wstring first, without 0 */
     length = pentry->variant.binvalue.parlength / sizeof(wchar_t) - 1;
     pc2 = yamiXDRWriteInt(pbuf, length);

     /* then, the data */
     for (i = 0; i < length; ++i)
     {
          pc2 = yamiXDRWriteInt(pc2, (int)*pwc1);
          ++pwc1;
     }
     
     /* type field was already stored */
     return pbuf + pentry->parsize - sizeof(int);
}

static char * _yamiPackByte(char *pbuf, struct PARAMSETentry *pentry)
{
     *(char*)pbuf = pentry->variant.cvalue;
     /* we force 4 for alignment */
     return pbuf + sizeof(int);
}

static char * _yamiPackDouble(char *pbuf, struct PARAMSETentry *pentry)
{
     return yamiXDRWriteDouble(pbuf, pentry->variant.dvalue);
}

static char * _yamiPackBinary(char *pbuf, struct PARAMSETentry *pentry)
{
     char *pb = pbuf;
     int size = pentry->variant.binvalue.parlength;
     
     /* size first */
     pb = yamiXDRWriteInt(pb, size);
     
     /* then, the data */
     memcpy(pb, pentry->variant.binvalue.pointer, (size_t)size);

     /* type field was already stored */
     return pbuf + pentry->parsize - sizeof(int);
}

int yamiGetPacked(HPARAMSET paramset, char *pbuf)
{
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int parcount, cc, i;

     if ((cc = _checkPARAMSET(paramset)) != YAMIOK)
          return cc;

     pimpl = (struct PARAMSETimpl*)paramset;

     parcount = pimpl->paramcount;

     /* number of params goes first */
     pbuf = yamiXDRWriteInt(pbuf, parcount);
     
     pentry = pimpl->entries;
     for (i = 0; i < parcount; ++i)
     {
          /* param's type */
          pbuf = yamiXDRWriteInt(pbuf, (int)pentry->partype);

          /* then, pack the param itself */          
          switch (pentry->partype)
          {
          case eString:
               pbuf = _yamiPackString(pbuf, pentry);
               break;
          case eWString:
               pbuf = _yamiPackWString(pbuf, pentry);
               break;
          case eInt:
               pbuf = _yamiPackInt(pbuf, pentry);
               break;
          case eDouble:
               pbuf = _yamiPackDouble(pbuf, pentry);
               break;
          case eByte:
               pbuf = _yamiPackByte(pbuf, pentry);
               break;
          case eBinary:
               pbuf = _yamiPackBinary(pbuf, pentry);
               break;
          default:
               return YAMINOTINITIALIZED;
          }
                    
          ++pentry;
     }

     return YAMIOK;
}

static
const char * _yamiUnpackString(const char *pbuf, struct PARAMSETimpl *pimpl,
               struct PARAMSETentry *pentry, int swap, int doit)
{
     int size;
     void *pointer;

     /* read the length */
     pbuf = yamiXDRReadInt(pbuf, (int*)&size, swap);

     if (size < 0 || size > MAXPARAMSIZE)
          return NULL;

     if (doit != 0)
     {
          pentry->partype = eString;
          pentry->parsize = 2 * sizeof(int) + yamiXDRRoundTo4(size);
          pentry->variant.binvalue.parlength = size + 1;

          /* store strings with terminating (char)0 */
          MALLOC(pointer, (size_t)(size + 1))
          
          if (pointer == NULL)
               return NULL;
               
          memcpy(pointer, pbuf, (size_t)size);
          ((char*)pointer)[size] = (char)0;

          pentry->variant.binvalue.pointer = pointer;
          pentry->variant.binvalue.owner = 1;

          pimpl->totalsize += pentry->parsize;
     }
     
     return pbuf + yamiXDRRoundTo4(size);
}

static
const char * _yamiUnpackWString(const char *pbuf, struct PARAMSETimpl *pimpl,
               struct PARAMSETentry *pentry, int swap, int doit)
{
     const char *pc1;
     wchar_t *pwc2;
     int size;
     int i, length, ival;

     /* read the length */
     pbuf = yamiXDRReadInt(pbuf, &length, swap);
     
     if (length <= 0 || length *sizeof(int) > MAXPARAMSIZE)
          return NULL;

     if (doit != 0)
     {
          pentry->partype = eWString;
          pentry->parsize = length * sizeof(int) + 2 * sizeof(int);
          
          size = length * sizeof(wchar_t);
          pentry->variant.binvalue.parlength = size + sizeof(wchar_t);
          
          pc1 = pbuf;
          
          /* store wstrings with terminating (wchar_t)0 */

          MALLOC(pwc2, size + sizeof(wchar_t))
          pentry->variant.binvalue.pointer = pwc2;

          if (pwc2 == NULL)
               return NULL;
          
          for (i = 0; i < length; ++i)
          {
               pc1 = yamiXDRReadInt(pc1, &ival, swap);
               pwc2[i] = (wchar_t)ival;
          }
          pwc2[length] = (wchar_t)0;
          
          pentry->variant.binvalue.owner = 1;

          pimpl->totalsize += pentry->parsize;
     }
     
     return pbuf + length * sizeof(int);
}

static
const char * _yamiUnpackInt(const char *pbuf, struct PARAMSETimpl *pimpl,
               struct PARAMSETentry *pentry, int swap, int doit)
{
     if (doit != 0)
     {
          pentry->partype = eInt;
          pentry->parsize = 2 * sizeof(int);
          
          pimpl->totalsize += pentry->parsize;
     
          return yamiXDRReadInt(pbuf, &(pentry->variant.ivalue), swap);
     }
     else
          return pbuf + sizeof(int);
}

static
const char * _yamiUnpackDouble(const char *pbuf, struct PARAMSETimpl *pimpl,
               struct PARAMSETentry *pentry, int swap, int doit)
{
     if (doit != 0)
     {
          pentry->partype = eDouble;
          pentry->parsize = (int)(sizeof(int) + sizeof(double));
          
          pimpl->totalsize += pentry->parsize;

          return yamiXDRReadDouble(pbuf, &(pentry->variant.dvalue), swap);
     }
     else
          return pbuf + sizeof(double);
}

static
const char * _yamiUnpackByte(const char *pbuf, struct PARAMSETimpl *pimpl,
               struct PARAMSETentry *pentry, int doit)
{
     if (doit != 0)
     {
          pentry->partype = eByte;
          pentry->parsize = 2 * sizeof(int);
          
          pentry->variant.cvalue = *(char*)pbuf;

          pimpl->totalsize += pentry->parsize;
     }
     
     /* we force 4 for alignment */
     return pbuf + sizeof(int);
}

static
const char * _yamiUnpackBinary(const char *pbuf, struct PARAMSETimpl *pimpl,
               struct PARAMSETentry *pentry, int swap, int doit)
{
     int size;
     void *pointer;

     /* first, read the size */     
     pbuf = yamiXDRReadInt(pbuf, (int*)&size, swap);

     if (size <= 0 || size > MAXPARAMSIZE)
          return NULL;

     if (doit != 0)
     {
          pentry->partype = eBinary;
          pentry->parsize = 2 * sizeof(int) + yamiXDRRoundTo4(size);
          pentry->variant.binvalue.parlength = size;
          
          MALLOC(pointer, (size_t)size)
          
          if (pointer == NULL)
               return NULL;

          memcpy(pointer, pbuf, (size_t)size);
          pentry->variant.binvalue.pointer = pointer;
          pentry->variant.binvalue.owner = 1;
     
          pimpl->totalsize += pentry->parsize;
     }
     
     return pbuf + yamiXDRRoundTo4(size);
}

int yamiCreatePSFromPacked(HPARAMSET *pps, int levelno,
     const char *pbuf, int swap)
{
     HPARAMSET ps;
     struct PARAMSETimpl *pimpl;
     struct PARAMSETentry *pentry;
     int parcount, i, itype;
     enum paramType partype;
     int cc;
     
     pbuf = yamiXDRReadInt(pbuf, &parcount, swap);

     if (parcount <= 0 || parcount > MAXPARAMS)
          return YAMIPARSEERROR;

     cc = yamiCreateParamSet(&ps, parcount);
     if (cc != YAMIOK)
          return cc;
     
     pimpl = (struct PARAMSETimpl*)ps;
          
     pentry = pimpl->entries;
     
     for (i = 0; i < parcount; ++i)
     {
          pbuf = yamiXDRReadInt(pbuf, &itype, swap);
          partype = (enum paramType)itype;

          switch (partype)
          {
          case eString:
               pbuf = _yamiUnpackString(pbuf, pimpl, pentry, swap, 1);
               break;
          case eWString:
               pbuf = _yamiUnpackWString(pbuf, pimpl, pentry, swap, 1);
               break;
          case eInt:
               if (levelno >= 2)
                    pbuf = _yamiUnpackInt(pbuf, pimpl, pentry, swap, 1);
               else
                    pbuf = NULL;
               break;
          case eDouble:
               if (levelno >= 2)
                    pbuf = _yamiUnpackDouble(pbuf, pimpl, pentry, swap, 1);
               else
                    pbuf = NULL;
               break;
          case eByte:
               if (levelno >= 2)
                    pbuf = _yamiUnpackByte(pbuf, pimpl, pentry, 1);
               else
                    pbuf = NULL;
               break;
          case eBinary:
               if (levelno >= 2)
                    pbuf = _yamiUnpackBinary(pbuf, pimpl, pentry, swap, 1);
               else
                    pbuf = NULL;
               break;
          default:
               (void)yamiDestroyParamSet((HPARAMSET)pimpl);
               return YAMIPARSEERROR;
          }
          
          if (pbuf == NULL)
          {
               (void)yamiDestroyParamSet((HPARAMSET)pimpl);
               return YAMIPARSEERROR;
          }
          
          ++pentry;
     }
     
     *pps = pimpl;
     return YAMIOK;
}

/* returns YAMIOK or YAMIPARSEERROR */
int yamiCheckRawBuffer(int levelno, const char *pbuf, int swap)
{
     int parcount, i, itype;
     enum paramType partype;
     
     pbuf = yamiXDRReadInt(pbuf, &parcount, swap);

     if (parcount <= 0 || parcount > MAXPARAMS)
          return YAMIPARSEERROR;

     for (i = 0; i < parcount; ++i)
     {
          pbuf = yamiXDRReadInt(pbuf, &itype, swap);
          partype = (enum paramType)itype;

          switch (partype)
          {
          case eString:
               pbuf = _yamiUnpackString(pbuf, NULL, NULL, swap, 0);
               break;
          case eWString:
               pbuf = _yamiUnpackWString(pbuf, NULL, NULL, swap, 0);
               break;
          case eInt:
               if (levelno >= 2)
                    pbuf = _yamiUnpackInt(pbuf, NULL, NULL, swap, 0);
               else
                    pbuf = NULL;
               break;
          case eDouble:
               if (levelno >= 2)
                    pbuf = _yamiUnpackDouble(pbuf, NULL, NULL, swap, 0);
               else
                    pbuf = NULL;
               break;
          case eByte:
               if (levelno >= 2)
                    pbuf = _yamiUnpackByte(pbuf, NULL, NULL, 0);
               else
                    pbuf = NULL;
               break;
          case eBinary:
               if (levelno >= 2)
                    pbuf = _yamiUnpackBinary(pbuf, NULL, NULL, swap, 0);
               else
                    pbuf = NULL;
               break;
          default:
               return YAMIPARSEERROR;
          }
          
          if (pbuf == NULL)
               return YAMIPARSEERROR;
     }
     
     return YAMIOK;
}
