/*
 * yamiqueue.c
 *
 * this is an implementation of the queue used between threads
 * the queue is automatically synchronized, if requested
 *
 * Copyright (C) 2001-2005 Maciej Sobczak
 * http://www.msobczak.com/prog/yami/
 *
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 */

#include "yamierrors.h"
#include "yamiqueue.h"
#include "yamioslayer.h"
#include "yamitypes.h"
#include <stdlib.h>


/* this is for tracking memory leaks */
#ifdef YAMI_TRACK_MEMORY

#include <stdio.h>

#define MALLOC(p, s) { \
     p = malloc(s); \
     fprintf(stderr, "%s(%d) malloc %p\n", __FILE__, __LINE__, p); \
}
#define FREE(a) { \
     fprintf(stderr, "%s(%d) free %p\n", __FILE__, __LINE__, a); \
     free(a); \
}

#else

#define MALLOC(p, s) { p = malloc(s); }
#define FREE(a) free(a);

#endif


struct queuenode
{
     void             *data1;
     void             *data2;
     size_t           size;
     struct queuenode *next;
};

struct queue
{
     enum eYamiType   rtti;
     struct queuenode *first;
     struct queuenode *last;
     size_t           maxlength;
     size_t           length;
     size_t           maxsize;
     size_t           size;
     HMUTEX           queue_mutex;
};

static int _checkQUEUE(HQUEUE q)
{
     if (q == NULL)
          return YAMIBADHANDLE;
     if (((struct queue*)q)->rtti != eQueue)
          return YAMIBADHANDLE;
     return YAMIOK;
}

int yamiCreateQueue(HQUEUE *pq, size_t maxlength, size_t maxsize, int synchro)
{
     struct queue *p;
     int cc;

     MALLOC(p, sizeof(struct queue))
     
     if (p == NULL)
          return YAMINOMEMORY;
     
     p->rtti = eQueue;
     
     p->first = NULL;
     p->last = NULL;
     p->maxlength = maxlength;
     p->length = 0;
     p->maxsize = maxsize;
     p->size = 0;

     if (synchro != 0)
     {     
          cc = yamiCreateMutex(&(p->queue_mutex));
          if (cc != YAMIOK)
          {
               FREE(p)
               return cc;
          }
     }
     else
     {
          p->queue_mutex = NULL;
     }
     
     *pq = p;
     
     return YAMIOK;
}

int yamiDestroyQueue(HQUEUE q)
{
     struct queue *pqueue;
     struct queuenode *p1, *p2;
     int cc;

     if (q == NULL)
          return YAMIOK;     
     if ((cc = _checkQUEUE(q)) != YAMIOK)
          return cc;

     pqueue = (struct queue*)q;

     (void)yamiMutexLock(pqueue->queue_mutex);
     
     p1 = pqueue->first;
     while (p1)
     {
          p2 = p1->next;
          FREE(p1->data1)
          FREE(p1->data2);
          FREE(p1)
          p1 = p2;
     }
     
     (void)yamiMutexUnlock(pqueue->queue_mutex);
     (void)yamiDestroyMutex(pqueue->queue_mutex);

     FREE(q)
     
     return YAMIOK;
}

int yamiQueuePushBack(HQUEUE q, void *data1, void *data2, size_t size)
{
     struct queue *pqueue;
     struct queuenode *pnew;
     int cc;
     
     if ((cc = _checkQUEUE(q)) != YAMIOK)
          return cc;

     pqueue = (struct queue*)q;

     cc = yamiMutexLock(pqueue->queue_mutex);
     if (cc != YAMIOK)
          return cc;
     
     if (size != 0)
     {
          if ((pqueue->size + size > pqueue->maxsize) ||
               (pqueue->length == pqueue->maxlength))
          {
               cc = yamiMutexUnlock(pqueue->queue_mutex);
               if (cc != YAMIOK)
                    return YAMISYNCHROERROR;

               return YAMIOVERFLOW;
          }
     }

     MALLOC(pnew, sizeof(struct queuenode))
     
     if (pnew == NULL)
     {
          cc = yamiMutexUnlock(pqueue->queue_mutex);
          if (cc != YAMIOK)
               return YAMISYNCHROERROR;

          return YAMINOMEMORY;
     }
     
     pnew->data1 = data1;
     pnew->data2 = data2;
     pnew->size = size;
     pnew->next = NULL;

     if (pqueue->first == NULL)
     {
          pqueue->first = pnew;
          pqueue->last = pnew;
     }
     else
     {
          pqueue->last->next = pnew;
          pqueue->last = pnew;
     }
     
     ++(pqueue->length);
     pqueue->size += size;
     
     cc = yamiMutexUnlock(pqueue->queue_mutex);
     if (cc != YAMIOK)
          return cc;

     return YAMIOK;
}

int yamiQueuePopFront(HQUEUE q, void **data1, void **data2, size_t *size)
{
     struct queue *pqueue;
     struct queuenode *p;
     int cc;
     
     if ((cc = _checkQUEUE(q)) != YAMIOK)
          return cc;
     pqueue = (struct queue*)q;
     
     cc = yamiMutexLock(pqueue->queue_mutex);
     if (cc != YAMIOK)
          return cc;

     if (pqueue->first == NULL)
     {
          cc = yamiMutexUnlock(pqueue->queue_mutex);
          if (cc != YAMIOK)
               return cc;
          return YAMIQUEUEEMPTY;
     }
     
     p = pqueue->first;
     *data1 = p->data1;
     *data2 = p->data2;
     *size = p->size;
     
     --(pqueue->length);
     pqueue->size -= p->size;
     
     pqueue->first = p->next;
     if (pqueue->first == NULL)
          pqueue->last = NULL;
     
     FREE(p)
     
     cc = yamiMutexUnlock(pqueue->queue_mutex);
     if (cc != YAMIOK)
          return cc;

     return YAMIOK;
}

int yamiQueueIsEmpty(HQUEUE q)
{
     struct queue *pqueue;
     int retval;
     int cc;
     
     if ((cc = _checkQUEUE(q)) != YAMIOK)
          return cc;

     pqueue = (struct queue*)q;

     cc = yamiMutexLock(pqueue->queue_mutex);
     if (cc != YAMIOK)
          return cc;

     retval = (pqueue->first == NULL) ? 1 : 0;

     cc = yamiMutexUnlock(pqueue->queue_mutex);
     if (cc != YAMIOK)
          return cc;

     return retval;
}
