/*
 * yamic.h
 *
 * this is the main file that defines the agent's API
 *
 * Copyright (C) 2001-2005 Maciej Sobczak
 * http://www.msobczak.com/prog/yami/
 *
 * Permission to copy, use, modify, sell and distribute this software
 * is granted provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied
 * warranty, and with no claim as to its suitability for any purpose.
 *
 */

#ifndef YAMIC_INCLUDED
#define YAMIC_INCLUDED

/*#include "yamiparams.h"
*/

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifndef HPARAMSET_DEFINED
typedef void* HPARAMSET;
#define HPARAMSET_DEFINED
#endif

/* this is a handle to the agent        */
/* all the functions expect this handle */
typedef void* HYAMIAGENT;

/* this is a handle to the message token                */
/* objects use it between sending and receiving replies */
typedef void* HMESSAGE;

/* this is a handle to the incoming message token         */
/* objects use it between receiving messages and replying */
typedef void* HINCMSG;

/* this is a helper enum definition for use when agent level is required */
enum eAgentLevel
{
     eLevel1 = 1,
     eLevel2 = 2
};

/* this structure defines policies that can be used in the agent's creation */
struct yamipolicies
{
     int agentlevel;    /* the level of the local agent    - default 2       */
     int objqmaxlength; /* max len of the object's queue   - default 10      */
     int objqmaxsize;   /* max size of the object's queue  - default 1048576 */
     int dispatchers;   /* number of dispatching threads   - default 1       */
     int hassender;     /* if the sender thread exists     - default 1       */
     int sqmaxlength;   /* max len of the sending queue    - default 256     */
     int sqmaxsize;     /* max size of the sending queue   - default 1048576 */
     int connpoolsize;  /* size of the connection pool     - default 10      */
     int sendtries;     /* number of tries before reporting network error    */
                        /*                                 - default 5       */
     int hassocket;     /* if the listening socket exists  - default 1       */
     int reuseaddr;     /* flag - allows to reuse socket   - default 0       */
                        /* this flag can help overcome TIME_WAIT problem     */
     int hasreceiver;   /* if the receiver thread exists   - default 1       */
     int haswaker;      /* if the waker thread exists      - default 1       */
     int mtuser;        /* the the user is multithreaded   - default 1       */
     int allowduplex;   /* allow duplex connections from calling clients     */
                        /*                                 - default 1       */
     int maxdplxconns;  /* max number of duplex connections active at a time */
                        /*                                 - default 100     */
     int receiveridle;  /* the idle wait time for the receiver if there is   */
                        /* no socket to wait for or no incoming connection   */
                        /*                                 - default 500     */
};

/* this function fills the yamipolicies structure with the default values */
void yamiGetDefaultPolicies(struct yamipolicies *ppolicies);

/*
 * constructors/destructor
 */

/* create the agent working on the given port              */
/* and with the given policies (may be NULL - use default) */
int yamiCreateAgent(HYAMIAGENT *pagent, int port,
               struct yamipolicies *ppolicies);

/* destroy the agent */
int yamiDestroyAgent(HYAMIAGENT agent);

/*
 * methods for domain set operation
 */

/* register new domain */
int yamiAgentDomainRegister(HYAMIAGENT agent, const char *name,
     const char *addr, int port, int level);

/* connection options for yamiAgentDomainRegisterEx */
enum connectionOptions
{
     eNone = 0,       /* use transient, one-way conection */
     eFixedDuplex = 1 /* use fixed, two-way connection    */
                      /* between client and server        */
};

/* register new domain, allow connection options */
int yamiAgentDomainRegisterEx(HYAMIAGENT agent, const char *name,
     const char *addr, int port, int level, int options);

/* unregister a domain */
int yamiAgentDomainUnregister(HYAMIAGENT agent, const char *name);

/*
 * methods for objects operation
 */

/* this defines type of the object registered with the agent */
enum objectType
{
     polling,                 /* object polls for messages             */
     passive_singlethreaded,  /* object is up-called,                  */
                              /* but no more than one thread at a time */
     passive_multithreaded    /* object is up-called,                  */
                              /* possibly by many threads              */
};

/* type of the function that will be up-called for passive objects */
/* when there is a message for it                                  */
/* inc is the handler to the incoming message structure            */
/* every function from the yamiAgentIncomingMsgXXX family applies  */
/* (but this function should not destroy the structure -           */
/* this is done automatically when the function returns)           */
/* if the function has not processed the message, it will be       */
/* automatically rejected                                          */
/* (note: yamiAgentIncomingMsgEat can "process" the message)       */
typedef void (*DISPFUNCTION)(HINCMSG inc);

/* this is the generic name, that will match *any* other name */
/* it can be useful for the default object semantics          */
#define YAMI_ANY_OBJECT NULL

/* register new object                                       */
/* otype describes type of the object being registered       */
/* if it's polling object, then df parameter does not matter */
/* maxmsgqueuelen and maxmsgqueuesize set limits on the      */
/* queue of incoming messages                                */
/* df is a dispatch function that will be called             */
/* when new message for this object arrives                  */
/* the hint parameter will be stored in every incoming       */
/* message for this object                                   */
int yamiAgentObjectRegister(HYAMIAGENT agent, const char *name,
               enum objectType otype, DISPFUNCTION df,
               void *hint);

/* unregister object - this destroys all data kept for the objects, */
/* including pending messages                                       */
int yamiAgentObjectUnregister(HYAMIAGENT agent, const char *name);

/*
 * methods for message invocations
 */

/* status of the waiting message */
enum msgStatus
{
     ePosted,       /* the message was posted, but not yet sent     */
     ePending,      /* the message was sent                         */
     eReplied,      /* the reply has arrived                        */
     eRejected,     /* the message was rejected                     */
     eUnknownObj,   /* the message was sent to unknown object       */
     eOverflow,     /* the message was rejected due to overflow     */
     eNetError,     /* there was an error while sending the message */
     eTimedOut,     /* the timeout set for this message expired     */
     eRejectByAgent /* the message was rejected by the destination  */
                    /* agent (may happen while sending level2       */
                    /* message to level1 agent)                     */
};

/* this is the name that corresponds to the domain though which the     */
/* message is being sent                                                */
/* this means that the destination object is managed (if exists) by the */
/* same agent as the originating object                                 */
#define YAMI_THIS_DOMAIN NULL

/* send a new message                                      */
/* this function retrieves the handle to a message token,  */
/* which can be used for querying for results              */
/* if phm is NULL, the message token is not created,       */
/* so there will be no way to query for anything (one-way) */
int yamiAgentMsgSend(HYAMIAGENT agent, const char *domainname,
          const char *objectname, const char *messagename,
          HPARAMSET paramset, HMESSAGE *phm);

/* send a new message, but with explicit address */
int yamiAgentMsgSendAddr(HYAMIAGENT agent,
     const char *addr, int port, int level,
     const char *objectname, const char *messagename,
     HPARAMSET paramset, HMESSAGE *phm);

/* send a new message with explicit address and allow connection options */
/* meaning of the connection options as with yamiAgentDomainRegisterEx   */
int yamiAgentMsgSendAddrEx(HYAMIAGENT agent,
     const char *addr, int port, int level,
     const char *objectname, const char *messagename,
     HPARAMSET paramset, HMESSAGE *phm, int options);

/* set up a timeout alarm for a sent message */
/* the timeout is in seconds (not precise)   */
int yamiAgentSetUpTimeOut(HMESSAGE message, int timeout);
     
/* get the status of the message sent                     */
int yamiAgentMsgGetStatus(HMESSAGE hm, enum msgStatus *status);

/* this function helps clients wait for responses         */
int yamiAgentMsgWait(HMESSAGE hm);

/* get the response from the server */
int yamiAgentMsgGetResponse(HMESSAGE hm, HPARAMSET *pps);

/* destroy the sent message token                         */
int yamiAgentMsgDestroy(HMESSAGE hm);

/*
 * methods for operating on incoming messages
 */

/* get the new incoming message               */
/* it can block or not, depending on the wait */
int yamiAgentIncomingMsgGetNext(HYAMIAGENT agent,
          const char *objname, HINCMSG *pinc, int wait);

/* get the message name from the incoming token */
int yamiAgentIncomingMsgGetMsgName(HINCMSG inc, const char **pname);

/* get the object name from the incoming token */
int yamiAgentIncomingMsgGetObjectName(HINCMSG inc, const char **pname);

/* get the parameter set from the incoming message token */
int yamiAgentIncomingMsgGetParameters(HINCMSG inc, HPARAMSET *pps);

/* get the level of the sender */
int yamiAgentIncomingMsgGetLevel(HINCMSG inc, int *levelno);

/* get the hint parameter that was set during object registration */
int yamiAgentIncomingMsgGetHint(HINCMSG inc, void **phint);

/* eat the message - just mark it as processed           */
/* so that no further action (reply, reject) is possible */
int yamiAgentIncomingMsgEat(HINCMSG inc);

/* respond with rejection notification */
int yamiAgentIncomingMsgReject(HINCMSG inc);

/* reply to the incoming message */
int yamiAgentIncomingMsgReply(HINCMSG incoming, HPARAMSET paramset);

/* forward the message to another object */
int yamiAgentIncomingMsgForward(HINCMSG inc,
          const char *domainname, const char *objectname,
          const char *messagename, HPARAMSET paramset);

/* forward the message to another object, with explicit address */
int yamiAgentIncomingMsgForwardAddr(HINCMSG inc,
          const char *domainaddr, int domainport, int domainlevel,
          const char *objectname, const char *messagename,
          HPARAMSET paramset);

/* destroy the incoming message token */
int yamiAgentDestroyIncomingMsg(HINCMSG inc);

/*
 * additional functions (provided for convenience)
 */

/* creates a minimalistic ad-hoc agent and sends   */
/* a one-way message to the given address          */
int yamiEasySend(const char *address, int port, int level,
          const char *objectname, const char *messagename,
          HPARAMSET paramset);

/* prints in the provided buffer the local address */
/* in the xxx.xxx.xxx.xxx form                     */
int yamiAgentGetLocalAddress(HYAMIAGENT agent, char *pbuf);

/* retrieves the actual port number                */
/* used by the listening socket                    */
int yamiAgentGetLocalPort(HYAMIAGENT agent, int* pport);

/* this is for the explicit event loop               */
/* normally, the responsibilities of this function   */
/* are taken by the receiver thread                  */
/* this function is for those who like experimenting */
/* and for platforms where no threading is possible  */
/* returns YAMINOTFOUND when there was no socket     */
/* and YAMINOTAVAILABLE when there was no event      */
int yamiAgentProcessEvent(HYAMIAGENT agent, int timeout);

/* network initialization                           */
/* (for the Windows platform and others,            */
/* that could possibly require such initialization) */
/* (on the Unix platform, this is empty function)   */
int yamiNetInitialize(void);

/* network cleanup */
/* (same remarks)  */
int yamiNetCleanup(void);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
